#include <algorithm>
#include <cstring>
#include <ctime>
#include <cerrno>
#include <cstdio>
#include <unistd.h>
#include <sys/ioctl.h>
#include <poll.h>
#include <iostream>
#include "logger.h"
#include "Fifo.h"

#define err_msg(fmt, args...) \
    if (errno) { \
        printf("***Error [%s:%d %s] " fmt "\n",  \
                __PRETTY_FUNCTION__, __LINE__, strerror(errno), ## args); \
    } \
    else {\
        printf("***Error [%s:%d] " fmt "\n",  \
                __PRETTY_FUNCTION__, __LINE__, ## args); \
    }


struct DAQpp::FifoPriv
{
        AtomicInt _in_loop;
        AtomicInt _end_poll;
        FifoPriv() : _in_loop(0), _end_poll(0) {}
};


bool DAQpp::DataCondition::condition()
{
    return !(fifo->empty());
}

void DAQpp::FifoThread::run()
{
    struct pollfd PS = { fifo->fd, POLLIN | POLLPRI | POLLERR, 0 };

    fifo->endpoll(false);
    while (true)
    {
        fifo->_priv->_in_loop.store(1);
        int iok = poll(&PS, 1, 1000);
        if (iok==1)
        {
            fifo->read_from_channel();
        }
        else
        {
            if (iok<0)
            {
                if (errno==EINTR)
                {
                    if (fifo->endpoll())
                        return;
                }
                else
                {
                    err_msg("error polling for data");
                    return;
                }
            }
            else if (fifo->endpoll())
            {
                return;
            }
        }

    }
    return;
}

static int _power_of_2_size(int count)
{
    int psz = 1;
    while (psz < count)
        psz *= 2;

    return psz;
}



DAQpp::Fifo::Fifo(unsigned int sz) :
    condition(this), thread(this), fd(-1), _priv(0), bfsiz(sz), buff(0)
{

    bfsiz = _power_of_2_size(sz);
    buff = reinterpret_cast<char *>(realloc(buff, bfsiz * sizeof(char)));
    rdptr = wrptr = buff;
    end = buff + bfsiz;
    _priv = new FifoPriv();
}

DAQpp::Fifo::~Fifo()
{
    detach();
    free(buff);
    delete _priv;
}

bool DAQpp::Fifo::running() const
{
    return _priv->_in_loop();
}

/// Tells if the Fifo should stop polling for data from the open file
bool DAQpp::Fifo::endpoll() const
{
    return _priv->_end_poll();
}

/// Make the Fifo stop polling for data from the open file
void DAQpp::Fifo::endpoll(bool x)
{
    _priv->_end_poll = (x ? 1 : 0);
}


void DAQpp::Fifo::resize(unsigned int sz)
{
    acquire();
    if (sz>bfsiz)
    {
        grow(sz-bfsiz);
    }
    release();
}

int DAQpp::Fifo::read(char *data, int count)
{
    int rc = 0;
    int rdtype = -1;
    condition.wait();

    acquire();
    if (wrptr > rdptr)
    {
        // write pointer has not wrapped yet
        if (count <= wrptr - rdptr)
        { // all the data is there
            rdtype = 1;
            memcpy(data, rdptr, count);
            rc = count;
            rdptr += count;
        }
        else
        { // some data is missing
            rdtype = 2;
            rc = wrptr - rdptr;
            memcpy(data, rdptr, rc);
            rdptr += rc;
        }
    }
    else
    { // we may have to wrap
        int left = (end - rdptr) + (wrptr-buff);
        if (count <= left)
        { // all the data is there
            if (count < end - rdptr)
            { // no need to wrap
                rdtype = 3;
                memcpy(data, rdptr, count);
                rdptr += count;
            }
            else
            { // we do need to wrap
                rdtype = 4;
                rc = end - rdptr;
                memcpy(data, rdptr, rc);
                rdptr = buff;
                memcpy(data + rc, rdptr, count - rc);
                rdptr += count - rc;
            }
            rc = count;
        }
        else
        { // only some data is there
            rdtype = 5;
            int sz;
            rc = end - rdptr;
            memcpy(data, rdptr, rc);
            rdptr = buff;
            sz = wrptr - rdptr;
            memcpy(data + rc, rdptr, sz);
            rdptr += sz;
            rc += sz;
        }
    }
    if (rc != count )
    {
        DAQpp::ostream os;
        os << loglevel(DAQpp::Log::warning )<< "***** Fifo::read (type " << rdtype << ") - expect problems " << rc << " != " << count << std::endl;
    }
    release();
    return rc;
}

void DAQpp::Fifo::grow(int count)
{
    // we make it a power of 2
    unsigned int psz = 1;
    while (true)
    {
        if ( psz < count + bfsiz )
            psz *= 2;
        else
            break;
    }
    if (psz > 0x80000 )
        return;

    count = psz - bfsiz;

    int nsz = bfsiz + count;
    int nw = (wrptr - buff);
    int nr = (rdptr - buff);
    int nt = data_available();
    if (wrptr > rdptr)
    {
        buff = reinterpret_cast<char *>(realloc(buff, nsz));
        if (buff==0)
        {
            err_msg("Ran out of memory");
            exit(1);
        }
        bfsiz = nsz;
        end = buff + bfsiz;
        wrptr = buff + nw;
        rdptr = buff + nr;
    }
    else
    {
        char *nbuff = reinterpret_cast<char *>(realloc(0, nsz));
        if (nbuff==0)
        {
            err_msg("Ran out of memory");
            exit(1);
        }
        memcpy(nbuff, rdptr, (end-rdptr));
        memcpy(nbuff+(end-rdptr), buff, (wrptr-buff));
        free(buff);
        buff = nbuff;
        bfsiz = nsz;
        end = buff + bfsiz;
        rdptr = buff;
        wrptr = buff+nt;
    }
}

int DAQpp::Fifo::write(char *data, int count, bool force)
{
    int rc = 0, left, sz;

    acquire();
    left = space_left();
    if (left <= count)
    {
        if ( !force)
        {
            release();
            return rc;
        }

        // we need to enlarge the buffer
        grow(count);
    }

    if (wrptr >= rdptr)
    { // The write pointer has not wrapped yet
        if (count < end - wrptr)
        { // no need to wrap
            memcpy(wrptr, data, count);
            wrptr += count;
        }
        else
        { // we will have to wrap
            sz = end - wrptr;
            memcpy(wrptr, data, sz);
            wrptr = buff;
            memcpy(wrptr, data + sz, count - sz);
            wrptr += count - sz;
        }
        rc = count;
    }
    else
    {
        // The write pointer is smaller than the read pointer.
        // We check that we can still write without overwriting
        if ( rdptr - wrptr > count)
        {
            memcpy(wrptr, data, count);
            wrptr += count;
            rc = count;
        }
        else
        {   // we are going to overwrite...
            // refuse to write
            DAQpp::ostream os;
            os << loglevel(DAQpp::Log::error) << "Fifo::write refused to write past the read pointer" << std::endl;
            rc = 0;
        }

    }

    condition.notify();
    release();
    return rc;
}


/// Returns the space left in the FIFO buffer
int DAQpp::Fifo::space_left()
{
    int left;
    acquire();
    if (rdptr == wrptr)
    {
        left = bfsiz;
    }
    else
    {
        left = ((rdptr + bfsiz - wrptr) % bfsiz);
    }
    release();
    return left;
}

/// Returns the ammount of data not yet read from the FIFO
int DAQpp::Fifo::data_available()
{
    int left;
    acquire();
    if (wrptr >= rdptr)
    { // no wrap yet
        left = wrptr - rdptr;
    }
    else
    {
        left = (end - rdptr) + (wrptr - buff);
    }
    release();
    return left;
}



int DAQpp::Fifo::attach(int fildes)
{
    fd = fildes;
    _priv->_in_loop.store(0);
    thread.start();
    // wait until the thread is actually polling
    while (!_priv->_in_loop())
    {
    	DAQpp::wait(0.00001);
    }
    return 0;
}

void DAQpp::Fifo::detach()
{
    endpoll(true);
    thread.stop();
}

void DAQpp::Fifo::read_from_channel()
{
    int rc, sz, left, nread=0;

    ioctl(fd, FIONREAD, &nread);
    if (nread<=0)
        return;

    acquire();
    left = space_left();
    if (left<=nread)
    {
        grow(nread);
    }

    if (wrptr >= rdptr)
    { // The write pointer has not wrapped yet
        if (nread < end-wrptr)
        {
            // no need to wrap yet
            rc = ::read(fd, wrptr, nread);
            if (rc>0)
            {
                wrptr += rc;
            }
        }
        else
        {
            // we will have to wrap
            sz = end-wrptr;
            rc = ::read(fd, wrptr, sz);
            wrptr = buff;
            rc = ::read(fd, wrptr, nread-sz);
            wrptr += (nread-sz);
        }
    }
    else
    {
        rc = ::read(fd, wrptr, nread);
        wrptr += nread;
    }
    condition.notify();
    release();
}
