#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <iostream>
#include <sstream>
#include <iomanip>
#include <csignal>
#include <cerrno>
#include <cstring>
#include <cstdio>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/poll.h>

#include "logger.h"
#include "FileIOManager.h"

using namespace DAQpp;


void DAQpp::FileIOManager::IOthread::run()
{
    if (iom)
        iom->spy_size();
}

DAQpp::FileIOManager::FileIOManager(bool use_compress, unsigned long mx_size)
        : fd(-1), zfd(0), thread(this),
        extension("data"), ifile(-1),
        zcomp(use_compress),
        _size(0), _max_size(mx_size),
        doSpy(true), get_header(0), header_data(0),
        get_tail(0), tail_data(0)
{
}
DAQpp::FileIOManager::~FileIOManager()
{
    if ( thread.is_running() )
        thread.stop();

}

std::string DAQpp::FileIOManager::new_name()
{
    std::ostringstream ostr;

    ostr << base_name << '_'
         << std::setw(3) << std::setfill('0') << ifile
         << '.'
         << extension
         ;
    return ostr.str();
}

int DAQpp::FileIOManager::write_header_tail( HeaderFunc ff, char *dt )
{
    int rc = 0;
    if ( ff )
    {
        const char *data;
        _size = (*ff)(&data, dt);
        if ( _size && data )
        {
            rc = (zcomp
                  ? gzwrite(zfd, (void *)data, _size)
                  : ::write(fd, (void *)data, _size)
                 );
            delete [] data;
        }
    }
    return rc;
}

int DAQpp::FileIOManager::new_file()
{
    int rc;
    if ( fd >= 0 )
    {
        write_header_tail(get_tail, tail_data);
        rc = (zcomp ? gzclose(zfd) : ::close(fd));
    }

    ifile++;
    std::string onam = new_name();
    file_name = onam;

    fd = creat(onam.c_str(), S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);

    if (fd == -1)
        return fd;

    if (zcomp)
    {
        zfd = gzdopen(fd, "wb2");
        if (zfd == NULL)
        {
            fd = -1;
            return fd;
        }
    }

    write_header_tail(get_header, header_data);
    return 0;
}

int DAQpp::FileIOManager::open(const std::string &fnam)
{

    if (fd >= 0)
        return -1;

    // build the file name
    unsigned int e = fnam.find_last_of(".");
    if ( e == fnam.npos )
    {
        base_name = fnam;
        extension = "data";
    }
    else
    {
        base_name = std::string(fnam, 0, e);
        extension = std::string(fnam, e + 1, fnam.npos);
    }
    ifile = -1;

    // make a local copy of the header

    // open the file
    int rc = new_file();
    if ( rc )
    {
        std::cerr << "FileIOManager::open failed with " << fnam << std::endl;
        return -1;
    }
    thread.start();
    return rc;
}

int DAQpp::FileIOManager::close()
{
    int rc = -1;

    if ( fd < 0 )
        return rc;

    doSpy = false;

    mtx.acquire();
    write_header_tail(get_tail, tail_data);

    if ( fd >= 0 )
        rc = (zcomp ? gzclose(zfd) : ::close(fd));
    fd = -1;
    zfd = NULL;
    mtx.release();

    if ( thread.is_running() )
        thread.stop();


    return rc;
}


int FileIOManager::write(const char *ptr, size_t bflen)
{
    int rc;
    mtx.acquire();
    if (zcomp)
    {
        rc = gzwrite(zfd, (void *)ptr, bflen);
    }
    else
    {
        rc =::write(fd, (void *)ptr, bflen);
    }
    mtx.release();
    return rc;
}

void DAQpp::FileIOManager::spy_size()
{
//    spy_size_intr();

    struct stat buf;
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 25000;


    while (fd >= 0)
    {
        if (select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &t) != 0)
        {
            if (errno != EINTR)
            {
                DAQpp::ostream os;
                os << loglevel(DAQpp::Log::error)
                   << "Error in FileIOManager::spy_size"
                   << strerror(errno)
                   << std::endl;
            }
        }
        fstat(fd, &buf);
        _size = buf.st_size;
        if (_size > _max_size)
        {
            mtx.acquire();
            new_file();
            mtx.release();
        }
    }
}
