/* -*- mode: c++ -*- */
#ifndef __DAQpp_FileIOManager_h__
#define __DAQpp_FileIOManager_h__

#include <zlib.h>
#include <DAQ++/Thread.h>
#include <fcntl.h>
#include <string>
#include <DAQ++/exception.h>

namespace DAQpp
{

    class FileHeader;

    /**
     *  \brief Class FileIOManager
     *
     *  \ingroup IO
     *
     *  This class takes care of data logging into a file. It provides the
     *  possibility of compressing the data to reduce the file size. To
     *  compress it uses zlib.
     *
     *  One should specify a maximum size for the output file in the
     *  constructor. FileIOManager will split the data into a sequence of
     *  files that will share a common base name and will have appended in
     *  the name the serial number.
     *
     *  \author Carlos Lacasta
     */

    class FileIOManager
    {
        public:
            typedef int (*HeaderFunc)(const char **, char *);

        private:
            /// File handler
            int fd;
            /// Compression file handle
            gzFile zfd;

            /// Mutex
            Mutex mtx;
            /** thread to spy the file size
             */
            class IOthread : public Thread
            {
                private:
                    /// A pointer to the FileIOmanager
                    FileIOManager *iom;
                public:
                    /// Constructor
                    IOthread(FileIOManager *_io) : iom(_io)
                    {}
                    /// method to run in the thread
                    void run();

            };
            friend class IOthread;

            /// This is the thread that will spy the file size
            IOthread thread;

            /// File name
            std::string base_name;

            /// File extension
            std::string extension;

            /// File index
            int ifile;

            /// flag to activate data compression
            bool zcomp;

            /// Current file size
            unsigned long _size;

            /// Maximum file size
            unsigned long _max_size;

            /// actual file name
            std::string file_name;

            /// spies the size of the current file
            void  spy_size();

            /// creates a new file in the file sequence
            int    new_file();



            /** \brief spying flag.
             *   If true, FileIOmanager will spy the size of the file to
             *   check that it is not bigger than the maximum  allowed
             */
            bool doSpy;

            /// calls the header/tail function
            int write_header_tail( HeaderFunc , char *);


        protected:
            /**
             * This is for child classes
             */
            void set_file_name(const std::string &fnam)
            {
                file_name = fnam;
            }
            void set_size(unsigned long val)
            {
                _size = val;
            }
        public:
            /** called when a new file is opened.
             *  Every time a new file is opened this fuction is called.
             */
            HeaderFunc get_header;

            /// Data that will be passed to HeaderFunc
            char *header_data;

            /** called when a file is about to close.
             *  Before a file is closed this function is closed
             */
            HeaderFunc get_tail;

            // Data that will be passed to HeaderFunc
            char *tail_data;

        public:
            /// constructor
            FileIOManager(bool use_compression, unsigned long mx_size );

            /// destructor
            virtual ~FileIOManager();

            /// Opens a new file with a given base name
            virtual int open(const std::string &);

            /// Closes the file sequence
            virtual int close();

            /// Writes data into the file
            virtual int write(const char *, size_t);

            /// changes the base name of the file
            virtual std::string new_name();

            /// returns the serial number of the output file
            int file_number()
            {
                return ifile;
            }

            /// sets the maximum size of each file in the sequence
            void set_max_size(unsigned long x)
            {
                _max_size = x;
            }

            /// Returns the maximum size for an output file
            unsigned long get_max_size() const
            {
                return _max_size;
            }
            /// returns the size of the current file
            unsigned long size()
            {
                return _size;
            }

            /// returns the relative size of the current file
            double relative_size()
            {
                return ((double)_size) / ((double)_max_size);
            }

            /// Sets the compresion option
            void Compress(bool t)
            {
                zcomp = t;
            }

            /// Returns the compression option
            bool Compress()
            {
                return zcomp;
            }

            /// Returns the name of the current file
            const std::string &get_file_name() const
            {
                return file_name;
            }

            /// Sets the header "getter"
            void set_header_func( HeaderFunc ff, char *dt )
            {
                header_data = dt;
                get_header = ff;
            }

            /// removes the header "getter" pointer
            void reset_header_func()
            {
                get_header = 0;
            }

            /// Sets the tail "getter"
            void set_tail_func( HeaderFunc ff, char *dt )
            {
                tail_data = dt;
                get_tail = ff;
            }

            // removes the tail "getter"
            void reset_tail_func()
            {
                get_tail = 0;
            }
    };



}

#endif
