#include <iostream>
#include <cstring>
#include <cerrno>
#include <DAQ++/exception.h>

using namespace DAQpp;

#if !defined(USE_MUTEX)
#define USE_SEM
#endif

#ifdef USE_SEM
#include <sys/ipc.h>
#include <sys/sem.h>

#define SEM_KEY ((key_t)0xface)
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/* union semun is defined by including <sys/sem.h> */
#else
/** according to X/OPEN we have to define it ourselves */
#if defined(_POSIX_C_SOURCE) && !defined( _DARWIN_C_SOURCE)
union semun {
    int val;                    /*!< value for SETVAL */
    struct semid_ds *buf;       /*!< buffer for IPC_STAT, IPC_SET */
    unsigned short int *array;  /*!< array for GETALL, SETALL */
    struct seminfo *__buf;      /*!< buffer for IPC_INFO */
};
#endif
#endif
#endif /* USE_SEM */

#include "IOhandler.h"


IOhandler *IOhandler::instance = 0;

IOhandler::IOhandler(int s):
    signum(s), nTrg(0.), newData(false), func(0), ptr(0)
{
    actIO.sa_handler = SIG_DFL;
    sigemptyset (&actIO.sa_mask);
    actIO.sa_flags = SA_RESTART;

#ifdef USE_SEM
    sem =  semget ((key_t) SEM_KEY, 1, 0600 | IPC_CREAT);
    if ( sem == -1 )
    {
        std::cerr << "Error initializing IOhandler: "
                  << strerror(errno)
                  << std::endl;
    }
    else
    {
        union semun data;
        unsigned short val = 1;
        data.array = &val;
        semctl (sem, 0, SETALL, data);
    }
#else
    pthread_mutex_init(&mtx, 0);
#endif
}

IOhandler::~IOhandler()
{
    if ( actIO.sa_handler != SIG_DFL)
        Unregister();

#ifdef USE_SEM
    semctl (sem, 0, IPC_RMID, 0);
#else
    pthread_mutex_destroy(&mtx);
#endif
}

bool IOhandler::trylock()
{
    int rc;
#ifdef USE_SEM
    rc = 0;
    return false;
#else
    rc = pthread_mutex_trylock(&mtx);
    return errno == EBUSY && rc == 0 ? true : false;
#endif
}

int IOhandler::lock ()
{
#ifdef USE_SEM
    int rc;
    struct sembuf pbuf;

    pbuf.sem_num = 0;
    pbuf.sem_op = -1;
    pbuf.sem_flg = SEM_UNDO;

    do
    {
        rc = semop (sem, &pbuf, 1);
    }
    while (rc == -1 && errno == EINTR);
    if (rc)
        std::cerr << "error locking IOhandler: "
                  << strerror(errno)
                  << std::endl;

    return rc;
#else
    return pthread_mutex_lock(&mtx);
#endif
}

int IOhandler::unlock()
{
#ifdef USE_SEM
    int rc;
    struct sembuf pbuf;

    pbuf.sem_num = 0;
    pbuf.sem_op = 1;
    pbuf.sem_flg = SEM_UNDO;

    do
    {
        rc = semop (sem, &pbuf, 1);
    }
    while (rc == -1 && errno == EINTR);
    if (rc)
        std::cerr << "Error unlocking IOhandler: "
                  << strerror(errno)
                  << std::endl;

    return rc;
#else
    return pthread_mutex_unlock(&mtx);
#endif
}

void IOhandler::PrepareForData()
{
    newData = false;
    lock ();
}

void IOhandler::WaitForData()
{
    lock ();
    unlock();
}

void IOhandler::Register( void (*f)(void *), void *p )
{
    struct sigaction oact;
    instance = this;
    func = f;
    ptr = p;
    actIO.sa_handler = gotIO;
    sigaction (signum, &actIO, &oact);
}
void IOhandler::Unregister()
{
    struct sigaction oact;
    instance = 0;
    func = 0;
    actIO.sa_handler = SIG_DFL;
    sigaction (SIGIO, &actIO, &oact);
}

void IOhandler::gotIO(int x)
{
    if ( x != SIGIO )
        return;

    if ( instance )
    {

        if ( instance->func )
            instance->func(instance->ptr);
        
        instance->nTrg++;
        instance->newData = true;

        instance->unlock();
    }
}

