/* -*- mode: c++ -*- */
#ifndef __DAQpp_IOhandler_h__
#define __DAQpp_IOhandler_h__
#include <csignal>
#include <pthread.h>

namespace DAQpp
{
    /**
     *  \brief class IOhandler
     *  \ingroup IO
     * 
     *  This class is a sort of wrap-around the sigaction call to
     *  set a handler on a given signal. One example of use is to register to
     *  the SIGIO signal to detect when there is data available
     */
    class IOhandler
    {
        private:
            /// Specifies the signal to catch
            int signum;
            /// number of signal interrupts
            double nTrg;
            /// set to true when an interrupt occurs
            bool   newData;
#ifdef USE_MUTEX
            /// Mutex
            pthread_mutex_t mtx;
#else
            /// semaphore
            int    sem;
#endif
            /// internal sigaction structure
            struct sigaction actIO;

            /// callback function
            void (*func)(void *);
            /// pointer
            void *ptr;
            /// instance
            static IOhandler *instance;

            /// lock
            int lock ();
            /// check if it is locked
            bool trylock();
            /// unlock
            int unlock();
        public:
            /// Constructor
            IOhandler(int s);
            /// Destructor
            ~IOhandler();

            /// Registers to a IOhandler
            void   Register( void (*f)(void *), void *p );
            /// Unregister
            void   Unregister();
            /// resets the interrupt handler
            void   ResetTrg()
            {
                nTrg = 0.;
                newData = false;
            }
            /// Prepares to catch interrupts
            void   PrepareForData();
            /// returns the number of interrupts
            double GetNtrg()
            {
                return nTrg;
            }
            /// returns true if a new interrupt has arrived
            bool   Data()
            {
                return newData;
            }
            /// waits for an interrupt
            void   WaitForData();
            /// actial signal catcher
            static void gotIO(int);
    };
}
#endif
