#include <iostream>
#include <iomanip>
#include <unistd.h>
#include <ctime>
#include <cstring>
#include "logger.h"
#include "DAQ++/Monitor.h"
#include "DAQ++/MutexPriv.h"
//#define DAQpp_DEBUG 1
#if defined(DAQpp_DEBUG)
#define WHOAMI(str) { \
        DAQpp::ostream os; \
        os << loglevel(DAQpp::Log::debug) << str << "::" << __FUNCTION__ \
           << " - [thread " << std::hex << pthread_self() << std::dec \
           << "]" << std::endl; \
}
#else
#define WHOAMI(str)
#endif

#define DBG_MSG(str) { \
        DAQpp::ostream os; \
        os << loglevel(DAQpp::Log::debug) << __FUNCTION__ << ": " << str \
           << " - [thread " << std::hex << pthread_self() << std::dec \
           << "]" << std::endl << std::flush; \
}

using namespace DAQpp;

std::allocator<char> DAQpp::Event::data_allocator;
std::allocator<Event> DAQpp::Event::event_allocator;

Event *Monitor::MonitorDie = (Event *)0xD1E00D1E;
Event *Monitor::MonitorTimeOut = (Event *)0xD1E00AAA;



Event::Event(int s, const char *d)
        : _size(s), _data(0)
{
    if (d)
    {
        _data = data_allocator.allocate(_size);
        memcpy(_data, d, _size);
    }
}
Event::~Event()
{
    free_data();
}

void Event::cpy(const Event &e)
{
    free_data();
    _size = e._size;
    if ( e._data )
    {
        _data = data_allocator.allocate(_size);
        memcpy(_data, e._data, _size);
    }
    else
        _data = 0;
}

void __do_clean(std::deque<Event *> &Q)
{
    std::deque<Event *>::iterator ip;
    for (ip=Q.begin(); ip!=Q.end(); ++ip)
    {
        Event *evt = *ip;
        if ( evt != Monitor::MonitorDie && evt!=Monitor::MonitorTimeOut )
            delete evt;
    }
    Q.clear();
}

Monitor::Monitor(int sz)
        : mxsize(sz), user_timeout(-1)
{
    _privm = new MutexPriv_t();
}

Monitor::~Monitor()
{
    while (!empty())
    {
        Event *ev = front();
        pop_front();
        if ( ev != MonitorDie && ev != MonitorTimeOut)
            delete ev;
    }
    delete _privm;
}

bool Monitor::condition()
{
    bool rc;
    WHOAMI(">>> Monitor")
    _privm->acquire();
    rc = !empty();
    _privm->release();
    WHOAMI("<<< Monitor")
    return rc;
}

bool Monitor::full()
{
    bool rc;

    _privm->acquire();
    rc = size() > mxsize;
    _privm->release();

    return rc;
}


void Monitor::push_event( Event *ev )
{
    WHOAMI(">>> Monitor")
    _privm->acquire();
    if ( size() <= mxsize || ev == MonitorDie )
    {
        push_back(ev);
        notify_all();
    }
    else
        delete ev;

    _privm->release();
    WHOAMI("<<< Monitor")
}


class ThreadCancelState
{
    private:
        Thread *thr;
    public:
        ThreadCancelState(Thread *T) : thr(T)
        {
            thr->set_cancel(Thread::disable);
        }
        ~ThreadCancelState()
        {
            thr->set_cancel(Thread::deferred);
        }
        void dummy() {}
};

Event *Monitor::get_event(int tmout)
{
    Event *ev;
    ThreadCancelState T(this);

    T.dummy();

    if (!wait(tmout))
        return MonitorTimeOut;

    WHOAMI(">>> Monitor");
    _privm->acquire();
    ev = front();
    pop_front();
    _privm->release();
    WHOAMI("<<< Monitor");

    return ev;
}


void Monitor::clean()
{
    WHOAMI(">>> Monitor");
    _privm->acquire();
    __do_clean(*this);
    _privm->release();
    WHOAMI("<<< Monitor");
}


void Monitor::start_monitor()
{
    if (is_running())
    {
        DAQpp::ostream os;
        os  << loglevel(DAQpp::Log::warning)
            << "Monitor::start_monitor - "
            << "there is already an instance of this monitor running\n"
            << "Trying to stop and start."
            << std::endl;
        stop_monitor();
    }
    WHOAMI(">>> Monitor");
    clean();
    start();
    on_start();
}

void Monitor::stop_monitor()
{
    WHOAMI(">>> Monitor");
    // Send the stop signal
    _privm->acquire();
    push_front(MonitorDie);
    _privm->release();

    notify_all();

    stop();
    clean();
    WHOAMI("<<< Monitor");
    on_stop();
}

void Monitor::run()
{
    Event *evt = 0;
    int n_timeout = 0;
    bool loop=true;

    while (loop)
    {
        // Get the event
        evt = get_event(user_timeout);

        // Check if the thread should finish
        test_cancel();

        // Check if there has been a timeout
        if (evt == Monitor::MonitorTimeOut)
        {
            n_timeout++;
            if (n_timeout>25)
            {
                DAQpp::ostream os;
                os << loglevel(DAQpp::Log::warning)
                   << "*** Repeated Monitor Timed out. Quitting the monitor ***"
                   << std::endl;
                push_front(MonitorDie);
                notify_all();
            }

            continue;
        }
        n_timeout=0;
        // We have been told to stop
        if (evt == Monitor::MonitorDie)
            break;

        // The queue is empty
        if (evt == 0)
            continue;

        // Call the user function
        loop = user_run(evt);

        // delete the event
        delete evt;
    }
    exit();
}

