#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstdlib>
#include <unistd.h>
#include <csignal>
#include <cerrno>
#include <cstring>
#include <cstdio>
#include <ctime>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "Progress.h"

namespace DAQpp
{

    static char sbuf[4096];
    int Progress::page_k_shift = 0;

    static void timeAlarm(int)
    {
        signal(SIGALRM, timeAlarm);
    }

    typedef std::pair<double, std::string> timeString;

    static inline std::pair<double, std::string> time_units(double t)
    {
        timeString rc(t, "s");
        if (rc.first>86400.)
        {
            rc.first /= 86400.;
            rc.second = "d";
        }
        else if (rc.first>3600.)
        {
            rc.first /= 3600.;
            rc.second = "h";
        }
        else if (rc.first>60.)
        {
            rc.first/=60.;
            rc.second = "m";
        }
        else if (rc.first>1)
        {
            rc.second = "s";
        }
        else if (rc.first>1.e-3)
        {
            rc.first /= 1.e-3;
            rc.second = "ms";
        }
        else if (rc.first>1.e-6)
        {
            rc.first /= 1.e-6;
            rc.second = "us";
        }
        else
        {
            rc.first /= 1.e-9;
            rc.second = "ns";
        }
        return rc;
    }

    std::string Progress::print_time(double t)
    {
        std::ostringstream ostr;
        timeString units = time_units(t);
        ostr.setf(std::ios::fixed);
        ostr << std::setprecision(2) << units.first << " " << units.second;
        return ostr.str();
    }

    void Progress::get_proc_filename()
    {

        std::ostringstream ostr;
        ostr << "/proc/" << getpid() << "/statm";
        proc_file = ostr.str();
    }

    Progress::Progress(double a, double b, int opt) :
        flg(opt), usec(500000), thread(this), ss(true), last_value(a), last_update(0.),
                proc_file("")
    {
        reset(a, b);

        if (page_k_shift == 0)
        {
            int j;
            for (j = getpagesize(); j > 1024; j >>= 1)
                page_k_shift++;
        }

        get_proc_filename();
    }

    Progress::Progress(const Progress &p)
        : flg(0), thread(this), mm0(0.0), active(false)
    {
        reset(p.vmin, p.vmax);
        usec = p.usec;
        ss = false;

        get_proc_filename();
    }

    Progress::~Progress()
    {
    }

    void Progress::reset(double a, double b)
    {
        value = 0;
        vmin = a;
        vmax = b;
        if (vmin > vmax)
        {
            double tmp = vmin;
            vmin = vmax;
            vmax = tmp;
        }
    }

    void Progress::start()
    {
        ss = true;
        if (flg & timing)
            timer_start();

        mm0 = mem();
        thread.start();
    }

    void Progress::stop()
    {
        ss = false;
        if (flg & timing)
            timer_stop();

        thread.stop();
        std::cout << std::endl;
    }

    std::string Progress::str()
    {
        int width = -1;
        const char *sc = getenv("COLUMNS");
        if (sc)
        {
            std::istringstream istr(sc);
            istr >> width;
            if (!istr)
                width = 80;
        }
        else
            width = 80;

        double now = time();
        std::string time_str = print_time(now);
        std::ostringstream out;
        out.setf(std::ios::fixed);
        out << '\r';
        if (flg & show_average)
        {
            timeString ts = time_units(value>0. ? now/(value) : 0.); //  dt/du;
            out << std::setw(5) << std::setprecision(2) << ts.first << " " << ts.second
                    <<"/upd ";
            last_value = value;
            last_update = now;
        }
        if (flg & show_time)
        {
            out << "Time: " << time_str << " ";
        }
        if (flg & show_left)
        {
            double tf = Percent();
            if (tf != 0.)
                tf = (1. - tf) * time() / tf;

            timeString ts = time_units(tf);
            out << "Left: " << std::setw(5) << std::setprecision(1) << ts.first
                    << ts.second <<" ";
        }
        if (flg & show_memory)
        {
            out << "Mem.: " << std::setw(5) << std::setprecision(1) << (mem())/1024.
                    << " Mb ";
        }
        if (flg & show_status)
        {
            int ncol = width-15-out.str().size();
            double fact = (value - vmin) / (vmax - vmin);
            if (fact>1.)
                fact=1.;

            int i, icol = int(ncol * fact);
            out << " [";
            for (i = 0; i < icol; i++)
                out << '#';
            for (; i < ncol; i++)
                out << ' ';
            out << "] ";

            if ( !(flg&show_value))
            {
                out << std::setw(5) << std::setprecision(1) << 100.*(value-vmin)/(vmax
                        -vmin) << " % ";
            }
            else
            {
                out << std::setw(8) << std::setprecision(1) << value;
            }

        }
        return out.str();
    }

    void Progress::update()
    {
        last_value = value;
        last_update = time();
    }

    void Progress::show()
    {
        if (flg == show_nothing)
            return;

        std::cout << str() << std::flush;

        update();
    }

    void Progress::PrgThread::run()
    {
        while (prg->ss)
        {
        	DAQpp::wait(prg->usec*1000000);
            prg->show();
        }
    }

    int Progress::read_file() const
    {
        int fd = open(proc_file.c_str(), O_RDONLY);
        if (fd < 0 && !proc_file.empty())
        {
            std::cout << "\n\nCould not open " << proc_file << std::endl << std::endl;
            return -1;
        }
        int iok, r = 0;
        *sbuf = 0;
        do
        {
            iok = read(fd, sbuf, 4096);
            if (iok > 0)
                r += iok;
        } while (iok < 0 && errno == EINTR);
        close(fd);
        return r;
    }

    double Progress::mem() const
    {
        int statlen;
        statlen = read_file();
        if (statlen > 0)
        {
            unsigned int size;
            //       istringstream istr(sbuf,sizeof(sbuf));
            //       istr >> size;
            sscanf(sbuf, "%u", &size);
            size <<= page_k_shift;
            return ((double)size);
        }
        return 0.;
    }

}

