/* -*- mode: c++ -*- */
#ifndef __Progress_h__
#define __Progress_h__

#include  <DAQ++/Thread.h>
#include <unistd.h>
#include <sys/time.h>
#include <iosfwd>

namespace DAQpp
{
    /**
     *  \brief class Progress.
     *  \ingroup Utils
     *
     *  This class is used to show the progress of a given activity.
     *  It is a generic class and the way the progress is shown depends
     *  on how the show method is implemented.
     */
    class Progress
    {
        private:
            /// flag indicating what to show
            unsigned int flg;
            /// lower limit of the range
            double vmin;
            /// maximum value of the range
            double vmax;
            /// current value
            double value;
            /// frequency of of update of the progress display
            unsigned long usec;
            /**
             *  Class to for the thread of the show method
             */
            class PrgThread : public Thread
            {
                private:
                    /// A pointer to th eProgress instance
                    Progress *prg;
                public:
                    /// constructor
                    PrgThread(Progress *p) :
                        prg(p)
                    {
                    }
                    // statements to be run on the thread
                    void run();
            };
            /// Instance of PrgThread
            PrgThread thread;
            /// flag set internally to true if the Progress monitor is running
            bool ss;

            /// internal variables
            double last_value;

            /// internal variables
            double last_update;

            /// Other utilities
            /// Monitoring process memory
            std::string proc_file;

            /** Reads the /proc file corresponding to this process and
             * gets the relevant information.
             */
            int read_file() const;

            // internal stuff
            float mm0;

            // internal stuff
            static int page_k_shift;

            /// for timing
            struct timeval _start;

            /// Says if th eprogress instance is active
            bool active;

            void get_proc_filename();
        public:
            /// Information to be show
            enum information
            {
                show_nothing = 0x000, ///< Show nothing
                show_status = 0x001, ///< Shows progress
                show_value = 0x002, ///<if show_status set, shows the value
                timing = 0x0f0,
                show_time = 0x010, ///< Shows total time elapsed
                show_average = 0x020, ///< Shows average time from one update to the next
                show_left = 0x040, ///< Shows time left for completion
                memory = 0xf00,
                show_memory = 0x100 ///< Shows memory used by the process
            };

            /** Constructor.
             *
             * @param a lower limit of range
             * @param b upper limit of range
             * @param opt says waht to show. It is an OR of options in information
             */
            Progress(double a = 0., double b = 0., int opt = show_status | show_average);

            /// Copy constructor
            Progress(const Progress &p);

            /// Destructor.
            virtual ~Progress();

            /// Starts monitoring
            void start();

            /// Stops monitoring
            void stop();

            /** Defines the way the activity is shown.
             *
             * This method is the responsible of showing the progress. It has to
             * be implemented.
             */
            virtual void show();

            /// Internal
            static void *_loop(void *);

            /// Updates the current value
            void update(double x)
            {
                value = x;
            }
            
            /// Increments the current value
            void increment(double x=1.)
            {
                value += x;
            }

            /// Redefines the range
            void reset(double a, double b);

            /// returns the lower limit of the range
            double Vmin() const
            {
                return vmin;
            }

            /// returns the upper limit of the range
            double Vmax() const
            {
                return vmax;
            }

            /// returns the current value
            double Value() const
            {
                return value;
            }

            /// returns the percentage of the progress made so far
            double Percent() const
            {
                return (value -vmin) / (vmax - vmin);
            }

            /// starts the timer
            void timer_start();
            /// stops the timer
            void timer_stop();
            /// resets the timer
            void timer_reset();
            /// returns the time elapsed
            double time() const;

            /// Memory
            double mem() const;

            /// sets the update time
            void set_update(unsigned long t)
            {
                usec = t;
            }

            /// sets flags
            void set_options(unsigned int o)
            {
                flg = o;
            }

            /** returns a string representation of Progress that depends
             on the values of flags
             */
            std::string str();

            /**
             Updates some values
             */
            void update();
			static std::string print_time(double t);
    };

    inline void Progress::timer_start()
    {
        active = true;
        gettimeofday(&_start, NULL);
    }
    inline void Progress::timer_stop()
    {
        active = false;
    }
    inline void Progress::timer_reset()
    {
        gettimeofday(&_start, NULL);
    }
    inline double Progress::time() const
    {
        double total;
        struct timeval elapsed;

        if (active)
            gettimeofday(&elapsed, NULL);
        if (_start.tv_usec > elapsed.tv_usec)
        {
            elapsed.tv_usec += 1000000;
            elapsed.tv_sec--;
        }
        elapsed.tv_usec -= _start.tv_usec;
        elapsed.tv_sec -= _start.tv_sec;

        total = elapsed.tv_sec + ((double) elapsed.tv_usec) / 1.e6;
        return total;
    }
}


#endif
