#include <iostream>
#include <iomanip>
#include <DAQ++/RunState.h>
#include <DAQ++/MutexPriv.h>

std::string DAQpp::RunState::str_state[RunState::last_state] = {
            "NotDefined",
            "NotReady",
            "GettingReady",
            "Ready",
            "Starting",
            "Paused",
            "Running",
            "Stopping",
            "Stopped",
            "TimeOut"
        };

std::ostream &operator<<(std::ostream &os, const DAQpp::RunState &st)
{
    double rate = st.rate();
    std::string rate_units = " Hz";

    if ( rate > 1.e6 )
    {
        rate /= 1.e6;
        rate_units = "MHz";
    }
    else if (rate > 1000.)
    {
        rate /= 1000.;
        rate_units = "kHz";
    }


    double throughput = st.get_throughput()/st.time();
    std::string thr_units;
    if (throughput > 1048576.0)
    {
        throughput /= 1048576.0;
        thr_units = "Mb/s";
    }
    else if (throughput > 1024.0)
    {
        throughput /= 1024.0;
        thr_units = "kb/s";
    }
    else if (throughput > 1.0 )
    {
        thr_units = "b/s";
    }
    else
    {
        throughput *= 1024;
        thr_units = " mb/s";
    }

    os   << "Evts "
         << std::setiosflags(std::ios::fixed | std::ios::right)
         << std::setw(7) << st.n_trigger()
         << " time "
         << std::setw(6) << std::setprecision(1)
        << st.time()
        << " rate "
        << std::setw(8) << std::setprecision(1)
        << rate
        << rate_units
        << " eff "
        << std::setw(5) << std::setprecision(1)
        << st.eff()
        << " - "
        << std::setw(7) << std::setprecision(1)
        << throughput
        << " " << thr_units
        << std::flush;
    return os;
}


void DAQpp::RunState::cpy(const RunState &s)
{
    daqid = s.daqid;
    name = s.name;
    nexpected = s.nexpected;
    ntrig = s.ntrig;
    nwritten = s.nwritten;
    throughput = s.throughput;
    status = s.status;

    timer = s.timer;
    timer.stop();
}

double DAQpp::RunState::rate(bool average) const
{
    double r, t, t0, n;

    DAQpp::RunState *st = const_cast<RunState *>(this);

    t0 = timer();

    if ( !isRunning())
        average = true;

    if ( average )
    {
        t = t0;
        n = (double)(ntrig);
    }
    else
    {
        t = t0 - tim0;
        n = ntrig - ntrg0;
    }
    r = (t != 0. ? n / t : 0.);
    st->tim0 = t0;
    st->ntrg0 = ntrig;
    return r;
}

DAQpp::RunState::RunState()
        : name(""), daqid(""), nexpected(0), ntrig(0), nwritten(0), throughput(0), status(NotDefined),
          ntrg0(0), tim0(0.)
{
    start();
    stop();
    reset();
    _priv = new MutexPriv_t();
}

DAQpp::RunState::RunState(const DAQid &id)
        : name(""), daqid(id), nexpected(0), ntrig(0), nwritten(0), throughput(0), status(NotDefined),
          ntrg0(0), tim0(0.)
{
    start();
    stop();
    reset();
    _priv = new MutexPriv_t();
}

DAQpp::RunState::RunState(const RunState &s) : daqid("<unnamed>")
{
    cpy(s);
    _priv = new MutexPriv_t();
}


DAQpp::RunState::~RunState()
{
    deleteObservers();
    clearActionList();
    delete _priv;
}

void DAQpp::RunState::addAction( RunStateAction *a)
{
    action_list.insert(a);
}

void DAQpp::RunState::deleteAction( RunStateAction *a )
{
    action_list.erase( a );
}

void DAQpp::RunState::clearActionList()
{
    ActionList::iterator it;
    for (it = action_list.begin();it != action_list.end();++it)
        delete *it;

    action_list.clear();
}

void DAQpp::RunState::executeActionList()
{
    std::for_each(action_list.begin(), action_list.end(), executeAction());
}

void DAQpp::RunState::reset()
{
    nexpected = 0;
    ntrig = 0;
    nwritten = 0;
    throughput = 0;
    timer.reset();
}

void DAQpp::RunState::start()
{
    timer.start();
}

void DAQpp::RunState::stop()
{
    timer.stop();
}


double DAQpp::RunState::eff() const
{
    double eff;
    double expected = nexpected;
    double registered = ntrig;
    if ( expected <= 0 )
        expected = ntrig;

    if (nwritten>0)
        registered = nwritten;

    eff = (registered > 0.0 ? (registered / expected) : 0. );
    return 100.*eff;
}

// double DAQpp::RunState::get_throughput() const {
//    if ( time()==0. )
//       return 0.;
//    else
//       return throughput*sizeof(char)/1024./time();
// }

DAQpp::RunState::StatusType DAQpp::RunState::set_status(DAQpp::RunState::StatusType st)
{
    volatile StatusType old;
    _priv->acquire();
    old = status;
    status = st;
    if (status == Stopping)
    {
        timer.stop();
    }
    _priv->release();

    if ( st != old )
    {
        Arg<StatusType> arg(st);
        notifyObservers(&arg);
    }
    return old;
}

void DAQpp::RunState::update(Observable *o, Argument *arg)
{
    Arg<StatusType> *st = static_cast< Arg<StatusType> * > (arg);
    set_status( *st );
}

const DAQpp::DAQid &DAQpp::RunState::get_daqid() const
{
    return daqid;
}

const std::string &DAQpp::RunState::get_name() const
{
    return name;
}
/// sets the name of the RunState
void DAQpp::RunState::set_name(const std::string &s)
{
    name = s;
}

/// returns the number of triggers
unsigned int DAQpp::RunState::n_trigger() const
{
    return ntrig;
}
/// Increases the number of triggers
void DAQpp::RunState::new_trigger()
{
    ntrig++;
}
/// Increases the number of saved events
void DAQpp::RunState::new_written()
{
    nwritten++;
}
/// Returns the number of saved events
unsigned int DAQpp::RunState::n_written() const
{
    return nwritten;
}

/// sets the number of expected events
void DAQpp::RunState::set_expected(unsigned int val)
{
    nexpected = val;
}

void DAQpp::RunState::increase_expected(unsigned int val)
{
    nexpected += val;
}

unsigned int DAQpp::RunState::n_expected() const
{
    return nexpected;
}

double DAQpp::RunState::time() const
{
    return timer();
}
///  returns the throughput
double  DAQpp::RunState::get_throughput() const
{
    return throughput;
}
/// resets the throughput counter
void DAQpp::RunState::reset_throughput()
{
    throughput = 0;
}
/// sets the throughput counter to a given value
void DAQpp::RunState::set_throughput(unsigned int x)
{
    throughput += x;
}

void DAQpp::RunState::set_time(double seconds)
{
    timer.set(seconds);
}
/// Sets the number of triggers
void DAQpp::RunState::set_ntrigger(int val)
{
    ntrig = val;
}

// Sets teh number of written events
void DAQpp::RunState::set_nwritten(int val)
{
    nwritten = val;
}

/// Returns the current status
DAQpp::RunState::StatusType DAQpp::RunState::get_status() const
{
    return status;
}

/// Returns true if the object is running
bool DAQpp::RunState::isRunning() const
{
    volatile bool out;
    _priv->acquire();
    out = (status == Running || status == Paused);
    _priv->release();
    return out;
}

/// Returns true if the object's state is Ready
bool DAQpp::RunState::isReady() const
{
    volatile bool out;
    _priv->acquire();
    out = (status == Ready);
    _priv->release();
    return out;
}

/// Returns true if the object's state is Paused
bool DAQpp::RunState::isPaused() const
{
    volatile bool out;
    _priv->acquire();
    out = (status == Paused);
    _priv->release();
    return out;
}

/// Returns true if the object's state is NotReady
bool DAQpp::RunState::isNotReady() const
{
    volatile bool out;
    _priv->acquire();
    out = (status == NotReady);
    _priv->release();
    return out;
}

/// Returns true if the object's state is TimeOut
bool DAQpp::RunState::isTimeout() const
{
    volatile bool out;
    _priv->acquire();
    out = (status == TimeOut);
    _priv->release();
    return out;
}
