#ifdef HAS_CONFIG_H
#include <config.h>
#endif
#ifndef HAS_PYTHON
#define Py_BEGIN_ALLOW_THREADS
#define Py_END_ALLOW_THREADS
#else
#include <pythread.h>
#endif
#include <cmath>
#include <cstring>
#include <iostream>
#include <cerrno>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <DAQ++/Thread.h>
#include <DAQ++/logger.h>

#define NO_PYTHON
#ifdef NO_PYTHON
#define Py_BEGIN_ALLOW_THREADS
#define Py_END_ALLOW_THREADS
#endif

using namespace DAQpp;

struct DAQpp::MutexPriv
{
        pthread_t _owner;
        AtomicInt _count;
        //int _count;
        MutexPriv() : _owner(0), _count(0) {}
};





Mutex::Mutex() : debug(false), _priv(0)
{
    _priv = new MutexPriv();
    pthread_mutex_init(&_mtx, 0);
}

Mutex::~Mutex()
{
    pthread_mutex_destroy(&_mtx);
    delete _priv;
}
int Mutex::acquire()
{
	int rc;
    pthread_t me = pthread_self();
    if ( pthread_equal(_priv->_owner,  me ) )
    {
        //_count++;
        //__sync_fetch_and_add(&_count, 1);
        _priv->_count++;
        if ( debug )
            fprintf(stdout, "Mutex::acquire (%x) count: %d\n", this, _priv->_count());
        return _priv->_count();
    }
    Py_BEGIN_ALLOW_THREADS
    rc = pthread_mutex_lock(&_mtx);
    Py_END_ALLOW_THREADS
    if (!rc)
    {
        _priv->_owner = me;
        _priv->_count = 1;

        if (debug)
        {
            //DAQpp::ostream os;
            //os << loglevel(Log::debug) << "Mutex::acquire" << std::endl;
            fprintf(stdout, "Mutex::acquire (%x)\n", this);
        }
        return 1;
    }
    DAQpp::ostream oss;
    oss << loglevel(Log::error) << "Mutex::acquire should not return 0" << std::endl;
    return 0;
}
bool Mutex::busy()
{
    //pthread_t me = pthread_self();
    int rc       = pthread_mutex_trylock( &_mtx );
    if ( rc && errno == EBUSY )
    {
        return true;
    }
    else
    {
        // unlock and say it is free
        //pthread_mutex_unlock(&_mtx);
        release();
        return false;
    }
}
int Mutex::release()
{
    //__sync_fetch_and_sub(&_count, 1);
    volatile int V = --_priv->_count;

    if (debug)
    {
//        DAQpp::ostream os;
//        os << loglevel(Log::debug) <<  "Mutex::release" << std::endl;
        fprintf(stdout, "Mutex::release (%x) count: %d\n", this, V);
    }

    if ( V<=0 )
    {
        _priv->_owner = 0;
        _priv->_count = 0;
        Py_BEGIN_ALLOW_THREADS
        pthread_mutex_unlock(&_mtx);
        Py_END_ALLOW_THREADS
    }
    return V;
}

Condition::Condition()
{
    pthread_cond_init(&_cond, 0);
}
Condition::~Condition()
{
    pthread_cond_destroy(&_cond);
}

bool Condition::wait(int timeout)
{
    bool good = true;
    if (debug)
    {
        DAQpp::ostream os;
        os << loglevel(Log::debug) << "Condition::wait - acquiring Mutex" << std::endl;
    }
    acquire();
    if (debug)
    {
        DAQpp::ostream os;
        os << loglevel(Log::debug) << "Condition::wait - ...done " << std::endl;
    }
    if ( timeout <= 0 )
    {
        while ( !condition() )
        {
            Py_BEGIN_ALLOW_THREADS
            pthread_cond_wait(&_cond, &_mtx);
            Py_END_ALLOW_THREADS
        }
        exec();
    }
    else
    {
        struct timeval now;
        struct timespec ts;
        gettimeofday(&now, 0);

        now.tv_usec += (timeout % 1000000);
        now.tv_sec += (timeout + now.tv_usec) / 1000000;
        now.tv_usec = now.tv_usec % 1000000;

        ts.tv_sec = now.tv_sec;
        ts.tv_nsec = now.tv_usec * 1000;

        int rc = 0;
        while ( !condition() )
        {
            Py_BEGIN_ALLOW_THREADS
            rc = pthread_cond_timedwait(&_cond, &_mtx, &ts);
            Py_END_ALLOW_THREADS
            if (rc)
            {
                good = false;
                break;
            }
        }
        if (good)
            exec();
    }

    if (debug)
    {
        DAQpp::ostream os;
        os << loglevel(Log::debug) << "Condition::wait - releasing mutex " << std::endl;
    }
    release();

    if (debug)
    {
        DAQpp::ostream os;
        os << loglevel(Log::debug) << "Condition::wait - ...done " << std::endl;
    }
    return good;
}

void Condition::notify()
{
    pthread_cond_signal(&_cond);
}

void Condition::notify_all()
{
    pthread_cond_broadcast(&_cond);
}

Semaphore::Semaphore(int ival)
{
    sem_init(&_sem, 0, ival);
}

Semaphore::~Semaphore()
{
    do
    {
        sem_destroy(&_sem);
    }
    while ( errno == EBUSY );
}

int Semaphore::wait(bool block)
{
    int rc;
    if ( ! block )
    {
        rc = sem_trywait(&_sem);
        if ( errno == EAGAIN )
        {
            // The count is already 0
            rc = 1;
        }
    }
    else
    {
        Py_BEGIN_ALLOW_THREADS
        rc = sem_wait(&_sem);
        Py_END_ALLOW_THREADS
    }
    return rc;
}

void Semaphore::post()
{
    sem_post(&_sem);
}

int Semaphore::count()
{
    int rc;
    sem_getvalue(&_sem, &rc);
    return rc;
}

Thread::Thread() : _thread(0), _started(false), _stopped(false), _nice(-1)
{
}


void Thread::start()
{
    _started = true;
    _stopped = false;

    /* Initialize and set thread detached attribute */
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    pthread_create(&_thread, &attr, _run, this);

    pthread_attr_destroy(&attr);
}

void Thread::stop()
{
    void *rc;
    if (!_stopped && _thread)
    {
        _stopped = true;
        pthread_cancel(_thread);
        pthread_join(_thread, &rc);
    }
    _started = false;
    _stopped = true;
    _thread = 0;
}

void Thread::exit(int val)
{
    pthread_exit((void *)val);
}

void Thread::test_cancel()
{
    pthread_testcancel();
}

void Thread::set_cancel(CancelType ct)
{
    switch (ct)
    {
        case deferred:
            pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
            pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, 0);
            break;
        case asynchronous:
            pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
            pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);
            break;
        case disable:
            pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, 0);
            break;
    }
}

void *Thread::_run(void *data)
{
    Thread *thrd = reinterpret_cast<Thread *>(data);
    //    if (thrd->_nice>=0)
    //       setpriority(PRIO_PROCESS,getpid(),thrd->_nice);

    thrd->run();

    thrd->_stopped = true;
    thrd->_started = false;
    pthread_exit(0);
    return 0;
}

void DAQpp::wait(double secs)
{
    struct timeval t;
    double frac;
    frac = fmod(secs, 1.0);
    secs = floor(secs);
    t.tv_sec = (long)secs;
    t.tv_usec = (long)(frac*1000000.0);
    Py_BEGIN_ALLOW_THREADS
    if (select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &t) != 0) {
        if (errno!=EINTR)
        {
            DAQpp::ostream os;
            os << loglevel(Log::error) << "Error in DAQpp::wait: " << strerror(errno) << std::endl;
        }
    }
    Py_END_ALLOW_THREADS
}



DAQpp::AtomicInt::AtomicInt( int __i)
    :  _mi(__i)
{
    _mtx = new pthread_mutex_t;
    pthread_mutex_init( (pthread_mutex_t *) _mtx, 0);
}
DAQpp::AtomicInt::~AtomicInt()
{
    pthread_mutex_destroy((pthread_mutex_t *)_mtx);
}

void DAQpp::AtomicInt::store(const int &__i)
{
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    _mi = __i;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
}

int DAQpp::AtomicInt::load() const
{
    volatile int val;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    val = _mi;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return val;
}

int DAQpp::AtomicInt::fetch_add(int v)
{
    volatile int val;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    val = _mi;
    _mi += v;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return val;

}
int DAQpp::AtomicInt::fetch_sub(int v)
{
    volatile int val;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    val = _mi;
    _mi -= v;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return val;
}
int DAQpp::AtomicInt::operator|=(int x)
{
    volatile int out;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    _mi |= x;
    out = x;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return out;
}
int DAQpp::AtomicInt::operator&=(int x)
{
    volatile int out;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    _mi &= x;
    out = x;
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return out;

}
int DAQpp::AtomicInt::operator|(int x) const
{
    volatile int out;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    out = (_mi | x);
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return out;

}
int DAQpp::AtomicInt::operator&(int x) const
{
    volatile int out;
    Py_BEGIN_ALLOW_THREADS
    pthread_mutex_lock((pthread_mutex_t *)_mtx);
    out = (_mi & x);
    pthread_mutex_unlock((pthread_mutex_t *)_mtx);
    Py_END_ALLOW_THREADS
    return out;

}
