/* -*- mode: c++ -*- */
#ifndef __DAQpp_thread_h__
#define __DAQpp_thread_h__

#include <string>
#include <pthread.h>
#include <semaphore.h>

namespace DAQpp
{

    struct MutexPriv;
    /** \brief class Mutex
     *   This class encapsulates a pthread Mutex
     */
    class Mutex
    {
        protected:
            /// debugflags
            bool             debug;

            /// the mutex
            pthread_mutex_t _mtx;

            /// A private structure
            MutexPriv *_priv;
            /// the thread that owns the mutex
            ///pthread_t       _owner;

            /// A counter
            ///int             _count;

            Mutex(const Mutex &);
            Mutex &operator=(const Mutex &);
        public:
            /// constructor
            Mutex();
            /// destructor
            ~Mutex();

            /// Sets the debug flags to the value in the input
            void set_debug(bool x)
            {
                debug = x;
            }

            /// Acquire or lock the mutex
            int  acquire();
            /** Checks whether the mutex is locked or not.
             *   @returns true if the mutex is locked, false otherwise.
             */
            bool busy();
            /** Releases the Mutex.
             *   Note: This class emulates a recursive mutex. One needs as
             *   many calls to release as calls where made to acquire.
             *
             *   @returns the number of times this function still needs to be
             *   called to really release the Mutex
             */
            int  release();

    };

    /**
     * \brief call acquire_guard
     *
     * Acquire the Mutex with a constructor call, then release it with
     * the destructor call
     */
    class acquire_guard
    {
        private:
            Mutex &_mutex;
            acquire_guard(const acquire_guard &);
            acquire_guard &operator=(const acquire_guard &);
        public:
            acquire_guard(Mutex &m)
            : _mutex(m)
            {
                _mutex.acquire();
            }
            ~acquire_guard()
            {
                _mutex.release();
            }
    };

    /** \brief class Condition.
     *
     * This class encapsulates a condition, that is, it allows a thread
     * to suspend execution until some condition is satisfied. The way to
     * use it is calling its wait() method. When the condition is
     * satisfied the class will notify it to the waiting threads.
     *
     * This is an abstract class and the objects deriving from it should
     * implement the condition() method.
     *
     */
    class Condition : public Mutex
    {
        private:
            /// a pthread condition that we are Wrapping
            pthread_cond_t _cond;

            Condition(const Condition &);
            Condition &operator=(const Condition &);
        public:
            /// Constructor
            Condition();

            /// Destructor
            virtual ~Condition();


            /** Waits until the condition is satisfied (returns true) or
			 * 	there is a timeout (returns false). The time out is in microseconds
             */
            bool wait(int timeout = -1);

            /// Notifies a change to one of the waiting threads
            void notify();

            /// Notifies a change to all the waiting threads
            void notify_all();

            /** Condition to be checked.
             *   This function should be overloaded and return true when the
             *   thread should wake up and false otherwise
             */
            virtual bool condition() = 0;
            /**
             *  This is an optional function that is called before releasing
             *  the lock of the condition
             */
            virtual void exec()
            {}

    };

    /** \brief class Semaphore.
     */
    class Semaphore
    {
        private:
            /// The semaphore we are wrapping
            sem_t _sem;
        public:
            /// Constructor
            Semaphore(int ifirst);
            /// Destructor
            ~Semaphore();

            /** Waits untils the sempahore value is not 0. It will decrease
             *   the semaphore count on exit.
             *   \param block if false it is a non-blocking call,
             *                otherwise it will block.
             *   \return  if in non blocking mode it will return 1
             *            value if the semaphore has a 0 count, 0 if the
             *            value could be decreased and -1 in case of
             *            error. In blocking mode it will return 0 in case
             *            of success and -1 in case of problems
             */
            int wait(bool block = true);

            /// increases the semaphore count
            void post();

            /// returns the value of the semaphore count
            int count();
    };

    /** \brief class Thread.
     *
     * This class represents a thread. It is an abstract class and the
     * objects deriving from it should implement the run() method, which
     * defines what the thread will execute. The thread will start
     * running when the start() method is called and can be stopped by
     * calling the stop() method.
     */
    class Thread
    {
        private:
            /// The thread we are wrapping
            pthread_t _thread;

            /// says if the thread has already started
            bool           _started;

            /// says if the thread has stopped
            bool           _stopped;

            /// The nice value for the thread process
            int            _nice;
        public:
            enum CancelType {
                deferred,
                asynchronous,
                disable
            };

            /// constructor
            Thread();

            /// destructor
            virtual ~Thread()
            {};

            /// Sets the nice value of the thread process
            void nice(int x)
            {
                _nice = x;
            }

            /// test if the thread has been cancelled
            void test_cancel();

            /// sets the way the thread can be cancelled. It is any of
            /// the options in CancelType
            void set_cancel(CancelType);

            /// exits the thread
            void exit(int v = 0);

            /// Starts the thread
            virtual void start();

            /// Stops the thread
            virtual void stop();

            /// The method that gets executed in the thread.
            virtual void run() = 0;

            /// Internal
            static void *_run(void *);

            /// Checks if the thread is running
            bool is_running()
            {
                return _started;
            }

            /// Checks if the thread has stopped
            bool is_stopped()
            {
                return _stopped;
            }

    };
    /**
     * class AtomicInt.
     * This is a class that uses a Mutex to emulate atomic operations.
     * It is defined as a template.
     */
    class AtomicInt
    {
        private:
            int _mi;
            void *_mtx;

            AtomicInt() : _mi(0), _mtx(0) {}
            AtomicInt (const AtomicInt &) : _mi(0), _mtx(0) {}
        public:
            AtomicInt( int __i);
            ~AtomicInt();
            int operator ()() const {
                return load();
            }
            int operator=(const int &__i){
                store(__i);
                return __i;
            }
            void store(const int &__i);
            int load() const;
            int fetch_add(int v);
            int fetch_sub(int v);
            int operator++(int) { return fetch_add(1); }
            int operator--(int)  { return fetch_sub(1); }
            int operator++() { return fetch_add(1) + 1; }
            int operator--() { return fetch_sub(1) - 1; }
            int operator|=(int x);
            int operator&=(int x);
            int operator|(int x) const;
            int operator&(int x) const;

    };


    void wait(double sec);
}
#endif
