/*
 * MsgManager.h
 *
 *  Created on: 13 Jan 2011
 *      Author: lacasta
 */
#ifndef MSGMANAGER_H_
#define MSGMANAGER_H_

#include <DAQ++/DAQObject.h>
#include <DAQ++/Thread.h>
#include <DAQ++/ObserverModel.h>
#include <map>

namespace DAQpp {

    class MsgManager : public Observer
    {
        public:
            /**
             * A notification has the DAQid as key and the parameter or state name
             * as value. This way we only store one notification per object.
             */
            typedef std::map<std::string, std::string> Notification;

            /**
             * A list of 'detected' DAQObjects
             */
            typedef std::map<std::string, DAQpp::DAQObject *> Objects;
        private:
            Notification notif_parameters;
            Notification notif_state;
            Objects daq_objects;
            DAQpp::Mutex mtx;

        protected:
            // Hide the constructors
            MsgManager();
            MsgManager(const MsgManager &)
            {
            }

            /**
             * Whenever a RunManager is created, MsgManager registers as
             * an observer to any change of state and transmits that to the
             * corresponding RunMenu. RunManager notifications come through
             * this subclass
             */
            class StateChanged : public Observer
            {
                public:
                    StateChanged() {}
                    void update(Observable *o, Argument *arg);
            };

            /**
             * DAQpp::Parameter objects represented by ParamSlider objects
             *  will notify their changes through this subclass
             *
             */
            class ParameterChanged : public Observer
            {
                public:
                    ParameterChanged() {}
                    void update(Observable *o, Argument *arg);
            };

            class ObjectAddedRemoved : public Observer
            {
                public:
                    ObjectAddedRemoved() {}
                    void update(Observable *o, Argument *arg);
            };

            public:
            virtual ~MsgManager()
            {
            }

            /// Returns a pointer to the manager
            static MsgManager *theMsgManager()
            {
                static MsgManager mngr;
                return &mngr;
            }

            void acquire() { mtx.acquire(); }
            void release() { mtx.release(); }

            Notification::const_iterator begin_par() { return notif_parameters.begin(); }
            Notification::const_iterator end_par() { return notif_parameters.end(); }
            Notification::const_iterator begin_stat() { return notif_state.begin(); }
            Notification::const_iterator end_stat() { return notif_state.end(); }
            void clear_notifications() {
                notif_parameters.clear();
                notif_state.clear();
            }




            StateChanged state;
            ParameterChanged parameter;
            ObjectAddedRemoved object;

            /**
             * This is the routine called by DAQmanager when the MsgManager
             * registers as an Observer
             */
            void update(Observable *o, Argument *arg);
    };

}

#endif /* MSGMANAGER_H_ */
