/*
 * daqppClient.cc
 *
 *  Created on: Dec 15, 2014
 *      Author: lacasta
 */

#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include "daqppClient.h"
#include "utils.h"
#include "logging.h"

std::ostream &operator<<(std::ostream &os, const ns1__Status &st)
{
    double rate = st.rate;
    std::string rate_units = " Hz";
    std::string thr_units;

    if (rate > 1.e6)
    {
        rate /= 1.e6;
        rate_units = "MHz";
    }
    else if (rate > 1000.)
    {
        rate /= 1000.;
        rate_units = "kHz";
    }

    double throughput = st.throughput / st.time;
    if (throughput > 1048576.0)
    {
        throughput /= 1048576.0;
        thr_units = "Mb/s";
    }
    else if (throughput > 1024.0)
    {
        throughput /= 1024.0;
        thr_units = "kb/s";
    }
    else if (throughput > 1.0)
    {
        thr_units = "b/s";
    }
    else
    {
        throughput *= 1024;
        thr_units = " mb/s";
    }

    std::cout << "Evts " << std::setiosflags(std::ios::fixed | std::ios::right)
            << std::setw(7) << st.ntrigger << " time " << std::setw(6)
            << std::setprecision(1) << st.time << " rate " << std::setw(8)
            << std::setprecision(1) << rate << rate_units << " eff " << std::setw(5)
            << std::setprecision(1) << st.eff << " - " << std::setw(7)
            << std::setprecision(1) << throughput << ' ' << thr_units
            << std::flush;
    return os;
}


void transform_state(DAQpp::RunState* dst, ns1__Status* st)
{
    dst->set_name(st->name);
    switch (st->status)
    {
        case ns1__StatusType__NotDefined:
            dst->set_status(DAQpp::RunState::NotDefined);
            break;
        case ns1__StatusType__NotReady:
            dst->set_status(DAQpp::RunState::NotReady);
            break;
        case ns1__StatusType__Ready:
            dst->set_status(DAQpp::RunState::Ready);
            break;
        case ns1__StatusType__Starting:
            dst->set_status(DAQpp::RunState::Starting);
            break;
        case ns1__StatusType__Paused:
            dst->set_status(DAQpp::RunState::Paused);
            break;
        case ns1__StatusType__Running:
            dst->set_status(DAQpp::RunState::Running);
            break;
        case ns1__StatusType__Stopping:
            dst->set_status(DAQpp::RunState::Stopping);
            break;
        case ns1__StatusType__TimeOut:
            dst->set_status(DAQpp::RunState::TimeOut);
            break;
        default:
            dst->set_status(DAQpp::RunState::NotDefined);
            break;
    }
    dst->set_expected(st->nexpected);
    dst->set_throughput(st->throughput);
    dst->set_ntrigger(st->ntrigger);
    dst->set_nwritten(st->nwritten);
    dst->set_time(st->time);
}


template<typename Type, typename anyType>
DAQpp::Parameter *get_vector(ns1__Parameter *P)
{
    std::vector<Type> vout;
    get_vpar_value<Type, anyType>(P, vout);
    return new DAQpp::Par< std::vector<Type> >(vout);
}

DAQpp::Parameter *get_parameter( ns1__Parameter *P )
{
    DAQpp::Parameter *out = 0;
    std::vector<xsd__anyType *> &V = P->value->value->item;
    if (V.size()>1)
    {
        // vectors
        xsd__anyType *any = V[0];
        switch (any->soap_type())
        {
            case SOAP_TYPE_bool:
            case SOAP_TYPE_xsd__boolean:
                out = get_vector<bool, xsd__boolean>(P);
                break;
            case SOAP_TYPE_unsignedShort:
            case SOAP_TYPE_xsd__unsignedShort:
                out = get_vector<unsigned short, xsd__unsignedShort>(P);
                break;
            case SOAP_TYPE_unsignedInt:
            case SOAP_TYPE_xsd__unsignedInt:
                out = get_vector<unsigned int, xsd__unsignedInt>(P);
                break;
            case SOAP_TYPE_unsignedLONG64:
            case SOAP_TYPE_xsd__unsignedLong:
                out = get_vector<unsigned long, xsd__unsignedLong>(P);
                break;
            case SOAP_TYPE_short:
            case SOAP_TYPE_xsd__short:
                out = get_vector<short, xsd__short>(P);
                break;
            case SOAP_TYPE_int:
            case SOAP_TYPE_xsd__int:
                out = get_vector<int, xsd__int>(P);
                break;
            case SOAP_TYPE_LONG64:
            case SOAP_TYPE_xsd__long:
                out = get_vector<long, xsd__long>(P);
                break;
            case SOAP_TYPE_float:
            case SOAP_TYPE_xsd__float:
                out = get_vector<float, xsd__float>(P);
                break;
            case SOAP_TYPE_double:
            case SOAP_TYPE_xsd__double:
                out = get_vector<double, xsd__double>(P);
                break;
            default:
                break;
        }
    }
    else
    {
        xsd__anyType *any = V[0];
        switch (any->soap_type())
        {
            case SOAP_TYPE_bool:
            case SOAP_TYPE_xsd__boolean:
                out = new DAQpp::Par<bool>( get_par_value<bool, xsd__boolean>(P));
                break;
            case SOAP_TYPE_unsignedShort:
            case SOAP_TYPE_xsd__unsignedShort:
                out = new DAQpp::Par<unsigned short>(get_par_value<unsigned short, xsd__unsignedShort>(P));
                break;
            case SOAP_TYPE_unsignedInt:
            case SOAP_TYPE_xsd__unsignedInt:
                out = new DAQpp::Par<unsigned int>(get_par_value<unsigned int, xsd__unsignedInt>(P));
                break;
            case SOAP_TYPE_unsignedLONG64:
            case SOAP_TYPE_xsd__unsignedLong:
                out = new DAQpp::Par<unsigned long>(get_par_value<unsigned long, xsd__unsignedLong>(P));
                break;
            case SOAP_TYPE_short:
            case SOAP_TYPE_xsd__short:
                out = new DAQpp::Par<short>(get_par_value<short, xsd__short>(P));
                break;
            case SOAP_TYPE_int:
            case SOAP_TYPE_xsd__int:
                out = new DAQpp::Par<int>(get_par_value<int, xsd__int>(P));
                break;
            case SOAP_TYPE_LONG64:
            case SOAP_TYPE_xsd__long:
                out = new DAQpp::Par<long>(get_par_value<long, xsd__long>(P));
                break;
            case SOAP_TYPE_float:
            case SOAP_TYPE_xsd__float:
                out = new DAQpp::Par<float>(get_par_value<float, xsd__float>(P));
                break;
            case SOAP_TYPE_double:
            case SOAP_TYPE_xsd__double:
                out = new DAQpp::Par<double>(get_par_value<double, xsd__double>(P));
                break;
            default:
                break;
        }
    }
    if (out)
    {
        out->set_id(P->daqid);
        out->set_name( P->value->name );
    }
    return out;
}


daqppClient::daqppClient(const char *uri)
    : server(0)
{
    if (uri)
        connect(uri);
}

daqppClient::~daqppClient()
{
    if (server)
        delete server;
}


void daqppClient::connect(const char *uri)
{
    if (server)
        delete server;

    server = new daqppSOAPProxy(uri);
}


int daqppClient::stopServer()
{
    int rc = -1;
    if (server)
        server->stopServer(0, rc);

    return rc;
}

void daqppClient::set_logging( int opt)
{
    if (server && (opt & 3) )
    {
        soap_register_plugin(server, logging);
        if ( opt & debugOUT)
            soap_set_logging_outbound(server, stdout);
        if ( opt & debugIN )
            soap_set_logging_inbound(server, stdout);

    }
}

void daqppClient::wait(double s)
{
    daqpp_wait(s);
}

int daqppClient::getStatus(const std::string &daqid, DAQpp::RunState &status)
{
    if (!server)
        return 1;
    ns1__getStatusResponse response;
    server->getStatus(daqid, response);

    transform_state(&status, response.status);

//    ns1__Status *ss = response.status;
//    status.name = ss->name;
//    status.daqid = ss->daqid;
//    status.status = ss->status;
//    status.nexpected = ss->nexpected;
//    status.ntrigger = ss->ntrigger;
//    status.ntrigger = ss->ntrigger;
//    status.eff = ss->eff;
//    status.rate = ss->rate;
//    status.throughput = ss->throughput;
//    status.time = ss->time;
    return 0;
}

int daqppClient::getModuleList(daqppClient::NameList &lst)
{
    ns1__getModuleListResponse mlist;
    int rc = server->getModuleList(0, mlist);

    lst.insert(lst.end(), mlist.idlist->name.begin(), mlist.idlist->name.end());

    return rc;
}

int daqppClient::getRunManagerList(daqppClient::NameList &lst)
{
    ns1__getRunManagerListResponse mlist;
    int rc = server->getRunManagerList(0, mlist);

    lst.insert(lst.end(), mlist.idlist->name.begin(), mlist.idlist->name.end());

    return rc;
}

int daqppClient::getRunManagerModules(const std::string &daqid, NameList &lst)
{
    ns1__getRunManagerModulesResponse mlist;
    int rc = server->getRunManagerModules(daqid, mlist);

    lst.insert(lst.end(), mlist.idlist->name.begin(), mlist.idlist->name.end());

    return rc;
}

int daqppClient::getParametersList(const std::string &daqid, daqppClient::NameList &lst)
{
    ns1__getParameterNamesResponse mlist;
    int rc = server->getParameterNames(daqid, mlist);

    lst.insert(lst.end(), mlist.idlist->name.begin(), mlist.idlist->name.end());

    return rc;
}

int daqppClient::StartRunManager(const std::string &daqid)
{
    int rc=-1;
    if (server)
        server->StartRunManager(daqid, rc);

    return rc;
}
int daqppClient::StopRunManager(const std::string &daqid)
{
    int rc=-1;
    if (server)
        server->StopRunManager(daqid, rc);

    return rc;

}

int daqppClient::GetReady(const std::string &daqid)
{
    int rc = -1;
    if (server)
        server->GetReady( daqid, rc);

    return rc;
}
int daqppClient::Pause(const std::string &daqid)
{
    int rc = -1;
    if (server)
        server->Pause(daqid, rc);

    return rc;
}
int daqppClient::Continue(const std::string &daqid)
{
    int rc = -1;
    if (server)
        server->Continue(daqid, rc);

    return rc;
}
int daqppClient::Reset(const std::string &daqid)
{
    int rc = -1;
    if (server)
        server->Reset(daqid, rc);
    return rc;
}

int daqppClient::setMaxEvents(const std::string &daqid, int nevts)
{
    int rc = -1;
    if (server)
    {
        ns1__MaxEvents mxevts;
        mxevts.daqid=daqid;
        mxevts.nevts= nevts;
        server->setMaxEvents(&mxevts, rc);
    }
    return rc;
}

int daqppClient::getMonitorData(const std::string &daqid)
{
    if (!server)
        return -1;

    xsd__base64Binary mdata;
    server->getMonitorData(daqid, mdata);
    std::istringstream is( std::string((const char *)mdata.__ptr, (size_t)mdata.__size ) );
    monitor_data(is);

    return 0;
}

int daqppClient::resetMonitorData(const std::string &daqid)
{
    int rc = -1;
    if (server)
        server->resetMonitorData(daqid, rc);
    return rc;
}


int daqppClient::sendUpdate( ParameterList &parList, StateList &stateList )
{
    int rc = 0;

    ns1__sendUpdateResponse update;
    rc = server->sendUpdate(0, update);

    // Start with the States
    if ( ! update.update->status->status.empty() )
    {
        std::vector<ns1__Status *>::iterator is;
        for (is=update.update->status->status.begin(); is!=update.update->status->status.end(); is++)
        {
            ns1__Status *st = *is;
            DAQpp::RunState *dst = new DAQpp::RunState( st->daqid );
            transform_state(dst, st);
            stateList.push_back(dst);
        }
    }

    if (!update.update->parameters->parameter.empty())
    {
        std::vector<ns1__Parameter *>::iterator ip;
        for (ip=update.update->parameters->parameter.begin(); ip!=update.update->parameters->parameter.end(); ip++)
        {
            parList.push_back( get_parameter(*ip) );
        }
    }

    return rc;
}

