#include "pydaqpp.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <DAQ++/Fifo.h>

using namespace DAQpp;

typedef struct {
      PyObject_HEAD
      /* Type-specific fields go here. */
      PyObject *dict;
      DAQpp::Fifo *fifo;
} daqpp_Fifo;

PyDoc_STRVAR(fifotype_doc,
"This class implements a fifo. It can be used as a data communication\n\
channel between different threads, for instance the\n\
RunManager::writeData() and a data logger.\n\
\n\
");

static PyObject *
fifo_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    daqpp_Fifo *self;
    self = PyObject_New(daqpp_Fifo ,type);
    if (self!=NULL) {
        self->dict = PyDict_New();
        self->fifo = 0;
    }
    return (PyObject *)self;
}

static void
fifo_dealloc(daqpp_Fifo* self)
{
    delete self->fifo;
    self->fifo = 0;
    Py_DECREF(self->dict);
    self->ob_type->tp_free((PyObject*)self);
}

static int
fifo_init(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    int size;

    if (!PyArg_ParseTuple(args, "i:daqpp.Fifo.__init__", &size))
        return -1;

    self->fifo = new  Fifo(size);


    return 0;
}

PyDoc_STRVAR(fifo_size_doc,"returns the size of the fifo.\n\
");
static PyObject *
fifo_size(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    return PyInt_FromLong( (long)self->fifo->size() );
}

PyDoc_STRVAR(fifo_empty_doc,"tells if the fifo is empty.\n\
");
static PyObject *
fifo_empty(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    return PyInt_FromLong( (long)self->fifo->empty() );
}

PyDoc_STRVAR(fifo_reset_doc,
"reset()\n\
resets the Fifo.\n\
");
static PyObject *
fifo_reset(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    self->fifo->reset();
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(fifo_read_doc,
"read(count)\n\
reads count bytes from the fifo. It returns a daqpp.Buffer whose size\n\
is the actual number of bytes read.\n\
@param count: number of bytes to read\n\
@returns: a DAQpp::Buffer with the data read\n\
");
static PyObject *fifo_read(daqpp_Fifo *self, PyObject *args, PyObject *kwargs)
{
    int size, rc=0;
    PyObject *ret= NULL;

    if (!PyArg_ParseTuple(args, "i:daqpp.Fifo.read", &size))
        return NULL;

    ret = PyDAQpp_Buffer_New('c', size, NULL);
    Py_ssize_t sz;
    char *ptr;
    if (PyObject_AsWriteBuffer(ret, (void **)&ptr, &sz) )
    {
        PyErr_SetString(PyExc_MemoryError, "daqpp.Fifo.read: problems getting the buffer");

        Py_DECREF(ret);
        ret = NULL;
        goto out;
    }
    rc = 0;
    while (rc<size)
    {
        Py_BEGIN_ALLOW_THREADS
        rc += self->fifo->read(ptr+rc,size-rc);
        Py_END_ALLOW_THREADS
    }

 out: 
    return ret;
}

PyDoc_STRVAR(fifo_write_doc,
"write(buffer,force)\n\
writes to the fifo and returns the number of bytes actually written.\n\
If force is false, the data will be lost when there is not enough\n\
space in the buffer. If force is true, the Fifo buffer will be\n\
enlarged to ensure that this block of data is not lost.\n\
");
static PyObject *fifo_write(daqpp_Fifo *self, PyObject *args, PyObject *kwargs)
{
    PyObject *ret=NULL;
    PyObject *buffer;
    int force=0;
    if (!PyArg_ParseTuple(args, "O|i:daqpp.Fifo.write", &buffer, &force))
        return NULL;

    int rc;
    Py_ssize_t sz;
    const char *ptr;
    if (PyObject_AsReadBuffer(buffer, (const void **)&ptr, &sz) )
    {
        PyErr_SetString(DAQpp_Error, "daqpp.Fifo.write: first argument shoud be a buffer like object");
        Py_DECREF(buffer);
        goto out;
    }
    Py_BEGIN_ALLOW_THREADS
    rc = self->fifo->write(const_cast<char *>(ptr), sz, force);
    Py_END_ALLOW_THREADS
    ret = PyInt_FromLong((long)rc);
    out: return ret;
}


PyDoc_STRVAR(fifo_left_doc,
"returns the space left in the fifo buffer\n\
");
static PyObject *
fifo_left(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    return PyInt_FromLong( (long)self->fifo->space_left() );
}

PyDoc_STRVAR(fifo_available_doc,
"returns the ammount of data not yet read from the fifo\n\
");
static PyObject *
fifo_available(daqpp_Fifo *self, PyObject *args, PyObject *kwargs) {
    return PyInt_FromLong( (long)self->fifo->data_available() );
}


static PyMethodDef fifo_methods[] = {
   {"size",(PyCFunction)fifo_size, METH_NOARGS,fifo_size_doc},
   {"empty",(PyCFunction)fifo_empty, METH_NOARGS,fifo_empty_doc},
   {"reset",(PyCFunction)fifo_reset, METH_NOARGS,fifo_reset_doc},
   {"read",(PyCFunction)fifo_read, METH_VARARGS, fifo_read_doc},
   {"write",(PyCFunction)fifo_write, METH_VARARGS, fifo_write_doc},
   {"space_left",(PyCFunction)fifo_left,METH_NOARGS,fifo_left_doc},
   {"data_available",(PyCFunction)fifo_available,METH_NOARGS,fifo_available_doc},
   {NULL}  /* Sentinel */
};
static PyMemberDef fifo_members[] = {
    {NULL}  /* Sentinel */
};

static PyTypeObject Fifo_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.Fifo",             /*tp_name*/
   sizeof(daqpp_Fifo),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)fifo_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   0,                         /*tp_getattro*/
   0,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE ,        /*tp_flags*/
   fifotype_doc,           /* tp_doc */
       0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    fifo_methods,             /* tp_methods */
    fifo_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    offsetof(daqpp_Fifo, dict),                         /* tp_dictoffset */
    (initproc)fifo_init,      /* tp_init */
    0,                         /* tp_alloc */
    fifo_new,                 /* tp_new */
};

void register_fifo(PyObject *dict) {
   daqpp_register_class(dict,"Fifo",
			&Fifo_Type,
			0);
}
