#include <Python.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <ctime>
#include "pydaqpp.h"
#include "pydaqpp_debug.h"

#include <DAQ++/FileIOManager.h>

using namespace DAQpp;

typedef struct {
      PyObject_HEAD
      /* Type-specific fields go here. */
      PyObject *header_data;
      PyObject *tail_data;
      FileIOManager *fm;
} daqpp_fileIOManager;

static PyObject *
fileIOManager_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
   daqpp_fileIOManager *self;
   WHOAMI;

   self = PyObject_New(daqpp_fileIOManager,type);
   if (self!=NULL) {
      self->fm  = 0;
      self->header_data = 0;
      self->tail_data = 0;
   }
   return (PyObject *)self;
}


static void
fileIOManager_dealloc(daqpp_fileIOManager* self)
{
   WHOAMI;
   delete self->fm;
   self->fm = 0;
   self->ob_type->tp_free((PyObject*)self);
}

static int
fileIOManager_init(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   WHOAMI;
   int mxsize=0;
   int compression=0;
   if (!PyArg_ParseTuple(args,"ii:FileIOManager.__init__",&compression,&mxsize))
      return -1;

   self->fm = new FileIOManager(compression!=0,mxsize);
   return 0;
}

static PyObject*
fileIOManager_open( daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   PyObject *fnam;
   if (!PyArg_ParseTuple(args,"S:FileIOManager.open",&fnam))
      return NULL;

   char *ptr = PyString_AsString( fnam );
   if (!ptr) {
      PyErr_SetString(DAQpp_Error,"FileIOManager::open Cannot get filename");
      return NULL;
   }
   self->fm->open( ptr );
   Py_INCREF(Py_None);
   return Py_None;   
}


static PyObject*
fileIOManager_close(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   int rc = -1;
   if (self->fm) 
      rc = self->fm->close();

   return PyInt_FromLong(rc);
}

static PyObject*
fileIOManager_write(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   PyObject *S=Py_None;
   int rc, sz=0;

   if (!PyArg_ParseTuple(args,"Si:FileIOManager.write",&S,&sz))
      return NULL;

   rc = self->fm->write( (const char *)PyString_AsString(S), sz);
   return PyInt_FromLong(rc);
}

static int header_tail_function(char **out, char *data) {
   PyObject *tuple = (PyObject *)data;
   PyObject *func, *dt=Py_None,*ret, *repr;
   Py_ssize_t size;

   *out = 0;
   if (!PyTuple_Check(tuple)) {
      PyErr_SetString(DAQpp_Error,"FileIOManager::calling header/tail func. This should not happen");
      return 0;
   }
   func = PyTuple_GetItem(tuple,0);
   if (PyTuple_Size(tuple)>1)
      dt = PyTuple_GetItem(tuple,1);

   ret  = PyObject_CallFunctionObjArgs(func,dt,NULL);
   repr = PyObject_Repr(ret);
   size = PyString_Size(repr);
   *out = new char[ size+1 ];
   strcpy(*out, PyString_AsString(repr));
   Py_DECREF(repr);
   Py_DECREF(ret);
   return size;
}

static PyObject*
process_header_tail_args(PyObject *func, PyObject *data)
{
   PyObject *tuple = data ? PyTuple_New(2) : PyTuple_New(1);
   Py_INCREF(func);
   PyTuple_SetItem(tuple, 0, func);
   if (data) {
       Py_INCREF(data);
       PyTuple_SetItem(tuple, 1, data);
   }

   return tuple;
}

static void
free_header_tail_data(PyObject *data) {
   Py_ssize_t sz = PyTuple_Size(data);
   Py_DECREF( PyTuple_GetItem(data,0) );
   if (sz>1)
   {
      Py_DECREF( PyTuple_GetItem(data,1) );
   }
   Py_DECREF( data );
}

static PyObject*
fileIOManager_set_header(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   PyObject *func=0, *data=0, *tuple=0;
   static char *kwlist[] = { (char *)"func", (char *)"data", NULL };
   if (!PyArg_ParseTupleAndKeywords(args, kwargs,"OO:FileIOManager.set_header",kwlist,&func,&data))
      return NULL;

   if (!func || !PyCallable_Check(func)) {
      PyErr_SetString(DAQpp_Error,"FileIOManager::set_header_func input is not a function");
      return NULL;
   }
   if (self->header_data) {
      free_header_tail_data(self->header_data);
      self->header_data=0;
   }
   tuple = self->header_data = process_header_tail_args(func, data);
   self->fm->set_header_func((DAQpp::FileIOManager::HeaderFunc)header_tail_function, (char *)tuple);
   
   Py_INCREF(Py_None);
   return Py_None;
}
   
static PyObject*
fileIOManager_reset_header(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   free_header_tail_data(self->header_data);
   self->fm->reset_header_func();
   self->header_data=0;
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject*
fileIOManager_set_tail(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   PyObject *func=0, *data=0, *tuple=0;
   static char *kwlist[] = { "func", "data", NULL };
   if (!PyArg_ParseTupleAndKeywords(args, kwargs,"O|O:FileIOManager.set_header",kwlist,&func,&data))
      return NULL;

   if (!func || !PyCallable_Check(func)) {
      PyErr_SetString(DAQpp_Error,"FileIOManager::set_tail_func input is not a function");
      return NULL;
   }
   if (self->tail_data) {
      free_header_tail_data(self->tail_data);
      self->tail_data=0;
   }
   tuple = self->tail_data = process_header_tail_args(func, data);
   self->fm->set_tail_func((DAQpp::FileIOManager::HeaderFunc)header_tail_function, (char *)tuple);

   Py_INCREF(Py_None);
   return Py_None;
}
   
static PyObject*
fileIOManager_reset_tail(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   free_header_tail_data(self->tail_data);
   self->fm->reset_tail_func();
   self->tail_data=0;
   Py_INCREF(Py_None);
   return Py_None;
}


static PyObject*
fileIOManager_file_number(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   int rc=self->fm->file_number();
   return PyInt_FromLong(rc);
}

static PyObject*
fileIOManager_compress(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   int rc=self->fm->Compress();
   return PyInt_FromLong(rc);
}


static PyObject*
fileIOManager_size(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   int rc=self->fm->size();
   return PyInt_FromLong(rc);
}

static PyObject*
fileIOManager_relative_size(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   double rc=self->fm->relative_size();
   return PyFloat_FromDouble(rc);
}

static PyObject*
fileIOManager_file_name(daqpp_fileIOManager *self, PyObject *args, PyObject *kwargs) {
   std::string rc=self->fm->get_file_name();
   return PyString_FromString( rc.c_str() );
}
      
static PyGetSetDef fileIOManager_getseters[] = {
   {NULL}  /* Sentinel */
};

static PyMethodDef fileIOManager_methods[] = {
   {"open",(PyCFunction)fileIOManager_open,METH_VARARGS,"Opens a file series"},
   {"close",(PyCFunction)fileIOManager_close,METH_NOARGS,"Closess a file series"},
   {"write", (PyCFunction)fileIOManager_write,METH_VARARGS,"Writes into a file"},
   {"file_number",(PyCFunction)fileIOManager_file_number,METH_NOARGS,"Returns the file number"},
   {"size",(PyCFunction)fileIOManager_size,METH_NOARGS,"Returns the maximum size of a file"},
   {"relative_size",(PyCFunction)fileIOManager_relative_size,METH_NOARGS,"Returns the how much of a file has been writen"},
   {"file_name",(PyCFunction)fileIOManager_file_name,METH_NOARGS,"Returns the current name of a file"},
   {"compress",(PyCFunction)fileIOManager_compress,METH_NOARGS,"Returns the compress option"},
   {"set_header_func",(PyCFunction)fileIOManager_set_header,METH_VARARGS|METH_KEYWORDS,"Sets the header function."},
   {"reset_header_func",(PyCFunction)fileIOManager_reset_header,METH_NOARGS,"Resets the header function"},
   {"set_tail_func",(PyCFunction)fileIOManager_set_tail,METH_VARARGS|METH_KEYWORDS,"Sets the tail function."},
   {"reset_tail_func",(PyCFunction)fileIOManager_reset_tail,METH_NOARGS,"Resets the tail function"},
   {NULL}  /* Sentinel */
};


static PyMemberDef fileIOManager_members[] = {
   {NULL}  /* Sentinel */
};

static PyTypeObject fileIOManager_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.FileIOManager",             /*tp_name*/
   sizeof(daqpp_fileIOManager),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)fileIOManager_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   0,                         /*tp_getattro*/
   0,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   "FileIOManager wrapper",           /* tp_doc */
   0,		               /* tp_traverse */
   0,		               /* tp_clear */
   0,		               /* tp_richcompare */
   0,		               /* tp_weaklistoffset */
   0,		               /* tp_iter */
   0,		               /* tp_iternext */
   fileIOManager_methods,             /* tp_methods */
   fileIOManager_members,             /* tp_members */
   fileIOManager_getseters,         /* tp_getset */
   0,                         /* tp_base */
   0,                         /* tp_dict */
   0,                         /* tp_descr_get */
   0,                         /* tp_descr_set */
   0,                         /* tp_dictoffset */
   (initproc)fileIOManager_init,      /* tp_init */
   0,    /* tp_alloc */
   fileIOManager_new,                 /* tp_new */
};

void register_fileIOManager(PyObject *dict)
{
    daqpp_register_class(dict, "FileIOManager", &fileIOManager_Type, 0);
}
