//#define DAQpp_DEBUG 1
#include "pydaqpp.h"
#include "pydaqpp_debug.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <ctime>
#include <DAQ++/Monitor.h>
#include <DAQ++/RunManager.h>
#include "pydaqpp_types.h"
#include "pydaqpp_buffer.h"
#include "DAQobserver.h"

using namespace DAQpp;

class pyMonitor : public Monitor, public PyDAQpp
{
    private:
        PyObject *rm;

    public:
        pyMonitor() :
            rm(0), timeout(-1)
            {
            WHOAMI;
            }
        ~pyMonitor()
        {
            WHOAMI;
            Py_XDECREF(rm);
        }
        int timeout;

        bool user_run(DAQpp::Event *);
        int check_attributes();
        void set_runmanager(PyObject *r)
        {
            WHOAMI;
            Py_XDECREF(rm);
            Py_INCREF(r);
            rm=r;
        }
        PyObject *get_runmanager()
        {
            //WHOAMI;
            return rm ? rm : Py_None;
        }
        void on_start();
        void on_stop();
};

void pyMonitor::on_start()
{
    call_void_method("on_start",NULL);
}

void pyMonitor::on_stop()
{
    call_void_method("on_stop", NULL);
}

bool pyMonitor::user_run(DAQpp::Event *evt)
{
    WHOAMI;
#if defined(DAQpp_DEBUG)
    std::cout << "####### Monitor run" << std::endl;
#endif
    if (!rm)
    {
        std::cout << "pyMonitor: no RunManager associated"<< std::endl;
        return false;
    }
    //unsigned int gs = DAQobserver::theObserver()->start();
    PyGILState_STATE gs = PyGILState_Ensure();
    if ( evt && evt!=DAQpp::Monitor::MonitorTimeOut )
    {
        PyObject *buffer = PyDAQpp_Buffer_New('c',evt->size(), const_cast<char *>(evt->data()));
        PyObject *ret = PyObject_CallMethod( self(), "run","O",buffer);
        Py_DECREF(buffer);
        if (ret)
        {
            Py_DECREF(ret);
        }
    }
    else
    {
        PyObject *ret = PyObject_CallMethod( self(), "run","O",Py_None);
        if (ret)
        {
            Py_DECREF(ret);
        }
    }
    //DAQobserver::theObserver()->stop(gs);
    PyGILState_Release(gs);
    return true;
}

int pyMonitor::check_attributes()
{
    WHOAMI;
    if (!has_object())
        return 0;

    if (!PyObject_HasAttrString(self(), "run") )
    {
        PyErr_SetString(DAQpp_Error, "Monitor should define its own run method");
        return -1;
    }
    return 0;
}

static PyObject *monitor_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    daqpp_Monitor *self;
    WHOAMI;
    self = (daqpp_Monitor *)type->tp_alloc(type, 0);
    if (!self)
        return (PyObject *)0;

    self->m = 0;
    self->dict = PyDict_New();
    self->timeout=-1;
    return (PyObject *)self;
}

static void monitor_dealloc(daqpp_Monitor* self)
{
    WHOAMI;
    delete self->m;
    self->m=0;
    Py_DECREF(self->dict);
    self->ob_type->tp_free((PyObject*)self);
}

static int set_object_pointer(daqpp_Monitor *self, Monitor *m)
{
    int rc=0;
    WHOAMI;
    self->m = m;

    pyMonitor *pm = dynamic_cast<pyMonitor *>(m);
    if (pm)
    {
        pm->set_object((PyObject *)self);
        if (pm->check_attributes()<0)
            rc=-1;
    }
    return rc;
}

static int monitor_init(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;

    pyMonitor *m = new pyMonitor();
    m->timeout = self->timeout;
    if (set_object_pointer(self, m))
        return -1;

    return 0;
}

static PyObject *monitor_push_event(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    PyObject *obj;

    if (!PyArg_ParseTuple(args, "O:Monitor.push_event", &obj))
        return NULL;

    Py_ssize_t sz;
    char *cbuf=0;
    if (PyObject_AsReadBuffer(obj, (const void **)&cbuf, &sz) )
    {
        PyErr_SetString(DAQpp_Error, "Monitor.push_event: needs a buffer "
            "type variable as input");
        return NULL;
    }
    else
    {
        self->m->push_event(new Event(sz,cbuf));
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *monitor_get_event(daqpp_Monitor *self)
{
    DAQpp::Event *evt=self->m->get_event(self->timeout);
    if (evt==(DAQpp::Event *)0xcafecafe )
    {
        Py_INCREF(Py_None);
        return Py_None;
    }
    if (evt==0)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    //    PyObject *buffer=Py_None;
    //    Py_INCREF(Py_None);
    PyObject *buffer = PyDAQpp_Buffer_New('c', evt->size(), 0);
    memcpy(((daqpp_buffer *)buffer)->item, evt->data(), evt->size());
    delete evt;
    return buffer;
}

static PyObject *monitor_start_monitor(daqpp_Monitor *self)
{
    WHOAMI;
    self->m->start_monitor();
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *monitor_stop_monitor(daqpp_Monitor *self)
{
#if defined(DAQpp_DEBUG)
    std::cout << "%% Stopping Monitor" << std::endl;
#endif
    self->m->stop_monitor();
#if defined(DAQpp_DEBUG)
    std::cout << "%% ...done" << std::endl;
#endif
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *monitor_get_max_size(daqpp_Monitor *self)
{
    WHOAMI;
    return PyInt_FromLong(self->m->get_max_size() );
}

PyDoc_STRVAR(monitor_full_doc,"\
Tells if the monitor Event queue is full.\n\
");
static PyObject *monitor_full(daqpp_Monitor *self)
{
    WHOAMI;
    return PyInt_FromLong(self->m->full() );
}

static PyObject *monitor_set_max_size(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    int nevts;
    WHOAMI;

    if (!PyArg_ParseTuple(args, "i:Monitor.set_max_size", &nevts))
        return NULL;

    self->m->set_max_size(nevts);

    Py_INCREF(Py_None);
    return Py_None;
}
;

static PyObject *monitor_set_runmanager(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    PyObject *rm;
    WHOAMI;

    if (!PyArg_ParseTuple(args, "O:Monitor.set_runmanager", &rm))
        return NULL;

    pyMonitor *pm = dynamic_cast<pyMonitor *>(self->m);
    if (pm)
        pm->set_runmanager(rm);

    Py_INCREF(Py_None);
    return Py_None;
}
;

static PyObject *monitor_get_runmanager(daqpp_Monitor *self)
{
    PyObject *rm= Py_None;

    //WHOAMI;
    pyMonitor *pm = dynamic_cast<pyMonitor *>(self->m);
    if (pm && pm->get_runmanager() )
        rm = pm->get_runmanager();

    Py_INCREF(rm);
    return rm;
}

static int monitor_set_tmout(daqpp_Monitor *self, PyObject *value, void *closure)
{
    if (value==NULL)
    {
        PyErr_SetString(DAQpp_Error, "Monitor: Cannot set timeout value");
        return -1;
    }
    if (!PyNumber_Check(value))
    {
        PyErr_SetString(DAQpp_Error, "Monitor.timeout: value must be a number");
        return -1;
    }
    self->timeout = PyInt_AsLong(value);
    pyMonitor *m= PyMONITOR(self);
    if (m)
        m->timeout = self->timeout;

    return 0;
}

static PyObject *monitor_get_tmout(daqpp_Monitor *self, void *closure)
{
    return PyInt_FromLong(self->timeout);

}

PyDoc_STRVAR(monitor_on_start_doc,"Called when the monitor is started\n\
");
static PyObject *monitor_on_start(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    
    pyMonitor *pm = PyMONITOR(self);
    if (pm)
    {
        if (!pm->external_call()) {
            pm->set_python_call();
            pm->on_start();
            pm->clear_python_call();
        }
    }
    
    Py_INCREF(Py_None);
    return Py_None;    
}


PyDoc_STRVAR(monitor_on_stop_doc,"Called when the monitor is stopped\n\
");
static PyObject *monitor_on_stop(daqpp_Monitor *self, PyObject *args, PyObject *kwargs)
{
    pyMonitor *pm = PyMONITOR(self);
    if (pm)
    {
        if (!pm->external_call()) {
            pm->set_python_call();
            pm->on_stop();
            pm->clear_python_call();
        }
    }

    Py_INCREF(Py_None);
    return Py_None;    
}

PyDoc_STRVAR(monitor_runs_doc,"\
Tells if the monitor thread is still running.\n\
");
static PyObject *monitor_is_running(daqpp_Monitor *self)
{
    return PyInt_FromLong( MONITOR(self)->is_running());
}

PyDoc_STRVAR(monitor_stopped_doc,"\
Tells if the monitor thread has been stopped.\n\
");
static PyObject *monitor_is_stopped(daqpp_Monitor *self)
{
    return PyInt_FromLong( MONITOR(self)->is_stopped());
}

      
static PyGetSetDef monitor_getseters[] = {
   {"timeout",(getter)monitor_get_tmout,(setter)monitor_set_tmout,"Monitor time out",NULL},
   {NULL}  /* Sentinel */
};

static PyMethodDef monitor_methods[] = {
   {"push_event",(PyCFunction)monitor_push_event,METH_VARARGS,"."},
   {"get_event",(PyCFunction)monitor_get_event,METH_NOARGS,"."},
   {"start_monitor",(PyCFunction)monitor_start_monitor,METH_NOARGS,"."},
   {"stop_monitor",(PyCFunction)monitor_stop_monitor,METH_NOARGS,"."},
   {"get_max_size",(PyCFunction)monitor_get_max_size,METH_NOARGS,"."},
   {"set_max_size",(PyCFunction)monitor_set_max_size,METH_VARARGS,"."},
   {"set_runmanager",(PyCFunction)monitor_set_runmanager,METH_VARARGS,"."},
   {"get_runmanager",(PyCFunction)monitor_get_runmanager,METH_NOARGS,"."},
   {"is_running",(PyCFunction)monitor_is_running,METH_NOARGS,monitor_runs_doc},
   {"is_stopped",(PyCFunction)monitor_is_stopped,METH_NOARGS,monitor_stopped_doc},
   {"full",(PyCFunction)monitor_full,METH_NOARGS,monitor_full_doc},
   {"on_start",(PyCFunction)monitor_on_start,METH_NOARGS,monitor_on_start_doc},
   {"on_stop",(PyCFunction)monitor_on_stop,METH_NOARGS,monitor_on_stop_doc},
   {NULL}  /* Sentinel */
};
static PyMemberDef monitor_members[] = {
   {NULL}  /* Sentinel */
};

static PyTypeObject Monitor_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.Monitor",             /*tp_name*/
   sizeof(daqpp_Monitor),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)monitor_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   PyObject_GenericGetAttr,                         /*tp_getattro*/
   PyObject_GenericSetAttr,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   "Monitor wrapper",           /* tp_doc */
   0,		               /* tp_traverse */
   0,		               /* tp_clear */
   0,		               /* tp_richcompare */
   0,		               /* tp_weaklistoffset */
   0,		               /* tp_iter */
   0,		               /* tp_iternext */
   monitor_methods,             /* tp_methods */
   monitor_members,             /* tp_members */
   monitor_getseters,         /* tp_getset */
   0,                         /* tp_base */
   0,                         /* tp_dict */
   0,                         /* tp_descr_get */
   0,                         /* tp_descr_set */
   offsetof(daqpp_Monitor, dict),                         /* tp_dictoffset */
   (initproc)monitor_init,      /* tp_init */
   0,    /* tp_alloc */
   monitor_new,                 /* tp_new */
};


void register_monitor(PyObject *dict)
{
    daqpp_register_class(dict, "Monitor", &Monitor_Type, 0);
}

PyObject *PyDAQpp_Monitor_New(void *ptr)
{
    daqpp_Monitor *ret;

    ret = PyObject_New( daqpp_Monitor, &Monitor_Type);
    if (ret == NULL)
        return PyErr_NoMemory();

    Monitor *m = dynamic_cast<Monitor *>((Monitor *)ptr);
    if (!m)
    {
        PyErr_SetString(DAQpp_Error, "Parameter 1 of Monitor_New is not a Monitor");
        return Py_None;
    }
    set_object_pointer((daqpp_Monitor *)ret, m);
    ret->dict = PyDict_New();
    return (PyObject *)ret;

}
