#include "pydaqpp.h"
#include <DAQ++/ObserverModel.h>
#include <DAQ++/exception.h>
#include "pydaqpp_types.h"
#include "pydaqpp_debug.h"
#include "Observer.h"

#if PY_VERSION_HEX < 0x02030000
static PyObject *
PyObject_SelfIter(PyObject *obj)
{
   Py_INCREF(obj);
   return obj;
}
#endif

static PyObject *observable_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    daqpp_Observable *self;

    WHOAMI;
    self = PyObject_New(daqpp_Observable,type);
    if (!self)
        return (PyObject *)0;

    self->obs = 0;
    self->dict = PyDict_New();
    return (PyObject *)self;
}

static void observable_dealloc(daqpp_Observable* self)
{
    WHOAMI;
    Py_DECREF(self->dict);
    self->ob_type->tp_free((PyObject*)self);
}


static int observable_init(daqpp_Observable *self, PyObject *args, PyObject *kwargs)
{
    int rc=0;
    PyObject *in=NULL;

    if (!PyArg_ParseTuple(args, "|O:Observable.__init__", &in))
        return -1;

    self->obs = 0;
    return rc;
}



PyDoc_STRVAR(count_observers_doc,"Returns the number of observers\n\
");
static PyObject *observable_countObservers(daqpp_Observable *self)
{
    int rc = 0;
    if (self->obs)
    {
        rc = self->obs->countObservers();
    }
    return PyInt_FromLong(rc);
}



PyDoc_STRVAR(add_observer_doc,"Adds a new observer\n\
");

static PyObject *observable_addObserver(daqpp_Observable *self, PyObject *args, PyObject *kwargs)
{
    PyObject *in;
    Observer *obs;
    if (!PyArg_ParseTuple(args, "O:Observable.addObserver", &in))
        return NULL;

    if (in==NULL)
    {
        PyErr_SetString(DAQpp_Error, "Something scary happened!!!");
        return NULL;
    }

    if (!PyObserver_Check(in) )
    {
        if (!PyObject_HasAttrString(in, "update") )
        {
            PyErr_SetString(DAQpp_Error, "Parameter addObserver is not an Observer");
            Py_INCREF(Py_None);
            return Py_None;
        }
        else
        {
            obs = new pyObserver(in);
        }
    }
    else
    {
        obs = OBSERVER(in);
    }

    self->obs->addObserver(*obs );

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(delete_observer_doc,"Adds a new observer\n\
");

static PyObject *observable_deleteObserver(daqpp_Observable *self, PyObject *args, PyObject *kwargs)
{
    PyObject *in;
    if (!PyArg_ParseTuple(args, "O:Observable.deleteObserver", &in))
        return NULL;

    if (!PyObserver_Check(in))
    {
        PyErr_SetString(DAQpp_Error, "Parameter 1 of Observable.addObserver is not an Observable");
        Py_INCREF(Py_None);
        return Py_None;
    }

    self->obs->deleteObserver(*((daqpp_Observer *)in)->obs);

    Py_INCREF(Py_None);
    return Py_None;
}


static PyMethodDef observable_methods[] = {
   {"countObservers",(PyCFunction)observable_countObservers,METH_NOARGS,count_observers_doc},
   {"addObserver",(PyCFunction)observable_addObserver,METH_VARARGS,add_observer_doc},
   {"deleteObserver",(PyCFunction)observable_deleteObserver,METH_VARARGS,delete_observer_doc},
   {NULL}  /* Sentinel */
};


static PyMemberDef observable_members[] = {
    {NULL}  /* Sentinel */
};

PyDoc_STRVAR(observabletype_doc,
"A Observable is an object that is notified whenever something is changed \n\
in an Observable. An example of an Observable is a RunState object. \n\
An Observable has to register to the Observable in order to get notifications.\n\
\n\
Notifications are send by invoquing the update method in the Observable\n\
whose argument is a python object corresponding to the argument sent\n\
by the Observable\n\
"
);

static PyObject *daqpp_observable_iter(daqpp_Observable *self);

PyTypeObject Observable_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.Observable",             /*tp_name*/
   sizeof(daqpp_Observable),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)observable_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0, //(getattrfunc)observable_getattr,                         /*tp_getattr*/
   0, //(setattrfunc)observable_setattr,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   PyObject_GenericGetAttr,                         /*tp_getattro*/
   PyObject_GenericSetAttr,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   observabletype_doc,           /* tp_doc */
       0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    (getiterfunc)daqpp_observable_iter,		               /* tp_iter */
    0,		               /* tp_iternext */
    observable_methods,             /* tp_methods */
    observable_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    offsetof(daqpp_Observable, dict),                         /* tp_dictoffset */
   (initproc)observable_init,      /* tp_init */
    0,                         /* tp_alloc */
    observable_new,                 /* tp_new */
};

/*
 * The iterator
 */
typedef struct {
      PyObject_HEAD
      const Observable *obs;
      ObserverList::const_iterator iter;
} daqpp_observable_iter_object;

static PyObject *
daqpp_observable_iter_next(daqpp_observable_iter_object *it) 
{
   if ( it->iter != it->obs->Observers().end() ) {
      return PyDAQpp_Observable_New( *it->iter++ );
   }  
   else 
      return NULL;
}

static void
daqpp_observable_iter_dealloc(daqpp_observable_iter_object *it)
{
   PyObject_GC_UnTrack(it);
   PyObject_GC_Del(it);
}

static int
daqpp_observable_iter_traverse(daqpp_observable_iter_object *it, visitproc visit, void *arg)
{
   return 0;
}

static PyTypeObject PyDAQppObservableIter_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                                      /* ob_size */
   "bufferiterator",                        /* tp_name */
   sizeof(daqpp_observable_iter_object),      /* tp_basicsize */
   0,                                      /* tp_itemsize */
   /* methods */
   (destructor)daqpp_observable_iter_dealloc,		/* tp_dealloc */
   0,                                      /* tp_print */
   0,                                      /* tp_getattr */
   0,                                      /* tp_setattr */
   0,                                      /* tp_compare */
   0,                                      /* tp_repr */
   0,                                      /* tp_as_number */
   0,                                      /* tp_as_sequence */
   0,                                      /* tp_as_mapping */
   0,                                      /* tp_hash */
   0,                                      /* tp_call */
   0,                                      /* tp_str */
   0,                /* tp_getattro */
   0,                                      /* tp_setattro */
   0,                                      /* tp_as_buffer */
   Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,/* tp_flags */
   0,                                      /* tp_doc */
   (traverseproc)daqpp_observable_iter_traverse,	/* tp_traverse */
   0,					/* tp_clear */
   0,                                      /* tp_richcompare */
   0,                                      /* tp_weaklistoffset */
   PyObject_SelfIter,			/* tp_iter */
   (iternextfunc)daqpp_observable_iter_next,		/* tp_iternext */
   0,					/* tp_methods */
};



static PyObject *daqpp_observable_iter(daqpp_Observable *self) {
   daqpp_observable_iter_object *it;
   it = PyObject_GC_New(daqpp_observable_iter_object, &PyDAQppObservableIter_Type);
   if (it == NULL)
      return NULL;

   Py_INCREF(self);
   it->obs = OBSERVABLE(self);
   it->iter = it->obs->Observers().begin();
   PyObject_GC_Track(it);
   return (PyObject *)it;
}



void register_observable(PyObject *dict) {
   daqpp_register_class(dict,"Observable",
			&Observable_Type,
			0);
}

PyObject *PyDAQpp_Observable_New(void *ptr) {
   WHOAMI;
   daqpp_Observable *ret;

   Observable *m =
      dynamic_cast<Observable *>((Observable*)ptr);

   if (!m) {
      PyErr_SetString(DAQpp_Error,"Parameter 1 of Observable_New is not an Observable");
      Py_INCREF(Py_None);
      return Py_None;
   }
   ret = PyObject_New(daqpp_Observable,&Observable_Type);
   if (ret == NULL) {
      PyErr_SetString(DAQpp_Error,"daqpp.Observable_New: no memory");
      return NULL;
   }
   ret->obs = m;
   //ret->x_attr = 0;
   ret->dict = PyDict_New();
   return (PyObject *)ret;
}
