//#define DAQpp_DEBUG 1
#include <Python.h>
#include <sstream>
#include <typeinfo>
#include <cxxabi.h>
#include "pydaqpp.h"
#include "pydaqpp_debug.h"

#include <DAQ++/exception.h>
#include <DAQ++/Parameter.h>
#include <map>
#include <vector>

#include "pydaqpp_types.h"

// a few utils
inline std::string trim_right(const std::string &s) {
   std::string b=" \t\n";
   std::string str = s;
   return str.erase( str.find_last_not_of( b ) +1 );
}
inline std::string trim_left(const std::string &s) {
   std::string b=" \t\n";
   std::string str = s;
   return str.erase( 0,str.find_first_not_of( b ) );
}
inline std::string trim_str(const std::string &s) {
   std::string str = s;
   return trim_left( trim_right( str ) );
}

typedef void (*SetterFunction)(DAQpp::Parameter *, PyObject *,bool);
typedef PyObject *(*GetterFunction)(DAQpp::Parameter *);
static std::map< std::string, SetterFunction > theSetMap;
static std::map< std::string, GetterFunction > theGetMap;
static bool theMap_init=false;
static std::string moduleDoc("\
This class represents a holder for named parameters.\n\
The problem is that DAQpp::ParamHolder store Par<T> object and before\n\
setting or retrieving we must know the type. To overcome this problem\n\
this module has a set of predefined types that can be figured out at\n\
runtime. For those parameter types there is not problem and the interaction\n\
with DAQ++ should be transparent. When the type cannot be figured out\n\
the module will call the get_as_string and get_from_string methods of the\n\
DAQ++ parameters. The supported types are:\n");

#define GET_PAR_VALUE(P,type) ((DAQpp::Par< type > *)(P))->get_value()
#define SET_PAR(P,type,v,notify) ((DAQpp::Par< type > *)(P))->set_value(v,notify)

// Set functions
static void set_bool(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,bool,PyInt_AsLong(o),notify);
}

static void set_short(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,short,PyInt_AsLong(o),notify);
}

static void set_int(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,int,PyInt_AsLong(o),notify);
}

static void set_long(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,long,PyInt_AsLong(o),notify);
}

static void set_ushort(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,unsigned short,PyInt_AsLong(o),notify);
}

static void set_uint(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,unsigned int,PyInt_AsLong(o),notify);
}

static void set_ulong(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,unsigned long,PyInt_AsLong(o),notify);
}

static void set_double(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,double,PyFloat_AsDouble(o),notify);
}

static void set_float(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,float,PyFloat_AsDouble(o),notify);
}

static void set_string(DAQpp::Parameter *p, PyObject *o,bool notify) {
   WHOAMI;
   SET_PAR(p,std::string,PyString_AsString(o),notify);
}

#define VECTOR_INT_SETTER(type,name)					\
static void set_vector_##name (DAQpp::Parameter *p, PyObject *o,bool notify) {	\
   WHOAMI;								\
   if (!PySequence_Check(o)) {						\
      PyErr_SetString(DAQpp_Error,"set_vector_" #name ": object is not a sequence"); \
      return;								\
   }									\
   PyObject *item=0;							\
   Py_ssize_t i,len = PySequence_Length(o);					\
   std::vector<type> v;							\
   if (len>0) {								\
      for (i=0;i<len;i++) {						\
          item = PySequence_GetItem(o,i);				\
          v. push_back( PyInt_AsLong(item) );				\
          Py_DECREF(item);						\
      }									\
   }									\
   SET_PAR(p,std::vector<type>,v,notify);                               \
}
#define VECTOR_FLOAT_SETTER(type,name)					\
   static void set_vector_##name (DAQpp::Parameter *p, PyObject *o, bool notify) {	\
   WHOAMI;								\
   if (!PySequence_Check(o)) {						\
      PyErr_SetString(DAQpp_Error,"set_vector_" #name ": object is not a sequence"); \
      return;								\
   }									\
   PyObject *item=0;							\
   Py_ssize_t i,len = PySequence_Length(o);					\
   std::vector<type> v;							\
   if (len>0) {								\
      for (i=0;i<len;i++) {						\
          item = PySequence_GetItem(o,i);				\
          v. push_back( PyFloat_AsDouble(item) );				\
          Py_DECREF(item);						\
      }									\
   }									\
   SET_PAR(p,std::vector<type>,v,notify);					\
}

VECTOR_INT_SETTER(bool, bool);
VECTOR_INT_SETTER(short, short);
VECTOR_INT_SETTER(int, int);
VECTOR_INT_SETTER(long,long);
VECTOR_INT_SETTER(unsigned short, ushort);
VECTOR_INT_SETTER(unsigned int, uint);
VECTOR_INT_SETTER(unsigned long,ulong);

VECTOR_FLOAT_SETTER(float,float);
VECTOR_FLOAT_SETTER(double,double);

static void set_object(DAQpp::Parameter *p, PyObject *o, bool notify) {
   WHOAMI;
   PyObject *old = DAQpp::get_par_value<PyObject *>(p);
   Py_DECREF(old);
   Py_INCREF(o);
   DAQpp::set_par_value<PyObject *>(p,o,notify);
}

// Get Functions
static PyObject *get_bool(DAQpp::Parameter *p) {
   WHOAMI;
   return PyInt_FromLong( GET_PAR_VALUE(p,bool) );
}

static PyObject *get_short(DAQpp::Parameter *p) {
   WHOAMI;
   return PyInt_FromLong( GET_PAR_VALUE(p,short) );
}

static PyObject *get_int(DAQpp::Parameter *p) {
   WHOAMI;
   return PyInt_FromLong( GET_PAR_VALUE(p,int) );
}

static PyObject *get_long(DAQpp::Parameter *p) {
   WHOAMI;
   return PyLong_FromLong( GET_PAR_VALUE(p,long) );
}

static PyObject *get_ushort(DAQpp::Parameter *p) {
   WHOAMI;
   return PyInt_FromLong( GET_PAR_VALUE(p,unsigned short) );
}

static PyObject *get_uint(DAQpp::Parameter *p) {
   WHOAMI;
   return PyInt_FromLong( GET_PAR_VALUE(p,unsigned int) );
}

static PyObject *get_ulong(DAQpp::Parameter *p) {
   WHOAMI;
   return PyLong_FromLong( GET_PAR_VALUE(p,unsigned long) );
}

static PyObject *get_float(DAQpp::Parameter *p) {
   WHOAMI;
   return PyFloat_FromDouble( GET_PAR_VALUE(p,float) );
}

static PyObject *get_double(DAQpp::Parameter *p) {
   WHOAMI;
   return PyFloat_FromDouble( GET_PAR_VALUE(p,double) );
}

static PyObject *get_string(DAQpp::Parameter *p) {
   WHOAMI;
   return PyString_FromString( GET_PAR_VALUE(p,std::string).c_str() );
}

#define VECTOR_INT_GETTER(type,name)				\
static PyObject *get_vector_##name (DAQpp::Parameter *p) {	\
   WHOAMI;							\
   std::vector<type> &v = GET_PAR_VALUE(p,std::vector<type>);	\
   std::vector<type>::iterator ip;				\
   PyObject *lst = PyList_New(0);				\
   PyObject *item=0;						\
   for (ip=v.begin();ip!=v.end();++ip) {			\
      item = PyInt_FromLong( *ip );				\
      PyList_Append(lst,item);					\
   }								\
   return lst;							\
}
#define VECTOR_FLOAT_GETTER(type,name)				\
static PyObject *get_vector_##name (DAQpp::Parameter *p) {	\
   WHOAMI;							\
   std::vector<type> &v = GET_PAR_VALUE(p,std::vector<type>);	\
   std::vector<type>::iterator ip;				\
   PyObject *lst = PyList_New(0);				\
   PyObject *item=0;						\
   for (ip=v.begin();ip!=v.end();++ip) {			\
      item = PyFloat_FromDouble( *ip );				\
      PyList_Append(lst,item);					\
   }								\
   return lst;							\
}

VECTOR_INT_GETTER(bool,bool);
VECTOR_INT_GETTER(short,short);
VECTOR_INT_GETTER(unsigned short,ushort);
VECTOR_INT_GETTER(int,int);
VECTOR_INT_GETTER(unsigned int,uint);
VECTOR_INT_GETTER(long,long);
VECTOR_INT_GETTER(unsigned long,ulong);
VECTOR_FLOAT_GETTER(float,float);
VECTOR_FLOAT_GETTER(double,double);


static PyObject *get_object(DAQpp::Parameter *p) {
   WHOAMI;
   PyObject *o = DAQpp::get_par_value<PyObject *>(p);
   return o;
}
static void init_map() {
   struct __S {
	 const char *name;
	 SetterFunction f;
   };
   struct __G {
	 const char *name;
	 GetterFunction f;
   };
   __S _slist[] = {
      {"bool",set_bool},
      {"short",set_short},
      {"int",set_int},
      {"long",set_long},
      {"unsigned short",set_ushort},
      {"unsigned int",set_uint},
      {"unsigned long",set_ulong},
      {"float",set_float},
      {"double",set_double},
      {"std::string",set_string},
      {"std::vector<bool, std::allocator<bool> >",set_vector_bool},
      {"std::vector<short, std::allocator<short> >",set_vector_short},
      {"std::vector<int, std::allocator<int> >",set_vector_int},
      {"std::vector<long, std::allocator<long> >",set_vector_long},
      {"std::vector<unsigned short, std::allocator<unsigned short> >",set_vector_ushort},
      {"std::vector<unsigned int, std::allocator<unsigned int> >",set_vector_uint},
      {"std::vector<unsigned long, std::allocator<unsigned long> >",set_vector_ulong},
      {"std::vector<float, std::allocator<float> >",set_vector_float},
      {"std::vector<double, std::allocator<double> >",set_vector_double},
      {"_object*",set_object},
      {0,0}
   };
   __G _glist[] = {
      {"bool",get_bool},
      {"short",get_short},
      {"int",get_int},
      {"long",get_long},
      {"unsigned short",get_ushort},
      {"unsigned int",get_uint},
      {"unsigned long",get_ulong},
      {"float",get_float},
      {"double",get_double},
      {"std::string",get_string},
      {"std::vector<bool, std::allocator<bool> >",get_vector_bool},
      {"std::vector<short, std::allocator<short> >",get_vector_short},
      {"std::vector<int, std::allocator<int> >",get_vector_int},
      {"std::vector<long, std::allocator<long> >",get_vector_long},
      {"std::vector<unsigned short, std::allocator<unsigned short> >",get_vector_ushort},
      {"std::vector<unsigned int, std::allocator<unsigned int> >",get_vector_uint},
      {"std::vector<unsigned long, std::allocator<unsigned long> >",get_vector_ulong},
      {"std::vector<float, std::allocator<float> >",get_vector_float},
      {"std::vector<double, std::allocator<double> >",get_vector_double},
      {"_object*",get_object},
      {0,0}
   };
   int i;
   for (i=0;_slist[i].name;i++)
      theSetMap[_slist[i].name]=_slist[i].f;

   for (i=0;_glist[i].name;i++)
      theGetMap[_glist[i].name]=_glist[i].f;

   // complete the module documentation
   std::map< std::string, SetterFunction >::iterator ip;
   for (ip=theSetMap.begin();ip!=theSetMap.end();++ip) {
      if (ip->first=="_object*")
	 moduleDoc += "\t+ PyObject *\n";
      else
	 moduleDoc += "\t+ " + ip->first + "\n";
   }
   moduleDoc.append("\n");

   theMap_init = true;
}

static std::string parse_type_name( const DAQpp::Parameter *par, std::string &name ) {
   const std::string ptype("DAQpp::Par<");
   const int psize=ptype.size();
   const std::type_info  &ti = typeid( *par );
   char *abi_out;
   int status;
   unsigned int zp,ip=0;
   abi_out = abi::__cxa_demangle(ti.name(), 0, 0, &status);
   name = abi_out;
   free(abi_out);
   ip = name.find(ptype);
   if (ip==name.npos)
      return std::string();

   zp = name.rfind(">");
   if (zp==name.npos)
      return std::string();

   return trim_str( name.substr(ip+psize,zp-ip-psize) );
}

int set_daqpp_parameter(DAQpp::Parameter *P, PyObject *O, bool notify) {
   std::map< std::string, SetterFunction >::iterator ip;
   std::string name;
   std::string type = parse_type_name(P,name);

#if defined(DAQpp_DEBUG)
   PyObject *pytype =  PyObject_Type(O);
   PyObject *str  =  PyObject_Str(pytype);
   PyObject *repr = PyObject_Str(O);

   std::cout << "Setting parameter " << P->get_name()
	     << " of type " << type
	     << " to something of type "
	     << PyString_AsString(str)
	     << "\n\t" << PyString_AsString(repr)
	     << std::endl;
   Py_DECREF(pytype);
   Py_DECREF(str);
   Py_DECREF(repr);
#endif

   ip = theSetMap.find(type);
   if (ip==theSetMap.end()) {
      PyObject *str = PyObject_Repr(O);
      std::string data( PyString_AsString( str ) );
      if ( PyString_Check(O) ) {
	 // Python strings have come between 2 '
	 data = data.substr(1,data.size()-2);
      }
      Py_DECREF(str);
      P->set_from_string( data, notify );
      return 0;
   }

   // set the parameter
   try {
       ip->second(P,O,notify);
       if (PyErr_Occurred())
       {
           daqpp_print_error();
           return 1;
       }
   }
   catch (DAQpp::ParameterTypeDoesNotMatch ex) {
       std::ostringstream ostr;
       ostr << "set_daqpp_parameter: " << ex.what();
       PyErr_SetString(DAQpp_Error,ostr.str().c_str());
       return 1;
   }
   return 0;
}

PyObject *get_daqpp_parameter(DAQpp::Parameter *P) {
   WHOAMI;
   std::map< std::string, GetterFunction >::iterator ip;
   std::string name;
   std::string type = parse_type_name(P,name);


#if defined(DAQpp_DEBUG)
   std::cout << "Getting parameter " << P->get_name()
	     << " of type " << type
	     << std::endl;
#endif


   ip = theGetMap.find(type);
   if (ip==theGetMap.end()) {
      /*
      std::ostringstream ostr;
      if (type.empty())
	 ostr << "get_daqpp_parameter: Cannot handle type: " << name;
      else
	 ostr << "get_daqpp_parameter: Cannot handle type: " << type;

      std::cout << ostr.str() << std::endl;
      std::cout << "Supported types: " << std::endl;
      for (ip=theGetMap.begin();ip!=theGetMap.end();++ip)
	 std::cout << "\t" << ip->first << std::endl;

      PyErr_SetString(DAQpp_Error,ostr.str().c_str());
      */
      std::string val = "|-<"+P->get_as_string()+">-|";
      return PyString_FromString(val.c_str());
   }

   PyObject *rc = 0;
   try {
      rc = ip->second(P);
   }
   catch (DAQpp::ParameterTypeDoesNotMatch ex) {
      std::ostringstream ostr;
      ostr << "get_daqpp_parameter: " << ex.what();
      PyErr_SetString(DAQpp_Error,ostr.str().c_str());
      return NULL;
   }

   return rc;
}

//--------------------------------------------------------------------
// The phyton object
//--------------------------------------------------------------------
class pyParamHolder : public DAQpp::ParameterHolder, public PyDAQpp {
   public:
      pyParamHolder() {}
};

static PyObject *
paramHolder_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
   daqpp_ParamHolder *self;

   WHOAMI;
   self = (daqpp_ParamHolder *)type->tp_alloc(type, 0);
   if (!self)
      return (PyObject *)0;

   self->h   = 0;
   self->dict = PyDict_New();
   return (PyObject *)self;
}

static void
paramHolder_dealloc(daqpp_ParamHolder* self)
{
   WHOAMI;
   Py_DECREF(self->dict);
   self->ob_type->tp_free((PyObject*)self);
}

static int
set_object_pointer(daqpp_ParamHolder *self, DAQpp::ParameterHolder *m) {
   WHOAMI;
   int rc=0;
   self->h = m;
   pyParamHolder *pm = dynamic_cast<pyParamHolder *>(m);
   if (pm)
      pm->set_object((PyObject *)self);

   return rc;
}

static int
paramHolder_init(daqpp_ParamHolder *self,PyObject *args, PyObject *kwargs) {
   int rc = 0;

   WHOAMI;

   pyParamHolder *m = new  pyParamHolder();

   if (set_object_pointer(self,m))
      rc=-1;

   return rc;
}



PyDoc_STRVAR(paramHolder_addpar_doc,
"ParamHolder.add_par(name,value)\n\
Adds a new 'named' parameter to the paramHolder.\n\
");
static PyObject *
paramHolder_add_par(daqpp_ParamHolder *self, PyObject *args, PyObject *kwargs)
{
    PyObject *par;
    char *name;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "sO:ParamHolder.add_par", &name, &par))
        return NULL;

    if (PyBool_Check(par))
    {
        self->h->add_par(name, new DAQpp::Par<bool>( par==Py_True ? true : false ));
    }
    else if (PyInt_Check(par))
    {
        self->h->add_par(name, new DAQpp::Par<long>(PyInt_AsLong(par)));
    }
    else if (PyLong_Check(par))
    {
        self->h->add_par(name, new DAQpp::Par<unsigned long>(PyLong_AsUnsignedLong(par)));
    }

    else if (PyFloat_Check(par))
    {
        self->h->add_par(name, new DAQpp::Par<double>(PyFloat_AsDouble(par)));
    }
//    else if (PySequence_Check(par))
//    {
//        PyObject *item=0;
//        Py_ssize_t i, len = PySequence_Length(par);
//
//        if (len==0)
//        {
//            // empty sequence
//            Py_INCREF(Py_None);
//            return Py_None;
//        }
//
//        // Now check that all the elements are of the same type
//        item = PySequence_Fast_GET_ITEM(par, 0);
//        bool good=true;
//        for (i=1; i<len; i++)
//        {
//            PyObject *obj = PySequence_Fast_GET_ITEM(par, i);
//            if (! PyObject_TypeCheck( obj, item->ob_type) )
//            {
//                good = false;
//                break;
//            }
//        }
//
//        if ( good )
//        {
//            if ( PyInt_CheckExact(item) )
//            {
//                std::vector<long int> iv;
//                for (i=0; i<len;i++)
//                {
//                    item = PySequence_Fast_GET_ITEM(par, i);
//                    iv.push_back(  PyInt_AsLong(item) );
//                }
//                self->h->add_par(name, new DAQpp::Par< std::vector<long int> >(iv));
//                Py_INCREF(Py_None);
//                return Py_None;
//            }
//
//            if ( PyFloat_CheckExact(item) )
//            {
//                std::vector<double> iv;
//                for (i=0; i<len;i++)
//                {
//                    item = PySequence_Fast_GET_ITEM(par, i);
//                    iv.push_back(  PyFloat_AsDouble(item) );
//                }
//                self->h->add_par(name, new DAQpp::Par< std::vector<double> >(iv));
//                Py_INCREF(Py_None);
//                return Py_None;
//            }
//
//            if ( PyString_Check(item) )
//            {
//                std::vector< std::string > iv;
//                for (i=0; i<len;i++)
//                {
//                    item = PySequence_Fast_GET_ITEM(par, i);
//                    iv.push_back(  PyString_AsString(item) );
//                }
//                self->h->add_par(name, new DAQpp::Par< std::vector< std::string> >(iv));
//                Py_INCREF(Py_None);
//                return Py_None;
//            }
//        }
//
//        Py_INCREF(par);
//        self->h->add_par(name, new DAQpp::Par<PyObject *>(par));
//
//    }
    else
    {
        Py_INCREF(par);
        self->h->add_par(name, new DAQpp::Par<PyObject *>(par));
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(paramHolder_findpar_doc,
"ParamHolder.get_par(name)\n\
Returns the parameter with the given name as an Observable.\n\
");
static PyObject *
paramHolder_find_par(daqpp_ParamHolder *self,PyObject *args, PyObject *kwargs)

{
    PyObject *ret=0;
    PyObject *par_name=0;
    int rc;
    char *name;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "s:ParamHolder.get_par", &name))
        return NULL;

    DAQpp::Parameter *par=0;

    try
    {
        par = self->h->get_par(name);
    }
    catch ( ... )
    {
        std::ostringstream ostr;
        ostr << "Parameter " << name << " does not exist";
        PyErr_SetString(DAQpp_Error,ostr.str().c_str());
        return NULL;
    }
    ret = PyDAQpp_Observable_New( (void *)par );
    par_name = PyString_FromString(par->get_name().c_str());
    rc = PyObject_SetAttrString(ret, "name", par_name);
    return ret;
}



PyDoc_STRVAR(paramHolder_getpar_doc,
"ParamHolder.get_par(name)\n\
Retrieves the value of a 'named' parameter from the paramHolder.\n\
");
static PyObject *
paramHolder_get_par(daqpp_ParamHolder *self,PyObject *args, PyObject *kwargs) {
   char *name;

   WHOAMI;
   if (!PyArg_ParseTuple(args,"s:ParamHolder.get_par",&name))
      return NULL;

   DAQpp::Parameter *par=0;

   try {
      par = self->h->get_par(name);
   }
   catch ( ... ) {
      std::ostringstream ostr;
      ostr << "Parameter " << name << " does not exist";
      PyErr_SetString(DAQpp_Error,ostr.str().c_str());
      return NULL;
   }
   return get_daqpp_parameter(par);
}

PyDoc_STRVAR(paramHolder_setpar_doc,
"ParamHolder.get_par(name,value)\n\
Sets the value of a 'named' parameter of the paramHolder.\n\
");
static PyObject *
paramHolder_set_par(daqpp_ParamHolder *self,PyObject *args, PyObject *kwargs) {
   PyObject *obj=Py_None;
   int  notify=0;
   char *name;

   WHOAMI;
   if (!PyArg_ParseTuple(args,"sO|i:ParamHolder.set_par",&name,&obj,&notify))
      return NULL;


   DAQpp::Parameter *par;
   try {
      par = self->h->get_par(name);
   }
   catch ( ... ) {
      std::ostringstream ostr;
      ostr << "Parameter " << name << " does not exist";
      PyErr_SetString(DAQpp_Error,ostr.str().c_str());
      return NULL;
   }
   if (set_daqpp_parameter(par,obj,notify!=0))
      return NULL;

   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(paramHolder_parnames_doc,
"Returns a tuple with the names of all defined parameters\n\
");
static PyObject *
paramHolder_par_names(daqpp_ParamHolder *self) {
   WHOAMI;

   // Count the parameters
   DAQpp::ParameterHolder::const_iterator ip;
   Py_ssize_t n=0;
   for (ip=self->h->begin();ip!=self->h->end();++ip)
   {
       if (ip->first != "__PyObject__")
           n++;
   }

   PyObject *o,*tuple = PyTuple_New(n);
   for (n=0,ip=self->h->begin();ip!=self->h->end();ip++)
   {
       if (ip->first != "__PyObject__")
       {
           o = PyString_FromString( ip->first.c_str() );
           PyTuple_SetItem(	tuple, n, o);
           n++;
       }
   }
   return tuple;
}

PyDoc_STRVAR(paramHolder_has_par_doc,
"Tells if the object has a parameter with the given name\n\
");
static PyObject *
paramHolder_has_par(daqpp_ParamHolder *self,PyObject *args, PyObject *kwargs)
{
   PyObject *obj=Py_None;
   char *name;

   WHOAMI;
   if (!PyArg_ParseTuple(args,"s:ParamHolder.has_par",&name))
         return NULL;

   bool rc = self->h->has_par(name);
   obj = PyBool_FromLong( rc ? 1 : 0);

   return obj;
}



static PyMethodDef paramHolder_methods[] = {
   {"add_par",(PyCFunction)paramHolder_add_par,METH_VARARGS,paramHolder_addpar_doc},
   {"get_par",(PyCFunction)paramHolder_get_par,METH_VARARGS,paramHolder_getpar_doc},
   {"set_par",(PyCFunction)paramHolder_set_par,METH_VARARGS,paramHolder_setpar_doc},
   {"par_names",(PyCFunction)paramHolder_par_names,METH_NOARGS,paramHolder_parnames_doc},
   {"find_par",(PyCFunction)paramHolder_find_par,METH_VARARGS,paramHolder_findpar_doc},
   {"has_par",(PyCFunction)paramHolder_has_par,METH_VARARGS,paramHolder_has_par_doc},
   {NULL}  /* Sentinel */
};


PyTypeObject ParamHolder_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.ParameterHolder",             /*tp_name*/
   sizeof(daqpp_ParamHolder),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)paramHolder_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   PyObject_GenericGetAttr,                         /*tp_getattro*/
   PyObject_GenericSetAttr,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   0,             /* tp_doc */
   0,		               /* tp_traverse */
   0,		               /* tp_clear */
   0,		               /* tp_richcompare */
   0,		               /* tp_weaklistoffset */
   0,		               /* tp_iter */
   0,		               /* tp_iternext */
   paramHolder_methods,       /* tp_methods */
   0,                         /* tp_members */
   0,                         /* tp_getset */
   0,                         /* tp_base */
   0,                         /* tp_dict */
   0,                         /* tp_descr_get */
   0,                         /* tp_descr_set */
   offsetof(daqpp_ParamHolder, dict),                         /* tp_dictoffset */
   (initproc)paramHolder_init,      /* tp_init */
   0,                         /* tp_alloc */
   paramHolder_new,                 /* tp_new */
};

void register_paramholder(PyObject *dict) {
   init_map();
   ParamHolder_Type.tp_doc = strdup(moduleDoc.c_str());
   daqpp_register_class(dict,"ParameterHolder",
			&ParamHolder_Type,
			0);
}
