//#define DAQpp_DEBUG 1
#include "pydaqpp.h"
#include <iostream>
#include <sstream>
#include <cstring>
#include <ctime>
#include <DAQ++/DAQmanager.h>
#include <DAQ++/RunManager.h>
#include "pydaqpp.h"
#include "DAQobserver.h"
#include "Observer.h"
#include "pydaqpp_types.h"
#include "pydaqpp_objects.h"
#include "pydaqpp_debug.h"

#define CHECK_PTR(self) {					\
   if ( RUN_MANAGER(self)==0 ) {					\
      std::ostringstream ostr;					\
      ostr << __FUNCTION__ << ": DAQpp::RunManager pointer is NULL";	\
      PyErr_SetString(DAQpp_Error,ostr.str().c_str());		\
      return NULL;						\
   }								\
}


void pyRunManager::PrepareForRun()
{
    call_void_method("PrepareForRun", NULL);
    clear_external_call();
}

void pyRunManager::EndOfEvent()
{
    call_void_method("EndOfEvent", NULL);
}

void pyRunManager::EndOfRun()
{
    WHOAMI;
    call_void_method("EndOfRun", NULL);
}

int pyRunManager::writeData(int size, const char *data)
{
    PyObject *obj, *buffer;
    int rc=size;
    WHOAMI;

    if ( !GetStatus().isRunning() )
        return 0;

    if (python_call() || !has_object() )
        return 0;

    PyGILState_STATE gs = PyGILState_Ensure();
    buffer = PyDAQpp_Buffer_New('c', size, const_cast<char *>(data));
    obj = PyObject_CallMethod(self(), "writeData", "O", buffer);
    if (!obj)
    {
        // Something bizarre happened
        if (PyErr_Occurred())
        {
            daqpp_print_error();
            PyErr_Print();
        }
        PyGILState_Release(gs);
        return 0;
    }
    rc = PyInt_AsLong(obj);
    Py_DECREF(obj);
    Py_DECREF(buffer);
    PyGILState_Release(gs);
    return rc;
}

int pyRunManager::check_attributes()
{
    int rc = 0;
    WHOAMI;
    if (!has_object())
        return 1;

    if (!PyObject_HasAttrString(self(), "writeData") )
    {
        PyErr_SetString(DAQpp_Error, "RunManager should define its own writeData method");
        return -1;
    }
    return rc;
}

static PyObject *runmanager_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    daqpp_RunManager *self;

    WHOAMI;
    self = (daqpp_RunManager *) type->tp_alloc(type, 0);
    if (!self)
        return (PyObject *)0;

    self->m = 0;

    return (PyObject *)self;
}

static PyObject *runmanager_refcount(PyObject *obj)
{
    return PyInt_FromLong(obj->ob_refcnt);
}

static void runmanager_dealloc(daqpp_RunManager* self)
{
    WHOAMI;
    if ( RUN_MANAGER(self))
        DAQobserver::theObserver()->remove_runmanager_ref((PyObject *)self);

    Py_DECREF(self->pH.dict);
    self->pH.ob_type->tp_free((PyObject*)self);
}

static int set_object_pointer(daqpp_RunManager *self, DAQpp::RunManager *m)
{
    int rc=0;
    WHOAMI;
    self->pH.h = dynamic_cast<DAQpp::ParameterHolder *>(m);
    self->m = m;

    pyRunManager *pm = dynamic_cast<pyRunManager *>(m);
    if (pm)
    {
        pm->set_object((PyObject *)self);
        if (pm->check_attributes()<0)
            rc=-1;
    }
    return rc;
}

int setup_daqpp_runmanager(DAQpp::RunManager *m, daqpp_RunManager *self)
{
    int rc = 0;
    DAQobserver::theObserver()->add_runmanager((void *)m, (PyObject *)self);
    if (set_object_pointer(self, m))
        rc=-1;

    return rc;
}

static int runmanager_init(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    int rc=0;
    PyObject *in=Py_None;
    PyObject *str;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &in))
        return -1;

    if (in!=Py_None)
    {
        PyObject *str = PyObject_Str(in);
        std::string v = PyString_AsString(str);
        pyRunManager *m = new pyRunManager(v );
        rc = setup_daqpp_runmanager(m, self);
        Py_DECREF(str);
    }
    return rc;
}


PyDoc_STRVAR(runmanager_delete_doc,"\
Deletes the DAQpp::RunManager. This immediatly invalidates the Python\n\
reference and the next access to it will raise a daqpp.DAQError\n\
");
static PyObject *runmanager_delete(daqpp_Module *self)
{
    CHECK_PTR(self);

    DAQobserver::theObserver()->remove_runmanager_ref((PyObject *)self);
    if ( RUN_MANAGER(self))
        delete RUN_MANAGER(self);

    RUN_MANAGER(self)=0;
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(runmanager_getid_doc,"Returns the RunManager id.");
static PyObject *runmanager_get_id(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    DAQpp::DAQid id = self->m->get_id();
    PyObject* daqid = PyDAQpp_DAQid_New((const char *)id);
    return daqid;
}

PyDoc_STRVAR(runmanager_GetStatus_doc,"\
Returns the RunManager's RunState.\n\
");
static PyObject *runmanager_GetStatus(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;

    PyObject* status = PyDAQpp_RunState_New();
    *RUN_STATE(status) = self->m->GetStatus();

    return status;
}


PyDoc_STRVAR(runmanager_addModule_doc,"\
Adds a Module. The input can be either a Module id or a\n\
module reference.\n\
");
static PyObject *runmanager_addModule(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    PyObject *module;
    CHECK_PTR(self);WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &module))
        return NULL;

    // we gave a module id instead of a module pointer
    if ( PyString_Check(module))
    {
        std::string v = PyString_AsString(module);
        DAQpp::Module *m = DAQpp::DAQmanager::theDAQmanager()->find_module(v);
        if (m)
        {
            RUN_MANAGER(self)->addModule(m);
        }
        else
        {
            std::ostringstream ostr;
            ostr << "RunManager.addModule: module "<< v << " does not exist";
            PyErr_SetString(DAQpp_Error, ostr.str().c_str());
            return NULL;
        }
    }
    else if ( PyDAQpp_Module_Check(module) )
    {
        Py_INCREF(module);
        self->m->addModule(MODULE(module));
    }
    else
    {
        std::ostringstream ostr;
        ostr << "RunManager.addModule: argument is not a DAQpp::Module";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(runmanager_deleteModule_doc,"\
Deletes a Module from the RunManager CommandRun List.\n\
The module still exists, but it is not associated with this\n\
RunManager anymore. The argument can be either a module reference\n\
of a Module id.\n\
");
static PyObject *runmanager_deleteModule(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    PyObject *module;
    WHOAMI;
    CHECK_PTR(self);

    if (!PyArg_ParseTuple(args, "O", &module))
        return NULL;

    if ( PyString_Check(module))
    {
        std::string v = PyString_AsString(module);
        DAQpp::Module *m = DAQpp::DAQmanager::theDAQmanager()->find_module(v);
        if (m)
        {
            RUN_MANAGER(self)->deleteModule(m);
        }
        else
        {
            std::ostringstream ostr;
            ostr << "RunManager.addModule: module "<< v << " does not exist";
            PyErr_SetString(DAQpp_Error, ostr.str().c_str());
            return NULL;
        }
    }
    else
    {
        Py_DECREF(module);
        self->m->deleteModule(MODULE(module));
    }

    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(runmanager_countModules_doc,"\
Returns the number of modules in the RunManager list\n\
");
static PyObject *runmanager_countModules(daqpp_RunManager *self)
{
    WHOAMI;
    CHECK_PTR(self);
    return PyInt_FromLong(self->m->countModules() );
}

PyDoc_STRVAR(runmanager_moveModules_doc,"\
Moves the Modules in the list to the given RunManager.\n\
");
static PyObject *runmanager_moveModules(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;
    CHECK_PTR(self);

    PyObject *pyrm;
    if (!PyArg_ParseTuple(args, "O", &pyrm))
        return NULL;

    DAQpp::RunManager *dest = dynamic_cast<DAQpp::RunManager *>(RUN_MANAGER(pyrm));
    if (dest)
        DAQpp::RunManager::moveModules(RUN_MANAGER(self),dest);
    else
    {
        std::ostringstream ostr;
        ostr << "RunManager.moveModules: argument is not a RunManager";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(runmanager_isActive_doc,"\
Tells if the RunManager is active, that is, if it will\n\
react to the DAQmanager commands.\n\
");
static PyObject *runmanager_isActive(daqpp_RunManager *self)
{
    WHOAMI;
    CHECK_PTR(self);
    return PyInt_FromLong(self->m->isActive() );
}

PyDoc_STRVAR(runmanager_Activate_doc,"\
Deactivates the RunManager. It will not listen to the \n\
DAQmanager commands until it is activated again.\n\
");
static PyObject *runmanager_Activate(daqpp_RunManager *self)
{
    WHOAMI;
    CHECK_PTR(self);
    self->m->Activate();
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(runmanager_deActivate_doc,"\
Deactivates the RunManager. It will not listen to the \n\
DAQmanager commands until it is activated again.\n\
");
static PyObject *runmanager_deActivate(daqpp_RunManager *self)
{
    WHOAMI;
    CHECK_PTR(self);
    self->m->deActivate();
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(runmanager_MxEvts_doc,"\
Sets the maximum number of events to acquire.\n\
");
static PyObject *runmanager_set_max_events(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    int nevts;
    CHECK_PTR(self);

    WHOAMI;
    if (!PyArg_ParseTuple(args, "i", &nevts))
        return NULL;

    self->m->set_max_events(nevts);

    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(runmanager_getMxEvts_doc,"\
Returns the maximum number of events to acquire.\n\
");
static PyObject *runmanager_get_max_events(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    CHECK_PTR(self);WHOAMI;
    return PyInt_FromLong(self->m->get_max_events() );
}

static PyObject *runmanager_PrepareForRun(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm;
    pm = PyRUN_MANAGER(self);

    if (pm->external_call())
        goto out;

    pm->set_python_call();
    self->m->PrepareForRun();
    pm->clear_python_call();

    out: Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_EndOfRun(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
    {
        pm->set_python_call();
        self->m->EndOfRun();
        pm->clear_python_call();
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_EndOfEvent(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
    {
        pm->set_python_call();
        self->m->EndOfEvent();
        pm->clear_python_call();
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Start(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Start();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Stop(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Stop();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Pause(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Pause();
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Continue(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Continue();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Reset(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;

    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Reset();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_PrepareToStop(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    int flg;
    CHECK_PTR(self);WHOAMI;

    if (!PyArg_ParseTuple(args, "i", &flg))
        return NULL;

    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->PrepareToStop(flg!=0);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_GetReady(daqpp_RunManager *self)
{
    CHECK_PTR(self);WHOAMI;

    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->GetReady();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_Trigger(daqpp_RunManager *self)
{
    WHOAMI;
    CHECK_PTR(self);

    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( !pm->external_call())
        self->m->Trigger();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_writeData(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    PyObject *buff=Py_None;
    int rc = 1;
    WHOAMI;
    CHECK_PTR(self);
    pyRunManager *pm= PyRUN_MANAGER(self);

    if ( pm->external_call() )
    {
        return PyInt_FromLong( 1);
    }

    if (!PyArg_ParseTuple(args, "O", &buff))
        return NULL;

    Py_ssize_t size;
    const char *cbuff;
    if (buff!=Py_None)
    {
        if (PyObject_AsReadBuffer(buff, (const void **)&cbuff, &size) )
        {
            if (PyErr_Occurred())
            {
                daqpp_print_error();
                PyErr_Print();
            }

            PyErr_SetString(DAQpp_Error, "RunManager.writeData should receive a buffer object");
            Py_DECREF(buff);
            return NULL;
        }
//        pyRunManager *pm= PyRUN_MANAGER(self);
//        if (pm)
//            pm->set_python_call();
//        rc = self->m->writeData(size, cbuff);
//        if (pm)
//            pm->clear_python_call();
        rc = size;
    }

    return PyInt_FromLong(rc );
}

static PyObject *runmanager_set_monitor(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    PyObject *monitor;

    WHOAMI;
    CHECK_PTR(self);
    if (!PyArg_ParseTuple(args, "O:RunManager.set_monitor", &monitor))
        return NULL;

    if (monitor==Py_None|| monitor==0)
    {
        self->m->set_monitor(0);
    }
    else
    {
        DAQpp::Monitor *m = ((daqpp_Monitor *)monitor)->m;
        if (!PyObject_CallMethod(monitor, "set_runmanager", "O", self))
        {
            if (PyErr_Occurred())
            {
                daqpp_print_error();
                PyErr_Print();
            }
        }
        else
            self->m->set_monitor(m);
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *runmanager_get_monitor(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    PyObject *monitor;
    CHECK_PTR(self);

    WHOAMI;
    DAQpp::Monitor *m = self->m->get_monitor();
    if (m==0)
    {
        monitor = Py_None;
        Py_INCREF(Py_None);
    }
    else
    {
        PyDAQpp *pm = dynamic_cast<PyDAQpp *>(m);
        if (!pm)
        {
            monitor = Py_None;
            Py_INCREF(Py_None);
        }
        monitor = pm->self();
        Py_INCREF(monitor);
    }
    return monitor;
}

PyDoc_STRVAR(runmanager_addObserver_doc,
"addObserver(param=o1, state=o2). \n\
 Adds an object that will be notified when any Parameter changes\n\
 if the param argument (o1) is given or when the RunState changes\n\
 if the state argumen(o2) is given\n\
");
static PyObject *runmanager_addObserver(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    static char *kwlist[] = { "param", "state", NULL };
    PyObject *params=0, *state=0;
    PyObject *ret=Py_None;
    CHECK_PTR(self);
    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OO:RunManager.addObserver", kwlist, &params, &state))
        return NULL;

    if (params!=NULL)
    {
        pyObserver *obs = new pyObserver(params);
        self->m->addObserver( *obs );
    }
    if ( state != NULL )
    {
        pyObserver *obs = new pyObserver(state);
        self->m->get_status().addObserver( *obs );
    }
    Py_INCREF(ret);
    return ret;
}

static PyObject *runmanager_get_modules(daqpp_RunManager *self, PyObject *args, PyObject *kwargs)
{
    int imod, nmodules;
    PyObject *tuple, *obj;
    CHECK_PTR(self);

    nmodules = self->m->countModules();
    tuple = PyTuple_New(nmodules);
    ObserverList::iterator im;
    for (imod=0, im=(*self->m)()->begin(); im != (*self->m)()->end(); im++, imod++)
    {
        DAQpp::Module *module = dynamic_cast<DAQpp::Module *>(*im);
        obj = PyString_FromString( module->get_id() );
        PyTuple_SetItem(tuple, imod, obj);
    }
    return tuple;
}


static PyMethodDef runmanager_methods[] = {
   {"refcnt",(PyCFunction)runmanager_refcount,METH_NOARGS,"."},
   {"delete",(PyCFunction)runmanager_delete, METH_NOARGS,runmanager_delete_doc},
   {"get_id",(PyCFunction)runmanager_get_id, METH_NOARGS,runmanager_getid_doc},
   {"GetStatus",(PyCFunction)runmanager_GetStatus, METH_NOARGS,runmanager_GetStatus_doc},
   {"addModule",(PyCFunction)runmanager_addModule, METH_VARARGS,runmanager_addModule_doc},
   {"deleteModule",(PyCFunction)runmanager_deleteModule, METH_VARARGS,runmanager_deleteModule_doc},
   {"countModules",(PyCFunction)runmanager_countModules, METH_NOARGS,runmanager_countModules_doc},
   {"moveModules",(PyCFunction)runmanager_moveModules, METH_VARARGS,runmanager_moveModules_doc},
   {"isActive",(PyCFunction)runmanager_isActive,METH_NOARGS,runmanager_isActive_doc},
   {"deActivate",(PyCFunction)runmanager_deActivate,METH_NOARGS,runmanager_deActivate_doc},
   {"Activate",(PyCFunction)runmanager_Activate,METH_NOARGS,runmanager_Activate_doc},
   {"set_max_events",(PyCFunction)runmanager_set_max_events, METH_VARARGS,runmanager_MxEvts_doc},
   {"get_max_events",(PyCFunction)runmanager_get_max_events, METH_NOARGS,runmanager_getMxEvts_doc},
   {"set_monitor",(PyCFunction)runmanager_set_monitor,METH_VARARGS,"Sets the Monitor object."},
   {"get_monitor",(PyCFunction)runmanager_get_monitor,METH_NOARGS,"Returns the Monitor object or None if there is none"},
   {"PrepareForRun",(PyCFunction)runmanager_PrepareForRun,METH_NOARGS,"Calls Prepare for Run"},
   {"EndOfRun",(PyCFunction)runmanager_EndOfRun,METH_NOARGS,"Calls EndOfRun at the end of the run"},
   {"EndOfEvent",(PyCFunction)runmanager_EndOfEvent,METH_NOARGS,"Calls EndOfEvent at the end of the event"},
   {"Start",(PyCFunction)runmanager_Start,METH_VARARGS,"Starts the runmanager acquisition"},
   {"Stop",(PyCFunction)runmanager_Stop,METH_VARARGS,"Stops the runmanager acquisition"},
   {"Pause",(PyCFunction)runmanager_Pause,METH_NOARGS,"Pauses the runmanager acquisition"},
   {"Continue",(PyCFunction)runmanager_Continue,METH_NOARGS,"Resumes the runmanager acquisition"},
   {"Reset",(PyCFunction)runmanager_Reset,METH_NOARGS,"Resets the runmanager"},
   {"GetReady",(PyCFunction)runmanager_GetReady,METH_NOARGS,"Configures the runmanager"},
   {"PrepareToStop",(PyCFunction)runmanager_PrepareToStop,METH_VARARGS,"Set the Stop a.s.a.p flag"},
   {"Trigger",(PyCFunction)runmanager_Trigger,METH_NOARGS,"The DAQ received a trigger"},
//   {"writeData",(PyCFunction)runmanager_writeData,METH_VARARGS,"writes teh data"},
   {"addObserver",(PyCFunction)runmanager_addObserver,METH_VARARGS|METH_KEYWORDS,runmanager_addObserver_doc},
   {"get_modules",(PyCFunction)runmanager_get_modules, METH_NOARGS, "Returns a list of Modules"},

   {NULL}  /* Sentinel */
};


PyDoc_STRVAR(runmanagertype_doc,
"A RunManager controls the acquisition of a number of Module objects\n\
that share a common acquisition style. It will also handle in\n\
the same way the storage of the data from the different\n\
Modules. This class is an interface between the DAQmanager and the\n\
Modules. It transmits the DAQ commands to the different Modules.\n\
\n\
A RunManager can define the following hooks:\n\
      - PrepareForRun(self) : This is called at the beginning of a run\n\
      - EndOfEvent(self)    : This is called at the end of an event\n\
      - EndOfRun(self)      : called at the end of a run\n\
\n\
If a RunManager wants to store the data, it has to define:\n\
      - writeData(self,buffer): It will receive the data of an event\n\
                                in buffer.\n\
\n\
Modules should be registered to a RunManager with RunManager.addModule()\n\
To unregister a Module, use RunManager.deleteModule.\n\
\n\
There are some differences with respect to the C++ RunManager object. \n\
They are related mainly with the associated RunState. In Python the\n\
RunState returned by GetStatus() is a copy of the actual RunState. \n\
As a consequence, RunManager implements addObserver, that calls \n\
directly to its associated RunState addObserver. \n\
");

#define DEFERRED_ADDRESS(ADDR) 0
PyTypeObject RunManager_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.RunManager",             /*tp_name*/
   sizeof(daqpp_RunManager),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)runmanager_dealloc, /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)0,                 /*tp_str*/
   0,                         /*tp_getattro*/
   0,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   runmanagertype_doc,           /* tp_doc */
   0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    runmanager_methods,             /* tp_methods */
    0,             /* tp_members */
    0,                         /* tp_getset */
    DEFERRED_ADDRESS(&ParamHolder_Type),                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)runmanager_init,      /* tp_init */
    0,                         /* tp_alloc */
    runmanager_new,                 /* tp_new */
};


void register_runmanager(PyObject *dict)
{
    RunManager_Type.tp_base= &ParamHolder_Type;
    daqpp_register_class(dict, "RunManager", &RunManager_Type, 0);
}

int PyDAQpp_RunManager_Check(PyObject *O)
{
    return PyType_IsSubtype(O->ob_type, &RunManager_Type);
}

PyObject *PyDAQpp_RunManager_New(void *ptr)
{
    daqpp_RunManager *ret;

    DAQpp::RunManager* m =dynamic_cast<DAQpp::RunManager *>((DAQpp::RunManager*)ptr);
    if (!m)
    {
        PyErr_SetString(DAQpp_Error, "Parameter 1 of RunManager_New is not a Runmanager");
        Py_INCREF(Py_None);
        return Py_None;
    }
    DAQpp::Parameter *par;
    if (m->has_par("__PyObject__"))
    {
        par = m->get_par("__PyObject__");
        ret = (daqpp_RunManager *)DAQpp::get_par_value<PyObject *>(par);
        Py_INCREF( (PyObject *)ret);
    }
    else
    {
        ret = PyObject_New(daqpp_RunManager,&RunManager_Type);
        if (ret == NULL)
        {
            PyErr_SetString(DAQpp_Error,"daqpp.Module_New: no memory");
            return NULL;
        }
        ret->pH.dict = PyDict_New();
        par = new DAQpp::Par<PyObject *>( (PyObject *)ret );
        m->add_par("__PyObject__",par);
        set_object_pointer((daqpp_RunManager *)ret,m);
        Py_INCREF( (PyObject *)ret);

        DAQobserver::theObserver()->add_runmanager((void *)m,(PyObject *)ret);
    }
//    try
//    {
//        par = m->get_par("__PyObject__");
//        ret = (daqpp_RunManager *)DAQpp::get_par_value<PyObject *>(par);
//        Py_INCREF( (PyObject *)ret);
//    }
//    catch ( DAQpp::ParameterNonExistent ex)
//    {
//        ret = PyObject_New(daqpp_RunManager,&RunManager_Type);
//        if (ret == NULL)
//        {
//            PyErr_SetString(DAQpp_Error,"daqpp.Module_New: no memory");
//            return NULL;
//        }
//        ret->pH.dict = PyDict_New();
//        par = new DAQpp::Par<PyObject *>( (PyObject *)ret );
//        m->add_par("__PyObject__",par);
//        set_object_pointer((daqpp_RunManager *)ret,m);
//
//        DAQobserver::theObserver()->add_runmanager((void *)m,(PyObject *)ret);
//    }
    return (PyObject *)ret;
}
