#include "pydaqpp.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <DAQ++/RunState.h>
#include "Observer.h"
#include "pydaqpp_types.h"

using namespace DAQpp;


std::ostream &operator<<(std::ostream &os, const RunState &st) {
   double rate = st.rate();
   std::string rate_units=" Hz";
   std::string thr_units = " Mb/s";

   if ( rate>1.e6 ) {
      rate /=1.e6;
      rate_units=" MHz";
   }
   else if (rate>1000.) {
      rate /=1000.;
      rate_units = " kHz";
   }

   double throughput = st.get_throughput()/st.time();
   if (throughput > 1048576.0)
   {
       throughput /= 1048576.0;
       thr_units = "Mb/s";
   }
   else if (throughput > 1024.0)
   {
       throughput /= 1024.0;
       thr_units = "kb/s";
   }
   else if (throughput > 1.0 )
   {
       thr_units = "b/s";
   }
   else
   {
       throughput *= 1024;
       thr_units = " mb/s";
   }

   std::cout << "Evts "   
	     << std::setiosflags(std::ios::fixed | std::ios::right)	
	     << std::setw(7) << st.n_trigger() 
	     << " time " 
	     << std::setw(6) << std::setprecision(1)
	     << st.time()
	     << " rate " 
	     << std::setw(8) << std::setprecision(1)
	     << rate 
	     << rate_units
	     << " eff " 
	     << std::setw(5) << std::setprecision(1)
	     << st.eff()
	     << " - "
	     << std::setw(7) << std::setprecision(1)
	     << throughput
	     << ' ' << thr_units
	     << std::flush;
   return os;
}


static PyObject *
runstate_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
   daqpp_RunState *self;
   self = PyObject_New(daqpp_RunState,type);
   if (self!=NULL) {
      self->state  = 0;
      self->dict = PyDict_New();
   }
   return (PyObject *)self;
}

static void
runstate_dealloc(daqpp_RunState* self)
{
   delete self->state;
   self->state = 0;
   Py_DECREF(self->dict);
   self->ob_type->tp_free((PyObject*)self);
}

static int
runstate_init(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state = new  RunState();
   return 0;
}

static PyObject *
runstate_str(daqpp_RunState *self) {
   std::ostringstream ostr;
   ostr << *self->state;
   PyObject *out = PyString_FromString(ostr.str().c_str());
   return out;
}


PyDoc_STRVAR(runstate_get_daqid_doc,"Returns the RunState daqid");
static PyObject *
runstate_get_daqid(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *name = PyString_FromString( (const char *)self->state->get_daqid() );
   return name;
}


PyDoc_STRVAR(runstate_get_name_doc,"Returns the RunState name");
static PyObject *
runstate_get_name(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *name = PyString_FromString( self->state->get_name().c_str() );
   return name;
}

PyDoc_STRVAR(runstate_set_name_doc,"Sets the RunState name");
static PyObject *
runstate_set_name(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   char *s;

   if (!PyArg_ParseTuple(args,"s:RunState.set_name",&s))
      return NULL;

   self->state->set_name(s);
   
   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(runstate_reset_doc,
"reset()\n\
Resets the internal counters.\n\
It will reset the run duration time, the number of triggers,\n\
saved events, etc.\n\
");
static PyObject *
runstate_reset(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state->reset();
   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(runstate_start_doc,
"start()\n\
Starts the timer");
static PyObject *
runstate_start(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state->start();
   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(runstate_stop_doc,
"stop()\n\
Stops the timer");
static PyObject *
runstate_stop(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state->stop();
   Py_INCREF(Py_None);
   return Py_None;
}


PyDoc_STRVAR(runstate_Nexp_doc,
"n_trigger()\n\
returns the number of expected triggers");
static PyObject *
runstate_Nexp(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyInt_FromLong( self->state->n_expected() );
   return ret;
}

PyDoc_STRVAR(runstate_Ntrig_doc,
"n_trigger()\n\
returns the number of acquired triggers");
static PyObject *
runstate_Ntrig(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyInt_FromLong( self->state->n_trigger() );	
   return ret;
}

PyDoc_STRVAR(runstate_Nwriten_doc,
"n_written()\n\
Returns the number of saved events");
static PyObject *
runstate_Nwriten(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyInt_FromLong( self->state->n_written() );	
   return ret;
}

PyDoc_STRVAR(runstate_new_trigger_doc,
"new_trigger()\n\
Increases the number of triggers");
static PyObject *
runstate_new_trigger(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state->new_trigger();
   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(runstate_new_writen_doc,
"new_written()\n\
Increases the number of saved events");
static PyObject *
runstate_new_writen(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   self->state->new_written();
   Py_INCREF(Py_None);
   return Py_None;
}

PyDoc_STRVAR(runstate_set_expected_doc,
"set_expected(ival)\n\
Sets the number of expected events");
static PyObject *
runstate_set_expected(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
    unsigned int ival;
    if (!PyArg_ParseTuple(args,"I:RunState.set_expected",&ival))
        return NULL;

    self->state->set_expected(ival);
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(runstate_inc_expected_doc,
"increase_expected(ival)\n\
Increases the number of expected events by the given value");
static PyObject *
runstate_increase_expected(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
    unsigned int ival;
    if (!PyArg_ParseTuple(args,"I:RunState.increase_expected",&ival))
        return NULL;

    self->state->increase_expected(ival);
    Py_INCREF(Py_None);
    return Py_None;
}



PyDoc_STRVAR(runstate_get_thrp_doc,
"Returns the throughput\n\
");
static PyObject *
runstate_get_thrp(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyFloat_FromDouble( self->state->get_throughput() );
   return ret;
}

PyDoc_STRVAR(runstate_eff_doc,
"Returns the efficiency.\n\
The Efficiency is defined as the ration between saved events\n\
and number of triggers.\n\
");
static PyObject *
runstate_eff(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyFloat_FromDouble( self->state->eff() );	
   return ret;
}

PyDoc_STRVAR(runstate_time_doc,"Returns the time elapsed since the start");
static PyObject *
runstate_time(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   PyObject   *ret = PyFloat_FromDouble( self->state->time() );	
   return ret;
}

PyDoc_STRVAR(runstate_rate_doc,
"RunState.rate(flg)\n\
It returns the acquisition rate.\n\
It can be the average rate over the whole acquisition period,\n\
if the argument is true, or the instantaneous rate if it is false.\n\
");
static PyObject *
runstate_rate(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   int z=0;
   PyObject   *ret;
   
   if (!PyArg_ParseTuple(args,"i:RunState.rate",&z))
      return NULL;

   ret =  PyFloat_FromDouble( self->state->rate(z!=0) );	
   
   return ret;
}

PyDoc_STRVAR(runstate_isRunning_doc,"Tells whether the state is Running or not");
static PyObject *
runstate_isRunning(daqpp_RunState *self) {
   return PyInt_FromLong( self->state->isRunning() );
}

PyDoc_STRVAR(runstate_isReady_doc,"Tells whether the state is Ready");
static PyObject *
runstate_isReady(daqpp_RunState *self) {
   return PyInt_FromLong( self->state->isReady() );
}

PyDoc_STRVAR(runstate_isPaused_doc,"Tells whether the state is Paused");
static PyObject *
runstate_isPaused(daqpp_RunState *self) {
   return PyInt_FromLong( self->state->isPaused() );
}

PyDoc_STRVAR(runstate_isTimeout_doc,"Tells whether the state is TimeOut");
static PyObject *
runstate_isTimeout(daqpp_RunState *self) {
   return PyInt_FromLong( self->state->isTimeout() );
}

PyDoc_STRVAR(runstate_addObserver_doc,
"Adds an object that will be notified when the state changes\n\
");
static PyObject *
runstate_addObserver(daqpp_RunState *self,PyObject *args, PyObject *kwargs) {
   PyObject *obj;
   PyObject *ret=Py_None;
   if (!PyArg_ParseTuple(args,"O:RunState.addObserver",&obj))
      return NULL;

   
   pyObserver *obs = new pyObserver(obj);
   if (obs)
      self->state->addObserver( *obs );
   else {
      PyErr_SetString(DAQpp_Error,"RunState.addObserver: input is not an Observer");
      return NULL;
   }
   Py_INCREF(ret);
   return ret;
}

PyDoc_STRVAR(runstate_set_status_doc,
"RunState.set_status(status).\n\
Changes the status.\n\
");
static PyObject *
runstate_set_status(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   int z=0;
   PyObject   *ret=Py_None;
   
   if (!PyArg_ParseTuple(args,"i:RunState.set_status",&z))
      return NULL;

   self->state->set_status((RunState::StatusType)z);

   Py_INCREF(ret);
   return ret;
}

PyDoc_STRVAR(runstate_status_as_str_doc,"Returns as a string the current state.");
static PyObject *
runstate_status_as_str(daqpp_RunState *self, PyObject *args, PyObject *kwargs) {
   return PyString_FromString(
      self->state->status_str( self->state->get_status() ).c_str()
      );
}

static PyMethodDef runstate_methods[] = {
   {"get_daqid",(PyCFunction)runstate_get_daqid, METH_NOARGS, runstate_get_daqid_doc},
   {"get_name",(PyCFunction)runstate_get_name, METH_NOARGS,runstate_get_name_doc},
   {"set_name",(PyCFunction)runstate_set_name, METH_VARARGS ,runstate_set_name_doc},
   {"reset",(PyCFunction)runstate_reset, METH_NOARGS,runstate_reset_doc},
   {"start",(PyCFunction)runstate_start,METH_NOARGS,runstate_start_doc},
   {"stop",(PyCFunction)runstate_stop,METH_NOARGS,runstate_stop_doc},
   {"n_expected",(PyCFunction)runstate_Nexp,METH_NOARGS,runstate_Nexp_doc},
   {"n_trigger",(PyCFunction)runstate_Ntrig,METH_NOARGS,runstate_Ntrig_doc},
   {"n_written",(PyCFunction)runstate_Nwriten,METH_NOARGS,runstate_Nwriten_doc,},
   {"new_trigger",(PyCFunction)runstate_new_trigger,METH_NOARGS,runstate_new_trigger_doc},
   {"new_written",(PyCFunction)runstate_new_writen,METH_NOARGS,runstate_new_writen_doc},
   {"set_expected",(PyCFunction)runstate_set_expected,METH_NOARGS,runstate_set_expected_doc},
   {"increase_expected",(PyCFunction)runstate_increase_expected,METH_NOARGS,runstate_inc_expected_doc},
   {"throughput",(PyCFunction)runstate_get_thrp,METH_NOARGS,runstate_get_thrp_doc},
   {"eff",(PyCFunction)runstate_eff,METH_NOARGS,runstate_eff_doc},
   {"rate",(PyCFunction)runstate_rate,METH_VARARGS ,runstate_rate_doc},
   {"time",(PyCFunction)runstate_time,METH_NOARGS,runstate_time_doc},
   {"isRunning",(PyCFunction)runstate_isRunning, METH_NOARGS,runstate_isRunning_doc},
   {"isReady",(PyCFunction)runstate_isReady, METH_NOARGS,runstate_isReady_doc},
   {"isPaused",(PyCFunction)runstate_isPaused, METH_NOARGS,runstate_isPaused_doc},
   {"isTimeout",(PyCFunction)runstate_isTimeout, METH_NOARGS,runstate_isTimeout_doc},
   {"addObserver",(PyCFunction)runstate_addObserver,METH_VARARGS,runstate_addObserver_doc},
   {"set_status",(PyCFunction)runstate_set_status,METH_VARARGS,runstate_set_status_doc},
   {"status",(PyCFunction)runstate_status_as_str, METH_NOARGS,runstate_status_as_str_doc},
   {NULL}  /* Sentinel */
};
static PyMemberDef runstate_members[] = {
    {NULL}  /* Sentinel */
};

static PyTypeObject RunState_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
   "daqpp.RunState",             /*tp_name*/
   sizeof(daqpp_RunState),       /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)runstate_dealloc,                         /*tp_dealloc*/
   0,                         /*tp_print*/
   0,                         /*tp_getattr*/
   0,                         /*tp_setattr*/
   0,                         /*tp_compare*/
   (reprfunc)0,                         /*tp_repr*/
   0,                         /*tp_as_number*/
   0,                         /*tp_as_sequence*/
   0,                         /*tp_as_mapping*/
   0,                         /*tp_hash */
   (ternaryfunc)0,                         /*tp_call*/
   (reprfunc)runstate_str,                 /*tp_str*/
   PyObject_GenericGetAttr,                         /*tp_getattro*/
   PyObject_GenericSetAttr,                         /*tp_setattro*/
   0,                         /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
   "RunState wrapper",           /* tp_doc */
   0,		               /* tp_traverse */
   0,		               /* tp_clear */
   0,		               /* tp_richcompare */
   0,		               /* tp_weaklistoffset */
   0,		               /* tp_iter */
   0,		               /* tp_iternext */
   runstate_methods,             /* tp_methods */
   runstate_members,             /* tp_members */
   0,                         /* tp_getset */
   0,                         /* tp_base */
   0,                         /* tp_dict */
   0,                         /* tp_descr_get */
   0,                         /* tp_descr_set */
   offsetof(daqpp_RunState, dict),                         /* tp_dictoffset */
   (initproc)runstate_init,      /* tp_init */
   0,                         /* tp_alloc */
   runstate_new,                 /* tp_new */
};


PyObject *PyDAQpp_RunState_New() {
   daqpp_RunState *ret;
   ret = PyObject_New(daqpp_RunState,&RunState_Type);
   if (ret == NULL)
      return PyErr_NoMemory();

   ret->state = new RunState();
   ret->dict = PyDict_New();
   return (PyObject *)ret;
}

void register_runstate(PyObject *dict) {
   daqpp_register_class(dict,"RunState",
			&RunState_Type,
			0);
   daqpp_add_int_constant(dict,"NotDefined",RunState::NotDefined);
   daqpp_add_int_constant(dict,"NotReady",RunState::NotReady);
   daqpp_add_int_constant(dict,"GettingReady",RunState::GettingReady);
   daqpp_add_int_constant(dict,"Ready",RunState::Ready);
   daqpp_add_int_constant(dict,"Starting",RunState::Starting);
   daqpp_add_int_constant(dict,"Paused",RunState::Paused);
   daqpp_add_int_constant(dict,"Running",RunState::Running);
   daqpp_add_int_constant(dict,"Stopping",RunState::Stopping);
   daqpp_add_int_constant(dict,"Stopped",RunState::Stopped);
   daqpp_add_int_constant(dict,"TimeOut",RunState::TimeOut);
}

