#include <Python.h>
#include <sstream>
#include "DAQobserver.h"
#include "pydaqpp.h"
#include "pydaqpp_debug.h"

DAQpp::Mutex PyDAQpp::mtx;
PyObject *DAQpp_Error=NULL;

/* Obtains a string from a Python traceback.
   This is the exact same string as "traceback.print_exc" would return.

   Pass in a Python traceback object (probably obtained from PyErr_Fetch())
   Result is a string which must be free'd using PyMem_Free()
*/
#define TRACEBACK_FETCH_ERROR(what) {errMsg = what; goto done;}

char *PyTraceback_AsString(PyObject *exc_tb)
{
    const char *errMsg = NULL; /* a static that hold a local error message */
    char *result = NULL; /* a valid, allocated result. */
    PyObject *modStringIO = NULL;
    PyObject *modTB = NULL;
    PyObject *obFuncStringIO = NULL;
    PyObject *obStringIO = NULL;
    PyObject *obFuncTB = NULL;
    PyObject *argsTB = NULL;
    PyObject *obResult = NULL;

    /* Import the modules we need - cStringIO and traceback */
    modStringIO = PyImport_ImportModule("cStringIO");
    if (modStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant import cStringIO\n");

    modTB = PyImport_ImportModule("traceback");
    if (modTB==NULL)
        TRACEBACK_FETCH_ERROR("cant import traceback\n");
    /* Construct a cStringIO object */
    obFuncStringIO = PyObject_GetAttrString(modStringIO, "StringIO");
    if (obFuncStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant find cStringIO.StringIO\n");
    obStringIO = PyObject_CallObject(obFuncStringIO, NULL);
    if (obStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cStringIO.StringIO() failed\n");
    /* Get the traceback.print_exception function, and call it. */
    obFuncTB = PyObject_GetAttrString(modTB, "print_tb");
    if (obFuncTB==NULL)
        TRACEBACK_FETCH_ERROR("cant find traceback.print_tb\n");

    argsTB = Py_BuildValue("OOO",
            exc_tb  ? exc_tb  : Py_None,
            Py_None,
            obStringIO);
    if (argsTB==NULL)
        TRACEBACK_FETCH_ERROR("cant make print_tb arguments\n");

    obResult = PyObject_CallObject(obFuncTB, argsTB);
    if (obResult==NULL)
        TRACEBACK_FETCH_ERROR("traceback.print_tb() failed\n");
    /* Now call the getvalue() method in the StringIO instance */
    Py_DECREF(obFuncStringIO);
    obFuncStringIO = PyObject_GetAttrString(obStringIO, "getvalue");
    if (obFuncStringIO==NULL)
        TRACEBACK_FETCH_ERROR("cant find getvalue function\n");
    Py_DECREF(obResult);
    obResult = PyObject_CallObject(obFuncStringIO, NULL);
    if (obResult==NULL)
        TRACEBACK_FETCH_ERROR("getvalue() failed.\n");

    /* And it should be a string all ready to go - duplicate it. */
    if (!PyString_Check(obResult))
            TRACEBACK_FETCH_ERROR("getvalue() did not return a string\n");

    { // a temp scope so I can use temp locals.
    char *tempResult = PyString_AsString(obResult);
    result = (char *)PyMem_Malloc(strlen(tempResult)+1);
    if (result==NULL)
        TRACEBACK_FETCH_ERROR("memory error duplicating the traceback string");

    strcpy(result, tempResult);
    } // end of temp scope.
done:
    /* All finished - first see if we encountered an error */
    if (result==NULL && errMsg != NULL) {
        result = (char *)PyMem_Malloc(strlen(errMsg)+1);
        if (result != NULL)
            /* if it does, not much we can do! */
            strcpy(result, errMsg);
    }
    Py_XDECREF(modStringIO);
    Py_XDECREF(modTB);
    Py_XDECREF(obFuncStringIO);
    Py_XDECREF(obStringIO);
    Py_XDECREF(obFuncTB);
    Py_XDECREF(argsTB);
    Py_XDECREF(obResult);
    return result;
}


static void format_exception(std::string &msg,
                             PyObject *exc_typ,
                             PyObject *exc_val,
                             PyObject *exc_tb)
{
    std::ostringstream ostr;
    msg.clear();
    if (!exc_typ)
        return;

    if (exc_tb)
    {
        char *traceback = PyTraceback_AsString(exc_tb);
        if (traceback == NULL)
            ostr << "Can't get the traceback info!";
        else
        {
            ostr << "Traceback (most recent call last):\n" << traceback;
            PyMem_Free(traceback);
        }
    }
    PyObject *temp = PyObject_Str(exc_typ);
   if (temp)
   {
       ostr << PyString_AsString(temp);
       Py_DECREF(temp);
   }
   else
   {
       ostr << "Can't convert exception to a string!";
   }
   ostr << ": ";
   if (exc_val != NULL)
   {
       temp = PyObject_Str(exc_val);
       if (temp)
       {
           ostr << PyString_AsString(temp);
           Py_DECREF(temp);
       }
       else
       {
           ostr << "Can't convert exception value to a string!";
       }
   }
   msg = ostr.str();
}
void daqpp_print_error()
{
    std::string msg;
    PyObject *exc_typ = NULL, *exc_val = NULL, *exc_tb = NULL;
    PyErr_Fetch(&exc_typ, &exc_val, &exc_tb);
    PyErr_NormalizeException(&exc_typ, &exc_val, &exc_tb);
    if (exc_typ)
    {
        format_exception(msg, exc_typ, exc_val, exc_tb);
        if (!msg.empty())
            std::cout << msg << std::endl;
    }
    PyErr_Restore(exc_typ, exc_val, exc_tb);

}

void daqpp_register_class(PyObject *dict, const char *type_name, PyTypeObject *type, PyObject *bases)
{
    //PyObject *o;
    const char *s, *class_name;

    class_name=type->tp_name;
    s = strrchr(class_name, '.');
    if (s!=NULL)
        class_name = s+1;

    type->ob_type = &PyType_Type;
    if (bases)
    {
        type->tp_bases = bases;
        type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
    }
    if (PyType_Ready(type) < 0)
    {
        std::cerr << "couldn't make the type "<< type->tp_name<< " ready"<< std::endl;
        return;
    }

    PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
}

void daqpp_add_int_constant(PyObject *dict, const char *name, int value)
{
    PyObject *obj;
    obj = PyInt_FromLong(value);
    PyDict_SetItemString(dict, const_cast<char *>(name), obj);
    Py_DECREF(obj);
}



int PyDAQpp::get_rc(PyObject *obj, const char *name)
{
    int rc = 0;
    if (obj!=NULL)
    {
        if (!PyInt_Check(obj))
        {
            std::ostringstream ostr;
            ostr << name << " does not return an integer";
            PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        }
        else
        {
            rc = PyInt_AsLong(obj);
        }
    }
    return rc;
}

//#define DAQpp_DEBUG 1


static PyObject *create_argument_list(pycharp_t fmt, va_list va)
{
    PyObject *pyargs=0;

    if (fmt && *fmt)
    {
        pyargs = Py_VaBuildValue(fmt, va);
    }
    else
        pyargs = PyTuple_New(0);

    if (!pyargs)
        return 0;

    if (!PyTuple_Check(pyargs))
    {
        PyObject *a;

        a = PyTuple_New(1);
        if (PyTuple_SetItem(a, 0, pyargs) < 0)
            return 0;

        pyargs = a;
    }
    return pyargs;
}

int PyDAQpp::call_int_method(pycharp_t name, pycharp_t fmt, ...)
{
    int rc=-1;
    va_list args;
    PyObject *ret, *func, *pyargs=0;

    if (python_call() || !has_object() )
        return rc;

    PyGILState_STATE gs = PyGILState_Ensure();
    func = PyObject_GetAttrString(self(), name);
    if ( !func )
    {
        PyGILState_Release(gs);
        return rc;
    }

    va_start(args, fmt);
    pyargs = create_argument_list(fmt, args);
    va_end(args);
    if (!pyargs)
        goto end;

    set_external_call();
    ret = PyObject_Call(func, pyargs, NULL);
    if (ret)
    {
        rc = get_rc(ret, name);
        Py_DECREF(ret);
    }
    else
    {
        if (PyErr_Occurred())
        {
            daqpp_print_error();
            PyErr_Print();
        }
    }
    clear_external_call();

 end:
    Py_DECREF(func);
    if (pyargs)
    {
        Py_DECREF(pyargs);
    }
    PyGILState_Release(gs);
    return rc;
}

PyObject * PyDAQpp::call_obj_method(pycharp_t name, pycharp_t fmt, ...)
{
    va_list args;
    PyObject *ret=0, *func, *pyargs=0;

    if (python_call() || !has_object() )
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    PyGILState_STATE gs = PyGILState_Ensure();
    func = PyObject_GetAttrString(self(), name);
    if ( !func )
    {
        PyGILState_Release(gs);
        Py_INCREF(Py_None);
        return Py_None;
    }

    va_start(args, fmt);
    pyargs = create_argument_list(fmt, args);
    va_end(args);
    if (!pyargs)
        goto end;

    set_external_call();
    ret = PyObject_Call(func, pyargs, NULL);
    if (!ret)
    {
        if (PyErr_Occurred())
        {
            daqpp_print_error();
            PyErr_Print();
        }
    }
    else
    {
        Py_INCREF(ret);
    }
    clear_external_call();

 end:
    Py_DECREF(func);
    if (pyargs)
    {
        Py_DECREF(pyargs);
    }
    if (!ret)
    {
        ret = Py_None;
        Py_INCREF(ret);
    }
    PyGILState_Release(gs);
    return ret;
}


void PyDAQpp::call_void_method(pycharp_t name, pycharp_t fmt, ...)
{
    PyObject *ret, *func, *pyargs;
    va_list args;

    if (python_call() || !has_object() )
        return;

    PyGILState_STATE gs = PyGILState_Ensure();
    func = PyObject_GetAttrString(self(), name);
    if ( !func )
    {
        PyGILState_Release(gs);
        return;
    }

    va_start(args, fmt);
    pyargs = create_argument_list(fmt, args);
    va_end(args);
    if (!pyargs)
        goto end;

    set_external_call();
    ret = PyObject_Call(func, pyargs, NULL);
    if (!ret)
    {
        if (PyErr_Occurred())
        {
            daqpp_print_error();
            PyErr_Print();
        }
    }
    else Py_DECREF(ret)
;

    clear_external_call();
    end: Py_DECREF(func);
    if (pyargs)
    {
        Py_DECREF(pyargs);
    }
    PyGILState_Release(gs);
    return;
}

PyDAQpp::~PyDAQpp()
{
    WHOAMI;
}

void PyDAQpp::set_object(PyObject *o)
{
    WHOAMI;
    _self=o;
    Py_INCREF(o);
}
