#!/usr/bin/env python

"""daqpp

   This module defines a wrapper for the DAQ++ package, allowing to
   build a data acquisition system from python taking advantage of
   DAQ++.

   One can also interface with existing libraries that implement
   different models for Module and RunManager objects. For that daqpp
   provides load_lib that will return a Hook object from with you can
   call the Module and RunManager factories. daqpp uses dl for loading
   the shared libraries. It also 'borrows' the RTLD_* constants from
   the dl module.

   Note: The libraries should provide Object factories, that is,
   functions declared with extern 'C' that are able to create your
   objects from the arguments passed. They should be something like:

   extern 'C' void create_module(const char *type) {
      if ( type=='type1') {
         new ModuleType1(...)
      }
      else if ( type='...') {
          ...
      }
   }

   The idea behind is that you can afterwards access the module with
   the DAQmanager.find_module() or DAQmanager.find_runmanager() functions.
   
   This module also provides a helper function that does exactly that:
   - create_module
   - create_runmanager
"""

import os, sys
import ctypes
import inspect

from _daqpp import *

class HookList(dict):
    """HookList.

    Maintains a list of all the existing Hooks. This is implemented as
    a singleton. The basic idea behind is that we do not want the
    libraries to be unloaded until the very end, so that all their
    routines are available. 
    """
    def __new__(cls,*args,**kwargs):
        """That makes the class a singleton"""
        if '_inst' not in vars(cls):
            cls._inst = dict.__new__(cls,*args,**kwargs)

        return cls._inst
    
    def __init__(self):
        super(HookList,self).__init__()

    def __setitem__(self,key,value):
        if value.handler:
            super(HookList,self).__setitem__(key,value)
        else:
            print 'Invalid handler. Hook not stored'

    def __del__(self):
        print 'Deleting the list of Hooks'
        self.clear()

def load_lib(lib):
    """load_lib(lib)
    
       Load a shared library into the system. It will return a
       reference to a Hook object that contains the library handler
       that will allow to access to the library symbols.

       @param lib: the path of the shared library
       @return: A Hook object that will allow to access the
                library symbols
       
    """
    class Hook(object):
        """Hook,

        This class defines what I call a hook. It is in fact a holder
        of a shared library handler as given by the dl module. 

        To access the symbols or routines exported by the library, just
        call

            hook.func_name(...)

        where func_name is the name of a routine in the library. One
        can also use the call operator. This follows the dl module
        convention. 
        """
        def __init__(self,lib):
            self.handler=None
            try:
                self.lib    = lib
                self.handler = ctypes.cdll.LoadLibrary(self.lib)
            except SystemError,e:
                print e
                sys.exit(1)
            except dl.error,e:
                print e

        def __getattr__(self,name):
            if name not in self.__dict__ and self.handler:
                try:
                    function = getattr(self.handler, name)
                    def func(str,*args):
                        return function(str,*args)

                    self.__dict__[name] = func
                    return func
                except AttributeError:
                    raise AttributeError,'Symbol %s is not in %s'%(name,self.lib)
            else:
                return super(Hook,self).__getattr__(name)

        def __call__(self,name, str, *args):
            if self.handler:
                ff = getattr(self.handler, name)
                return ff(str,*args)
            else:
                raise AttributeError,'Invalid handler'

        def __del__(self):
            print '...deleting Hook(%s)' % self.lib
            del self.handler

    return HookList().setdefault(lib, Hook(lib))
    

def create_module(name, func_name, libpath):
    """ creates a module from a given library.
        @param name: The DAQid of the new module
        @param func_name: the name of the factory function that actually 
                          creates the module 
        @param libpath: the path of the library where dl will search for func_name
        @return: an instance of the module 
    """
    hook = load_lib(libpath)
    func = getattr(hook, func_name)
    func(name)
    return find_module(name)

def create_runmanager(name, func_name, libpath):
    """ creates a module from a given library.
        @param name: The DAQid of the new RunManager
        @param func_name: the name of the factory function that actually 
                          creates the RunManager 
        @param libpath: the path of the library where dl will search for func_name
        @return: an instance of the RunModule
    """
    hook = load_lib(libpath)
    func = getattr(hook, func_name)
    func(name)
    return find_runmanager(name)


import re
_spc=re.compile('\|-<(?P<str>.*)>-\|',re.MULTILINE|re.DOTALL)
def isSpecialParam(s):
    """ isSpecialParam(s)

    Parameters whose type cannot be easily figured out are sent and
    received as strings. When receiving, they are coded in a very
    'naive' way. This method tells you if this is one of those
    'special' parameter by returning the decoded string, and None if
    the type was properly gotten
    @param s: the parameter
    @returns: the string representation of the parameter value
    """

    try:
        r = _spc.match(s)
        if r:
            return r.group('str')
        else:
            return None
    except TypeError: # this is not a string
        return None


#------------------------------------------------------------------
