#include <DAQ++/python/pydaqpp.h>
#include <pythread.h>
#include <cstdarg>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <DAQ++/DAQmanager.h>
#include <DAQ++/ObserverModel.h>
#include <DAQ++/python/DAQobserver.h>
#include <DAQ++/python/Observer.h>
#include <DAQ++/python/pydaqpp_types.h>
#include <DAQ++/python/pydaqpp_trigger.h>
#include <DAQ++/python/pydaqpp_debug.h>
#include <DAQ++/soap/DAQppSOAP.h>

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif

PyDoc_STRVAR(module_doc,
"This module defines a Python wrapper for the DAQ++ package.\n\
It  allows to build a data acquisition system from Python taking advantage\n\
of DAQ++.\n\
\n\
This Python module (_daqpp) will be automatically imported by daqpp and all\n\
its symbols will be available from daqpp itself.\n\
\n\
All the functions defined here correspond to methods or utilities built up\n\
from methods belonging to the DAQpp::DAQmanager class.\n\
");

//====================================================================
// This is called when the library is unloaded
//====================================================================
#include <DAQ++/python/ModuleDecRef.h>

//static void the_end()  __attribute__((destructor));
static void the_end()
{
    const DAQpp::ModulesMap &mmap =DAQpp::DAQmanager::theDAQmanager()->ModuleMap();

    // First the RunManagers
    const DAQpp::RunManagerMap &rmap =DAQpp::DAQmanager::theDAQmanager()->RunMmap();
    DAQpp::RunManagerMap::const_iterator ir;
    for (ir=rmap.begin(); ir!=rmap.end(); ++ir)
    {
        // Free the modules attached to this run manager
        DAQpp::RunManager *rm = ir->second;
        rm->for_each_module(free_modules(rm) );
    }

    DAQpp::ModulesMap::const_iterator im;
    for (im=mmap.begin(); im!=mmap.end(); ++im)
    {
        module_decref(im->second);
        delete im->second;
    }

    std::cout << "...done"<< std::endl;
}


DAQpp::DAQppSOAP *the_soap_server = 0;
PyDoc_STRVAR(start_soap_server_doc,
"start_soap_server(port, threaded)\n\
Starts the DAQ++ soap server.\n\
@param port: the port where the server will listen\n\
@param threaded: An optional boolean (default is False). If true, the server will run in a different thread\n\
");
static PyObject *daqmanager_start_soap_server(PyObject *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;
    PyObject *rc = Py_None;
    int port;
    int threaded=0;

    if (!PyArg_ParseTuple(args, "i|i", &port, &threaded))
        return NULL;

    if (the_soap_server)
    {
        std::cout << "There is already a SOAP server running" << std::endl;
    }
    the_soap_server = new DAQpp::DAQppSOAP(port);
    the_soap_server->serve(threaded);
    if (threaded)
    {
        DAQpp::wait(0.1);
        rc = the_soap_server->is_running() ? Py_True : Py_False;
    }
    Py_IncRef(rc);
    return rc;
}

PyDoc_STRVAR(stop_soap_server_doc,
"stop_soap_server()\n\
Stops the DAQ++ soap server.\n\
");
static PyObject *daqmanager_stop_soap_server(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *rc = Py_None;
    if ( the_soap_server )
    {
        the_soap_server->stop_serving();
        delete the_soap_server;
        the_soap_server = 0;
    }
    Py_IncRef(rc);
    return rc;
}

PyDoc_STRVAR(get_module_list_doc,
"get_list_of_modules()\n\
Returns the names of all the modules.\n\
@returns: a list with the Module IDs\n\
");
static PyObject *daqmanager_get_list_of_modules(PyObject *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;
	PyObject *lst = PyList_New(0);
	PyObject *item=0;

    const DAQpp::ModulesMap &mmap =DAQpp::DAQmanager::theDAQmanager()->ModuleMap();
    DAQpp::ModulesMap::const_iterator im;
    for (im=mmap.begin(); im!=mmap.end(); ++im)
    {
    	item = PyString_FromString( im->first.c_str() );
    	PyList_Append(lst, item);
    }
	return lst;
}

PyDoc_STRVAR(get_runm_list_doc,
"get_runmanager_list()\n\
Returns the names of all defined RunManager\n\
@returns: a tuple with all the RunManager IDs\n\
");
static PyObject *daqmanager_get_list_of_runm(PyObject *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;
	PyObject *lst = PyList_New(0);
	PyObject *item=0;

    const DAQpp::RunManagerMap &rmap =DAQpp::DAQmanager::theDAQmanager()->RunMmap();
    DAQpp::RunManagerMap::const_iterator im;
    for (im=rmap.begin(); im!=rmap.end(); ++im)
    {
    	item = PyString_FromString( im->first.c_str() );
    	PyList_Append(lst, item);
    }
	return lst;
}

PyDoc_STRVAR(addModule_doc,
"addModule(module)\n\
Adds a module into the DAQmanager module list. The manager \n\
will take ownership of the module pointer.\n\
@param module: The module to add (Note that you still need to attach it to a RunManager)\n\
");
static PyObject *daqmanager_addModule(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *module;
    WHOAMI;

    if (!PyArg_ParseTuple(args, "O", &module))
        return NULL;

    if ( !PyDAQpp_Module_Check(module) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.addModule: argument is not a valid DAQpp::Module";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    DAQpp::Module *m= MODULE(module);
    if (m)
    {
        Py_INCREF(module);
        DAQpp::DAQmanager::theDAQmanager()->addModule(m);
    }
    else
    {
        std::ostringstream ostr;
        ostr << "Module object does not contain a valid DAQpp::Module pointer";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(deleteModule_doc,
"deleteModule(module)\n\
Deletes a module from the DAQmanager module list.\n\
@param module: the module object to be removed\n\
");
static PyObject *daqmanager_deleteModule(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *module;
    WHOAMI;

    if (!PyArg_ParseTuple(args, "O", &module))
        return NULL;

    if ( !PyDAQpp_Module_Check(module) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.addModule: argument is not a valid DAQpp::Module";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    DAQpp::Module *m= MODULE(module);
    if (m)
    {
        Py_DECREF(module);
        DAQpp::DAQmanager::theDAQmanager()->deleteModule(m);
        Py_INCREF(Py_None);
        return Py_None;
    }
    else
    {
        std::ostringstream ostr;
        ostr << "Module object does not contain a valid DAQpp::Module pointer";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
}

PyDoc_STRVAR(find_module_doc,
"find_module(id)\n\
@param id: the ID of the module to be retrieved\n\
@returns: the module with the given ID if it exists and None otherwise \n\
");
static PyObject *daqmanager_find_module(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *module=Py_None;
    char *id;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "s", &id))
        return NULL;

    DAQpp::Module *m = DAQpp::DAQmanager::theDAQmanager()->find_module(id);
    if (!m)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    PyDAQpp *ptr = dynamic_cast<PyDAQpp *>(m);
    if (ptr)
    {
        module = ptr->self();
        Py_INCREF(module);
    }
    else
    {
        module = PyDAQpp_Module_New(m);
    }

    return module;
}

PyDoc_STRVAR(countModules_doc,
"countModules()\n\
Returns the number of Module objects defined.\n\
");
static PyObject *daqmanager_countModules(PyObject *self)
{
    return PyInt_FromLong(DAQpp::DAQmanager::theDAQmanager()->countModules() );
}

PyDoc_STRVAR(addRM_doc,
"addRunManager(run_manager)\n\
Adds a new RunManager.\n\
@param run_manager: the RunManager to add\n\
");
static PyObject *daqmanager_addRunManager(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *runmanager;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &runmanager))
        return NULL;

    if ( !PyDAQpp_RunManager_Check(runmanager) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.addRunManager: argument is not a valid DAQpp::RunManager";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }

    DAQpp::RunManager *m= RUN_MANAGER(runmanager);

    DAQpp::DAQmanager::theDAQmanager()->addRunManager(m);
    Py_INCREF(runmanager);
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(delRM_doc,
"deleteRunManager(run_manager)\n\
Removes the given RunManager from the system.\n\
@param run_manager: the RunManager object to be removed\n\
");
static PyObject *daqmanager_deleteRunManager(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *runmanager;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &runmanager))
        return NULL;

    if ( !PyDAQpp_RunManager_Check(runmanager) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.deleteRunManager: argument is not a valid DAQpp::RunManager";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    DAQpp::RunManager *m= RUN_MANAGER(runmanager);

    Py_DECREF(runmanager);
    DAQpp::DAQmanager::theDAQmanager()->deleteRunManager(m);

    Py_INCREF(runmanager);
    return runmanager;
}

PyDoc_STRVAR(find_RM_doc,
"find_runmanager(id)\n\
@param id: the ID of the RunManager to be retrieved\n\
@returns: the RunManager with the given ID if it exists and None otherwise \n\
");
static PyObject *daqmanager_find_runmanager(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *runmanager=Py_None;
    char *id;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "s", &id))
        return NULL;

    DAQpp::RunManager *m = DAQpp::DAQmanager::theDAQmanager()->find_runmanager(id);

    if (!m)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    PyDAQpp *ptr = dynamic_cast<PyDAQpp *>(m);
    if (ptr)
    {
        runmanager = ptr->self();
        Py_INCREF(runmanager);
    }
    else
    {
        runmanager = PyDAQpp_RunManager_New(m);
    }

    return runmanager;
}

PyDoc_STRVAR(countRM_doc,
"countRunManager()\n\
Returns the number of RunManager objects defined.\n\
");
static PyObject *daqmanager_countRunManager(PyObject *self)
{
    WHOAMI;
    return PyInt_FromLong(DAQpp::DAQmanager::theDAQmanager()->countRunManager() );
}


PyDoc_STRVAR(startrun_doc,
"StartRun(flags)\n\
Starts a new run.\n\
This method starts a new run. It will set the running flag, \
update the list of active RunManager, reset their RunState, \
and call PrepareForRun for all of them. Once this is done, \
it will call Acquire for each of the active RunManager. Depending \
on the flags provided it can be executed in a new thread \
and/or run endlessly, regardless of the number of active RunManager, \
until the runnig flag is erased.\n\
@param flags: Tells how the manager will run. Possible values, if given, are Threaded or Loop\n\
");
static PyObject *daqmanager_StartRun(PyObject *self, PyObject *args, PyObject *kwargs)
{
    long flags;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "i", &flags))
        return NULL;

    pyDAQtrigger::running=true;
    DAQpp::DAQmanager::theDAQmanager()->StartRun(flags);

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(startRM_doc,
"startRunManager(run_manager)\n\
Starts the given RunManager.\n\
");
static PyObject *daqmanager_startRunManager(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *runmanager;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &runmanager))
        return NULL;

    if ( !PyDAQpp_RunManager_Check(runmanager) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.startRunManager: argument is not a valid DAQpp::RunManager";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }
    DAQpp::RunManager *m= RUN_MANAGER(runmanager);

    DAQpp::DAQmanager::theDAQmanager()->startRunManager(m);
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(stopRM_doc,
"stopRunManager(run_manager)\n\
Stops the given RunManager.\n\
");
static PyObject *daqmanager_stopRunManager(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *runmanager;

    WHOAMI;
    if (!PyArg_ParseTuple(args, "O", &runmanager))
        return NULL;

    if ( !PyDAQpp_RunManager_Check(runmanager) )
    {
        std::ostringstream ostr;
        ostr << "DAQpp.stopRunManager: argument is not a valid DAQpp::RunManager";
        PyErr_SetString(DAQpp_Error, ostr.str().c_str());
        return NULL;
    }

    DAQpp::RunManager *m= RUN_MANAGER(runmanager);

    DAQpp::DAQmanager::theDAQmanager()->stopRunManager(m);
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(waitEOR_doc,
"WaitForEndOfRun(tmout=-1)\n\
Waits for completion.\n\
Wait for completion means, in this case, that there are no active \
RunManager left.\n\
It is not equivalent to StopRun. WaitForEndOfRun may return while the run \
remains active. This is particularly true when StartRun is called in \
threaded mode.\n\
@keyword tmout: if not given or negative, the call blocks the thread execution,\
if positive it will quit after that time regardless of the run state\n\
");
static PyObject *daqmanager_WaitForEndOfRun(PyObject *self, PyObject *args, PyObject *kwargs)
{
    WHOAMI;
    int tmout=-1;
    if (!PyArg_ParseTuple(args, "|i", &tmout))
        return NULL;

    Py_BEGIN_ALLOW_THREADS
    DAQpp::DAQmanager::theDAQmanager()->WaitForEndOfRun(tmout);
    Py_END_ALLOW_THREADS
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(stoprun_doc,
"StopRun()\n\
Stops the current run, that is, the loop on the Active RunManager.\n\
");

static PyObject *daqmanager_StopRun(PyObject *self)
{

    WHOAMI;
    pyDAQtrigger::running=false;
    DAQpp::DAQmanager::theDAQmanager()->StopRun();
#if defined(DAQpp_DEBUG)
    std::cout << "...End of Run issued" << std::endl;
#endif

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(isrunning_doc,
"isRunning()\n\
Checks if the DAQ is running.\n\
");
static PyObject *daqmanager_isRunning(PyObject *self)
{
    WHOAMI;
    return PyInt_FromLong(DAQpp::DAQmanager::theDAQmanager()->isRunning() );
}

PyDoc_STRVAR(islooping_doc,
"isLooping()\n\
Checks if the DAQ is looping (runs in the same thread).\n\
");
static PyObject *daqmanager_isLooping(PyObject *self)
{
    WHOAMI;
    return PyInt_FromLong(DAQpp::DAQmanager::theDAQmanager()->isLooping() );
}

PyDoc_STRVAR(isthread_doc,
"isThreaded()\n\
Checks if the DAQ is running in a different thread.\n\
");
static PyObject *daqmanager_isThreaded(PyObject *self)
{
    WHOAMI;
    return PyInt_FromLong(DAQpp::DAQmanager::theDAQmanager()->isThreaded() );
}

PyDoc_STRVAR(settrigger_doc,
"set_trigger(trigger)\n\
Sets the DAQTrigger object that will decide when to execute one \
iteration of the DAQ loop.\n\
@param trigger: Should be a callable object returning True or False\n\
");
static PyObject *daqmanager_set_trigger(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *in;
    WHOAMI;
    if (!PyArg_ParseTuple(args, "O:DAQmanager.set_trigger", &in))
        return NULL;

    if (in==Py_None)
    {
        delete theTrigger;
        DAQpp::DAQmanager::theDAQmanager()->set_trigger(0);
    }
    else
    {
        if (!PyCallable_Check(in))
        {
            PyErr_SetString(DAQpp_Error, "DAQtrigger_init: parameter must be callable");
            return NULL;
        }

        Py_INCREF(in);

        delete theTrigger;
        theTrigger = new pyDAQtrigger(in);

        DAQpp::DAQmanager::theDAQmanager()->set_trigger(theTrigger);
    }

    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(add_observer_doc,
"add_observer(O)\n\
Adds a new observer to the DAQmanager.\n\
The argument must be an object of type Observer.\n\
This new object will receive the DAQmanager notifications\n\
@param O: This is an Observer object, that is, it implements and\n\
update method.\n\
");
static PyObject *daqmanager_add_observer(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *in, *ret=Py_None;
    WHOAMI;
    if (!PyArg_ParseTuple(args, "O:DAQmanager.add_observer", &in))
        return NULL;

    if (in==Py_None)
    {
    	PyErr_SetString(DAQpp_Error, "add_observer: parameter must not be None");
    	return NULL;
    }
    pyObserver *obs = new pyObserver(in);
    if (obs)
    {
    	DAQpp::DAQmanager::theDAQmanager()->addObserver( *obs );
    }
    else
    {
          PyErr_SetString(DAQpp_Error,"add_observer: input is not an Observer");
          return NULL;
    }
    Py_INCREF(ret);
    return ret;
}



PyDoc_STRVAR(runstate_status_str_doc,
"run_state(ival)\n\
Returns the string corresponding to the given RunState, if not argument is given.\n\
If a valid integer is given it should be a StatusType (Ready, Running, Stopping, ...).\n\
");
static PyObject *runstate_status_str(PyObject *self, PyObject *args, PyObject *kwargs)
{
    int s=0;
    WHOAMI;
    if (!PyArg_ParseTuple(args, "i:DAQmanager.run_state", &s))
        return NULL;
    return PyString_FromString(DAQpp::RunState::status_str((DAQpp::RunState::StatusType)s).c_str() );
}

PyDoc_STRVAR(daqpp_clean_doc,"clean()\n\
Deletes all the existing RunManager and Module objects.\n\
");
static PyObject *daqmanager_clean(PyObject *self, PyObject *args, PyObject *kwargs)
{

    the_end();

    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(daqpp_get_observable_doc,
"observable(O)\n\
Returns an object of type observable from the DAQ++\
object at the input. If the object does not derive from\
Observable returns None.\n\
This is quite handy to recover the Observable methods of a \
genuine DAQ++ object extending Observable, for instance Module \
or a RunManger.\n\
");
static PyObject *daq_get_observable(PyObject *self, PyObject *args, PyObject *kwargs)
{
    PyObject *in, *out;

    if (!PyArg_ParseTuple(args, "O:DAQmanager.get_observable", &in))
        return NULL;

    out = PyDAQpp_Observable_New( OBSERVABLE(in));
    return out;
}


static PyMethodDef daqpp_functions[] = {
   {"addModule",(PyCFunction)daqmanager_addModule, METH_VARARGS,addModule_doc},
   {"deleteModule",(PyCFunction)daqmanager_deleteModule, METH_VARARGS,deleteModule_doc},
   {"find_module",(PyCFunction)daqmanager_find_module, METH_VARARGS,find_module_doc},
   {"countModules",(PyCFunction)daqmanager_countModules, METH_NOARGS, countModules_doc},
   {"addRunManager",(PyCFunction)daqmanager_addRunManager, METH_VARARGS, addRM_doc},
   {"deleteRunManager",(PyCFunction)daqmanager_deleteRunManager, METH_VARARGS,delRM_doc},
   {"find_runmanager",(PyCFunction)daqmanager_find_runmanager, METH_VARARGS,find_RM_doc},
   {"countRunManager",(PyCFunction)daqmanager_countRunManager, METH_NOARGS,countRM_doc},
   {"StartRun",(PyCFunction)daqmanager_StartRun, METH_VARARGS,startrun_doc},
   {"startRunManager",(PyCFunction)daqmanager_startRunManager, METH_VARARGS,startRM_doc},
   {"stopRunManager",(PyCFunction)daqmanager_stopRunManager, METH_VARARGS,stopRM_doc},
   {"WaitForEndOfRun",(PyCFunction)daqmanager_WaitForEndOfRun, METH_VARARGS,waitEOR_doc},
   {"StopRun",(PyCFunction)daqmanager_StopRun, METH_NOARGS,stoprun_doc},
   {"isRunning",(PyCFunction)daqmanager_isRunning, METH_NOARGS, isrunning_doc},
   {"isLooping",(PyCFunction)daqmanager_isLooping, METH_NOARGS,islooping_doc},
   {"isThreaded",(PyCFunction)daqmanager_isThreaded, METH_NOARGS,isthread_doc},
   {"set_trigger",(PyCFunction)daqmanager_set_trigger,METH_VARARGS,settrigger_doc},
   {"run_state",(PyCFunction)runstate_status_str,METH_VARARGS,runstate_status_str_doc},
   {"clean",(PyCFunction)daqmanager_clean, METH_NOARGS,daqpp_clean_doc},
   {"observable",(PyCFunction)daq_get_observable,METH_VARARGS,daqpp_get_observable_doc},
   {"get_module_list",(PyCFunction)daqmanager_get_list_of_modules, METH_NOARGS,get_module_list_doc},
   {"get_runmanager_list",(PyCFunction)daqmanager_get_list_of_runm, METH_NOARGS,get_runm_list_doc},
   {"add_observer",(PyCFunction)daqmanager_add_observer,METH_VARARGS,add_observer_doc},
   {"start_soap_server", (PyCFunction)daqmanager_start_soap_server,METH_VARARGS,start_soap_server_doc},
   {"stop_soap_server", (PyCFunction)daqmanager_stop_soap_server,METH_VARARGS,stop_soap_server_doc},

   { NULL, NULL, 0 }
};



#if PY_VERSION_HEX < 0x2030000
extern "C" void
#else
PyMODINIT_FUNC
#endif
init_daqpp(void)

{
    PyObject *m,*d;
    DAQpp::DAQmanager *daqm = DAQpp::DAQmanager::theDAQmanager();

    if (!PyEval_ThreadsInitialized())
    	PyEval_InitThreads();

    m = Py_InitModule3("_daqpp",daqpp_functions,module_doc);
    d = PyModule_GetDict(m);

    register_daqid(d);
    register_paramholder(d);
    register_module(d);
    register_runmanager(d);
    register_monitor(d);
    register_runstate(d);
    register_progress(d);
    register_daqpp_buffer(d);
    register_fifo(d);
    register_gtimer(d);
    register_observer(d);
    register_observable(d);
    register_fileIOManager(d);

    if (PyErr_Occurred ())
    {
        Py_FatalError ("can't initialise module daqpp");
    }

    if ( DAQpp_Error == NULL )
    {
        DAQpp_Error = PyErr_NewException("_daqpp.DAQError", NULL, NULL);
        Py_INCREF(DAQpp_Error);
        PyModule_AddObject(m, "DAQError", DAQpp_Error);
    }

    daqpp_add_int_constant(d,"Threaded",DAQpp::DAQmanager::Threaded);
    daqpp_add_int_constant(d,"Loop",DAQpp::DAQmanager::Loop);
    daqpp_add_int_constant(d,"startDAQ",DAQpp::DAQmanager::startDAQ);
    daqpp_add_int_constant(d,"stopDAQ",DAQpp::DAQmanager::stopDAQ);
    daqpp_add_int_constant(d,"nwModule",DAQpp::DAQmanager::nwModule);
    daqpp_add_int_constant(d,"nwRunmanager",DAQpp::DAQmanager::nwRunManager);
    daqpp_add_int_constant(d,"rmModule",DAQpp::DAQmanager::rmModule);
    daqpp_add_int_constant(d,"rmRunManager",DAQpp::DAQmanager::rmRunManager);


    // Get prepared for the threading...
    PyEval_InitThreads();
    daqm->addObserver( *DAQobserver::theObserver() );
    //PyEval_ReleaseLock();
}

