/**
 * This example shows how to load a module defined in python
 */

#include <Python.h>
#include <iostream>
#include <iomanip>
#include <csignal>
#include <cmath>
#include <cerrno>
#include <cstring>

#include "PythonLoader.h"
#include <DAQ++/DAQmanager.h>
#include <DAQ++/RunManager.h>
#include <DAQ++/Module.h>
#include <DAQ++/GTimer.h>

using namespace DAQpp;
bool doRun = true;
void quit(int)
{
    doRun = false;
    std::cout << std::endl << "Run stop forced"<< std::endl;
}



class MyMonitor : public DAQpp::Monitor
{
    private:
        unsigned long counter;
        DAQpp::RunManager *rm;
        DAQpp::GTimer timer;
    public:
        MyMonitor(DAQpp::RunManager *r, int size = 50)
        : DAQpp::Monitor(size), counter(0), rm(r)
        {}
        unsigned long counts() const
        {
            return counter;
        }
        void on_start();
        void on_stop();
        void reset_counts() { counter=0; }
        bool user_run(DAQpp::Event *evt);
};

void MyMonitor::on_start()
{
    std::cerr << "Starting monitor" << std::endl;
    timer.start();
    reset_counts();
}

void MyMonitor::on_stop()
{
    timer.stop();
    std::cerr << "Monitor Stop. Events monitored: " << counter << std::endl;
}

bool MyMonitor::user_run(DAQpp::Event *evt)
{
    counter++;
    if ( timer()>0.25 )
    {
        std::cerr << "\r" << rm->GetStatus() << std::flush;
        timer.reset();
    }
    return true;
}



class SimpleRunManager : public DAQpp::RunManager
{
    public:
        SimpleRunManager(const DAQpp::DAQid &id, DAQpp::RunCommand *rc=0);
        virtual ~SimpleRunManager();

        void GetReady();
        void PrepareForRun();
        void Stop();
        void Start();
        void EndOfRun();
};
SimpleRunManager::SimpleRunManager(const DAQpp::DAQid &id, DAQpp::RunCommand *rc)
    : DAQpp::RunManager(id, rc)
{
}

SimpleRunManager::~SimpleRunManager()
{
    std::cout << "Destroying " << get_id() << std::endl;
}


void SimpleRunManager::GetReady()
{
    std::cout << get_id() << "-> GetReady" << std::endl;
    RunManager::GetReady();
}
void SimpleRunManager::PrepareForRun()
{
    std::cout << get_id() << "-> Starting the run" << std::endl;
    RunManager::PrepareForRun();
}
void SimpleRunManager::EndOfRun()
{
    std::cout << get_id() << "-> End of run" << std::endl;
    RunManager::EndOfRun();
}
void SimpleRunManager::Stop()
{
    std::cout << get_id() << "-> Stop" << std::endl;
    RunManager::Stop();
}
void SimpleRunManager::Start()
{
    std::cout << get_id() << "-> Start" << std::endl;
    RunManager::Start();
}

void show_stats(const RunManager *rm)
{
    std::cout << "*** Run statistics:\n"
              << rm->GetStatus() << std::endl
              << "data size/event: " 
              << rm->GetStatus().get_throughput()/rm->GetStatus().n_trigger()
              << std::endl;    
}

int main(int argc, char **argv)
{
    char *pyfile="simpleModule";
    char *pyhook="create_module";
    
    int nevts = 10000;
 
    if (argc>1)
        pyfile = argv[1];
    
    if (argc>2)
        pyhook = argv[2];
    
    // set up signal handler
    signal(SIGINT, quit);
    
    // Get the pointer to the DAQmanager
    DAQmanager *daq_manager = DAQmanager::theDAQmanager();
    
    // Create the run manager
    SimpleRunManager *run_manager = new SimpleRunManager("RunMngr");
    run_manager->set_monitor( new MyMonitor(run_manager) );
    
    // Create the module
    DAQpp::Module *module = PythonLoader::getInstance().create_module("MyModule", pyfile, pyhook);
    if (!module)
        return 0;
    run_manager->addModule( module );

    // We now start a run in another thread
    GTimer timer;
    std::cout << std::endl
              << "### A run in another thread. It should last for 5 sec."
              << std::endl;
    run_manager->set_max_events(0);
    run_manager->Activate();
    daq_manager->StartRun(DAQmanager::Threaded);

    timer.start();
    while ( doRun && timer() < 5.)
    {
        DAQpp::wait(0.5);
    }
    if (!doRun) {
        std::cout << "We had to quit..." << std::endl;
        return 0;
    }
    std::cout << "\n\nAfter " << timer() << " seconds we try to stop the run"
              << std::endl;
    
    daq_manager->StopRun();
    show_stats(run_manager);

    // Set the number of triggers so that it takes about
    // the same time as the other run
    int nhint = int(pow(10., int( log10(run_manager->GetStatus().n_trigger()))));
    nevts = int(nhint*(run_manager->GetStatus().n_trigger()/nhint));
    run_manager->set_max_events(nevts);


    // We now try to use wait for end of run
    std::cout << std::endl
              << "### A run in another thread. We wait for completion after "
              << run_manager->get_max_events()
              << " events..."
              << std::endl;
    run_manager->Activate();
    std::cout << "Starting Run" << std::endl;
    daq_manager->StartRun(DAQmanager::Threaded);
    daq_manager->WaitForEndOfRun();
    std::cout << "...run finished" << std::endl;
    std::cout << "manager is " << (daq_manager->isRunning() ? "running" : "stopped") << " and " << (daq_manager->isLooping() ? "looping" : "not looping")<< ". Active run managers " << daq_manager->countActiveRunManager() << std::endl;
    //daq_manager->StopRun();
    show_stats(run_manager);

    // We now start a run in this very same thread
    std::cout << std::endl
              << "### A run in this thread with "
              << run_manager->get_max_events()
              << " events "
              << std::endl;
    run_manager->Activate();
    daq_manager->StartRun(0);
    show_stats(run_manager);

    
    
    
    // Done
    return 0;
}
