#include <iostream>
#include <fstream>
#include <iomanip>
#include <csignal>
#include <sstream>
#include <DAQ++/DAQmanager.h>
#include <DAQ++/logger.h>
#include "SimpleRunManager.h"
#include "SimpleModule.h"
#include "random.h"
#include "Histogram.h"
#include "utils.h"

/*
  This example implements a DAQ system in which the data from the
  modules registered in the RunManager are stored in a histogram. The
  data from the modules are just numbers following a gaussian
  distribution.

  It also shows how to define named parameters in a module.
*/


bool DoRun = true;
void quit(int )
{
    DAQpp::ostream os;
    DoRun = false;
    os << DAQpp::loglevel(DAQpp::Log::message) << std::endl << "Run stop forced" << std::endl;
}

/**
   This is the run manager. I will store the data in a histogram
   instead of using a file .
*/
class Rmanager: public SimpleRunManager
{
    public:
        Rmanager(const DAQpp::DAQid &id) :
            SimpleRunManager(id)
        {
            hst = new Histogram("hst", "gauss", 100, -10., 10.);
        }
        ~Rmanager()
        {
            std::ofstream fout("test.histos");
            fout << *hst;
            fout.close();
            delete hst;
        }

    private:
        Histogram *hst;
};

/**
   This is a module.
   It implemets get data so that it returns a random number.
   It defines the mean and the standard deviation as a Parameter
*/
class MyModule: public SimpleModule
{
    public:
        MyModule(const DAQpp::DAQid &id, double mean, double std);
        bool hasData()
        {
            return true;
        }
        char *getData(int &size);

};

MyModule::MyModule(const DAQpp::DAQid &id, double mean, double std) :
    SimpleModule(id)
{
    add_par("mean", new DAQpp::Par<double>(mean));
    add_par("std", new DAQpp::Par<double>(std));
}

char *MyModule::getData(int &size)
{
    char   *buff = allocate(sizeof(double));
    double *val  = reinterpret_cast<double *>(buff);
    double _mean = DAQpp::get_par_value<double>(get_par("mean"));
    double _std  = DAQpp::get_par_value<double>(get_par("std"));
    *val = randgauss ( _mean, _std );
    size = sizeof(double);
    return buff;
}



template<class T> struct print_modules : public std::unary_function<T, void>
{
    print_modules(std::ostream& out) : os(out), count(0)
    {}
    void operator() (T x)
    {
        MyModule *m = dynamic_cast<MyModule *>(x);
        if ( m )
        {
            os << m->get_id() << ' ';
            ++count;
        }
    }
    std::ostream& os;
    int count;
};



int main(int argc, char  **argv)
{
    int max_nevts = 250000;
    const double run_duration = 5.;
    signal(SIGINT, quit);

    if ( argv[1] )
        max_nevts = atoi(argv[1]);

    // Create the run manager
    Rmanager *run_manager = new Rmanager("main");

    // Add the modules
    run_manager->addModule( new MyModule("1", -2., 0.25) );
    run_manager->addModule( new MyModule("2", -0., 1.5) );
    run_manager->addModule( new MyModule("3", 2., 0.25) );

    // test the for_each function in RunManager
    // This should just print the module IDs
    struct print_modules<Observer *> P =run_manager->for_each_module(print_modules<Observer *>(std::cout) );
    std::cout << std::endl<< "Number of Modules "<< P.count<< std::endl;

    // We should activate the RunManager
    DAQpp::GTimer timer;
    std::cout << std::endl<< std::endl<< "Starting a real run stopping after "<< run_duration << " sec."<< std::endl;
    run_manager->Activate();

    // Set an infinite number of events
    run_manager->set_max_events(0);

    /* ----------------------------------------------------------------------
       Start the DAQ
       ----------------------------------------------------------------------*/
    // Get the DAQmanager
    DAQpp::DAQmanager *daq_manager= DAQpp::DAQmanager::theDAQmanager();

    // start the run in a separate thread
    daq_manager->StartRun(DAQpp::DAQmanager::Threaded);
    timer.start();
    while ( timer() < run_duration && DoRun )
    {
        DAQpp::wait(0.1);
        std::cout << '\r' << run_manager->GetStatus() << std::flush;
    }
    std::cout << std::endl;
    daq_manager->stopRunManager(run_manager);
    daq_manager->StopRun();

    std::cout << std::endl << std::flush;

    // print a summary
    std::cout << "Run Stats. "<< std::endl<< *run_manager<< std::endl;

    // allow running again
    DoRun = true;


    /**
       Now we demonstrate the Loop way of running.
       First we deactivate the RunManager and we start the run.
       In principle the DAQmanaer loop will be waiting for active
       RunManager objects
    */
    if ( run_manager->isActive() )
        run_manager->deActivate();

    // Now activate the DAQmanager in the background
    DAQpp::DAQmanager::theDAQmanager()->StartRun(DAQpp::DAQmanager::Loop);



    /**
       Activate the run manager and start the run several times putting
       in local some of the modules
    */
    int i;
    DAQpp::Module *m;
    for (i = 0;i < 5;i++)
    {
        std::ostringstream ostr;
        ostr << ((i % 3) + 1);
        m = DAQpp::DAQmanager::theDAQmanager()->find_module(ostr.str());
        m->in_local();
        std::cout << "#### " << i << " module " << m->get_id() << " in local " << std::endl;
        run_manager->set_max_events(max_nevts);
        daq_manager->startRunManager(run_manager);
        DAQpp::RunState status;
        do
        {
            DAQpp::wait(0.5);
            status = run_manager->GetStatus();
            std::cout << '\r' << status << std::flush;
        } while ( status.isRunning() && DoRun );

        DAQpp::DAQmanager::theDAQmanager()->WaitForEndOfRun(100000);
        if ( run_manager->GetStatus().isRunning() )
            daq_manager->stopRunManager(run_manager);

        std::cout << "Run Stats. "<< std::endl<< *run_manager<< std::endl;
        m->in_global();
    }
    // And stop the DAQmanager loop
    daq_manager->StopRun();
}
