#include <iostream>
#include <iomanip>
#include <csignal>

#include <DAQ++/DAQmanager.h>
#include <DAQ++/logger.h>
#include "random.h"
#include "utils.h"
#include "SimpleModule.h"
#include "SimpleRunManager.h"

using namespace DAQpp;
bool doRun = true;
void quit(int)
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::none) << std::endl
       << DAQpp::loglevel(DAQpp::Log::message) << "Run stop forced" << std::endl;
    doRun = false;
}




template<class T> struct print_modules : public std::unary_function<T, void>
{
    print_modules(std::ostream& out) : os(out), count(0)
    {}
    void operator() (T x)
    {
        SimpleModule *m = dynamic_cast<SimpleModule *>(x);
        if ( m )
        {
            os << *m << std::endl;
            ++count;
        }
    }
    std::ostream& os;
    int count;
};

class RunSpy : public DAQpp::Thread
{
    private:
        DAQpp::RunManager *_rm;
    public:
        RunSpy(DAQpp::RunManager *rm) : _rm(rm)
        {}
        void run()
        {
            DAQpp::ostream os;
            while (is_running())
            {
                test_cancel();
                DAQpp::wait(0.1);
                DAQpp::RunState state = _rm->GetStatus();
                if (_rm->isActive() && state.isRunning())
                    os << DAQpp::loglevel(DAQpp::Log::none) << '\r' << state << std::flush;
            };
            os << DAQpp::loglevel(DAQpp::Log::none) << std::endl;
        }
};


int main(int argc, char **argv)
{
    int nevts = 100000;
    // set up signal handler
    signal(SIGINT, quit);
    
    if (argv[1])
    {
        nevts = atoi(argv[1]);
    }
    std::cout << "usage: simpleDaq [nevts]\n" << std::endl; 
    // Get the pointer to the DAQmanager
    DAQmanager *daq_manager = DAQmanager::theDAQmanager();
    // Create the run manager
    SimpleRunManager *run_manager = new SimpleRunManager("RunMngr");

    // Create an object that monitors the state of the RunManager
    RunSpy spy(run_manager);
    spy.start();

    // Create a few modules and add them to the RunManager
    run_manager->addModule( new SimpleModule("Mod1") );

    std::cout << "*** List of modules" << std::endl;
    run_manager->for_each_module(
                        print_modules<Observer *>(std::cout) );
    std::cout << "\n\n" << std::endl;

    // Activate the run manager
    run_manager->set_max_events(nevts);
    run_manager->Activate();

    // We now start a run in this very same thread
    std::cout << "### A run in this thread with "
              << run_manager->get_max_events()
              << " events "
              << std::endl;
    daq_manager->StartRun(0);
    std::cout << "*** " << *run_manager << std::endl;

    // We now start a run in another thread
    std::cout << std::endl
    << "### A run in another thread. It should last for 5 sec."
    << std::endl;
    GTimer timer;
    run_manager->set_max_events(0);
    run_manager->Activate();
    daq_manager->StartRun(DAQmanager::Threaded);

    timer.start();
    while ( doRun && timer() < 5.)
    {
//      std::cout << '\r' << run_manager->GetStatus() << std::flush;
        DAQpp::wait(0.1);
    }
    if (!doRun) {
        std::cout << "We had to quit..." << std::endl;
        return 0;
    }
    std::cout << "\n\nAfter " << timer() << " seconds we try to stop the run"
              << std::endl;
    DAQmanager::theDAQmanager()->StopRun();
    std::cout << "*** " << *run_manager << std::endl;

    // We now try to use wait for end of run
    run_manager->set_max_events(nevts);
    std::cout << std::endl
              << "### A run in another thread. We wait for completion after "
              << run_manager->get_max_events()
              << " events..."
              << std::endl;
    run_manager->Activate();
    daq_manager->StartRun(DAQmanager::Threaded);
    daq_manager->WaitForEndOfRun();
    if (!doRun) {
        std::cout << "We had to quit..." << std::endl;
        return 0;
    }
    spy.stop();
    std::cout << "*** " << *run_manager << std::endl;

    // Done
    return 0;
}
