#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <cmath>
#include "Histogram.h"



/*------------------------------------------------------------------------
 * Histogram output on files.
 *------------------------------------------------------------------------*/
std::ostream & operator<< ( std::ostream& os, const Histogram& h )
{
    h.Print(os);
    return os;
}

/*@(#)******************************************************************
 * constructors
 *
 *
 **********************************************************************/
Histogram::Histogram(const char *nam, const char *tit,
                     int nch, double x1, double x2)
        : nx(nch), xmin(x1), xmax(x2), hsum(0.), tsum(0.)
{
    name = strdup(nam);
    title = strdup(tit);

    init(0.0);

    step = (xmax - xmin) / ((double)nx);

}

int Histogram::write(std::ostream &os)
{
    int size = 0;
    os.write((char *)&nx, sizeof(int32_t)); size += sizeof(int32_t);
    os.write((char *)&xmin, 5*sizeof(double)); size += 5*sizeof(double);
    os.write((char *)data, nx*sizeof(double)); size += nx*sizeof(double);

    int sz = strlen(name)+1;
    os.write((char *)&sz, sizeof(int32_t)); size += sizeof(int32_t);
    os.write(name, sz); size += sz;

    sz = strlen(title)+1;
    os.write((char *)&sz, sizeof(int32_t)); size += sizeof(int32_t);
    os.write(title, sz); size += sz;

    return size;
}

void Histogram::read(std::istream &is)
{
    int sz;
    delete [] data;

    is.read((char *)&nx, sizeof(int32_t));
    is.read((char *)&xmin, 5*sizeof(double));

    data = new double[nx];
    is.read((char *)data, nx*sizeof(double));

    is.read((char *)&sz, sizeof(int32_t));
    name = (char *)malloc(sz);
    is.read(name, sz);

    is.read((char *)&sz, sizeof(int32_t));
    title = (char *)malloc(sz);
    is.read(title, sz);

}



/*@(#)******************************************************************
 * destroy
 *
 *
 **********************************************************************/
void Histogram::destroy()
{
    if ( name ) free(name);
    if ( title) free(title);
    if ( data ) delete [] data;
}

Histogram::Histogram( Histogram const &h )
{
    copy ( h ) ;
}

Histogram::~Histogram()
{
    destroy();
}

void Histogram::copy(Histogram const &h)
{
    nx = h.nx;
    xmin = h.xmin;
    xmax = h.xmax;
    hsum = h.hsum;
    tsum = h.tsum;

    if ( data )
        delete [] data;

    data = new double [nx];
    memcpy(data, h.data, nx*sizeof(double));


    step = (xmax - xmin) / ((double)nx);
}


void Histogram::init(double val)
{
    int i;

    // Allocate space for data
    data = new double [nx];
    assert( data != 0 );

    for ( i = 0;i < nx;i++ )
    {
        data[i] = val;
        hsum += val;
        tsum += val;
    }
}



Histogram const &Histogram::operator=(Histogram const &h)
{
    if ( this != &h )
        copy (h);

    return (*this);
}

double &Histogram::operator[](int index)
{
    if ( index<0 || index >= nx )
    {
        std::cerr << "Histogram: boundary overflow or underflow" << std::endl;
        exit(1);
    }
    return data[index];
}


Histogram &sqrt(const Histogram &hx)
{
    static Histogram h = hx;
    int i;

    for (i = 0; i < hx.nx; i++)
        h.data[i] = sqrt(hx.data[i]);

    return h;
}
void Histogram::Fill(double x, double y, double wgt)
{
    Fill(x, y*wgt);
}

void Histogram::Fill(double x, double wgt)
{
    int i;

    i = Bin(x);
    tsum += wgt;
    if ( i >= 0 && i < nx )
    {
        *(data + i) += wgt;
        hsum += wgt;
    }

}

void Histogram::reset()
{
    int i;
    for (i = 0; i < nx; i++)
        data[i] = 0.0;
}


/*@(#)******************************************************************
 * mean, var, stdDev
 *
 * computes Histogram mean. Only 1D histograms at the moment
 **********************************************************************/
double Histogram::Mean()
{
    int i;
    double xv, xw, xm, xsum;

    if ( hsum == 0.)
        return 0.;

    xm = xsum = 0.;
    for ( i = 0, xv = xmin + 0.5 * step;i < nx;i++, xv += step)
    {
        xw    = data[i];
        xsum += xw;
        xm   += xv * xw;
    }

    xm /= xsum;

    return xm;
}

double Histogram::Var()
{
    int i;
    double xv, xm, xw, xd, xs, xa;

    xm = Mean();
    xa = xs = 0.;
    for (i = 0, xv = xmin + 0.5 * step;i < nx;i++, xv += step)
    {
        xw = data[i];
        xd = xv - xm;
        xa += xw * xd;
        xs += xw;
    }

    xa /= xs;
    return xa;
}

double Histogram::stdDev()
{
    return sqrt( Var() );
}

void Histogram::Print(std::ostream &os) const
{
    int i;
    os << 0 << " " << nx << " " << xmin << " " << xmax << std::endl
    << name << std::endl
    << title << std::endl;

    std::ios::fmtflags  iflgs = os.flags();
    os.setf( std::ios::scientific );
    for (i = 0;i < nx;i++)
    {
        if ( i && ! (i % 6 ) )
            os << std::endl;
        os << std::setw(12) << std::setprecision(6) << data[i] << " ";
    }
    os << std::endl;
    os.flags( iflgs );
}

HProfile::HProfile(const char *name, const char *title,
                   int n, double x1, double x2, double y1, double y2)
        : Histogram(name, title, n, x1, x2), sum(0), ymin(y1), ymax(y2)
{
    int i;
    sum = new double [nx];
    err = new double [nx];
    for (i = 0;i < nx;i++)
        sum[i] = data[i] = err[i] = 0.;

}

HProfile::~HProfile()
{
    delete [] sum;
    delete [] err;
}

void HProfile::Fill(double x, double y)
{
    std::cerr << "HProfile::Fill(double x, double y) not defined" << std::endl;
}

void HProfile::Fill(double x, double y, double w)
{
    int i;
    double e;
    if ( y <= ymin || y >= ymax )
        return;

    i = Bin(x);
    tsum += w;
    if ( i >= 0 && i < nx )
    {
        e = err[i] * err[i];
        e = ( e + data[i] * data[i] ) * sum[i] + w * y * y;
        data[i] = ( data[i] * sum[i] + y * w ) / ( sum[i] + w );
        sum[i] += w;
        err[i] = sqrt(fabs(e / sum[i] - data[i] * data[i]));
        hsum += w;
    }
}

void HProfile::Print(std::ostream &os) const
{
    int i;
    os << 1 << " "
    << nx << " "
    << xmin << " "
    << xmax << " "
    << ymin << " "
    << ymax << std::endl
    << name << std::endl
    << title << std::endl;

    std::ios::fmtflags iflgs = os.flags();
    os.setf( std::ios::scientific );
    for (i = 0;i < nx;i++)
    {
        if ( i && ! (i % 6 ) )
            os << std::endl;
        os << std::setw(12) << std::setprecision(6) << data[i] << " ";
    }
    os << std::endl;
    for (i = 0;i < nx;i++)
    {
        if ( i && ! (i % 6 ) )
            os << std::endl;
        os << std::setw(12) << std::setprecision(6) << err[i] << " ";
    }
    os << std::endl;
    os.flags( iflgs );

}
