/*
 * client.cc
 *
 *  Created on: Dec 21, 2010
 *      Author: lacasta
 */

#include <cstdlib>
#include <cstdio>
#include <string>
#include <iterator>
#include <unistd.h>
#include <DAQ++/logger.h>
#include <DAQ++/GTimer.h>
#include <DAQ++/soap/daqppClient.h>
#include "Histogram.h"

double randflat(double a,double b) {
   double x = ((double)rand())/((double)RAND_MAX);
   return a+x*(b-a);
}


class MyClient : public daqppClient
{
    public:
        typedef std::vector<Histogram *> HstList;

    private:
        HstList histograms;

    public:
        MyClient(const char *uri)
            : daqppClient(uri) {}

        ~MyClient();

        Histogram *get_histogram(int ix) { return histograms[ix]; }
        int nhst() const { return histograms.size(); }
        void reset_histograms();

        void monitor_data(std::istream &is);
};

MyClient::~MyClient()
{
    reset_histograms();
}

struct HstEraser : public std::unary_function<Histogram *, void>
{
        HstEraser() {}
        void operator() (Histogram *hst) {
            delete hst;
        }
};
void MyClient::reset_histograms()
{
    std::for_each(histograms.begin(), histograms.end(), HstEraser() );
    histograms.clear();
}

void MyClient::monitor_data(std::istream &is)
{
    Histogram *hst = new Histogram("a", "b", 1, 0., 1.);
    hst->read(is);
    histograms.push_back(hst);
}


int main(int argc, char **argv)
{
    DAQpp::ostream odaq;
    DAQpp::RunState status;
    MyClient client("http://localhost:50000");

    /*
     * Send an Udate request to clear the past history...
     */
    daqppClient::StateList stateList;
    daqppClient::ParameterList parList;
    client.sendUpdate(parList, stateList);

    daqppClient::NameList modules;
    client.getModuleList(modules);

    std::vector<std::string>::iterator ip, jp;
    odaq << "List modules and their parameters" << std::endl;
    for ( ip=modules.begin(); ip!=modules.end(); ip++)
    {
        std::cout << "Module: " << *ip << std::endl;

        client.getStatus(*ip, status);
        std::cout << "Status: " << status.get_status()
                  << " time: " << status.time()
                  << " Ntrig: " << status.n_trigger()
                  <<  std::endl;

        daqppClient::NameList parlist;
        client.getParametersList(*ip, parlist);
        for (jp=parlist.begin(); jp!=parlist.end(); jp++)
        {
            std::cout << "\tPar: " << *jp << std::endl;
        }
    }


    // Test parameters
    std::vector<double> pedestals;
    client.get_vector_value<double, xsd__double>("mod.1", "pedestals", pedestals);
    std::cout << "Pedestals" << std::endl;
    std::cout << "<";
    std::copy(pedestals.begin(), pedestals.end(), std::ostream_iterator<double>(std::cout, " "));
    std::cout << ">" << std::endl;

    std::cout << "Random: "
              << client.get_value<bool, xsd__boolean>("mod.1", "random")
              << std::endl;


    double mean = client.get_value<double, xsd__double>("mod.1", "mean");
    std::cout << "mean: " << mean  << std::endl;
    for (int ival=0; ival<10;ival++)
    {
        double xval = randflat(0.75*mean, 1.25*mean);
        std::cout << " setting mean to " << xval;
        client.set_value<double, xsd__double>("mod.1", "mean", xval);
        std::cout << " --> reading mean " << client.get_value<double, xsd__double>("mod.1", "mean")
                  << std::endl;
    }
    std::cout << "Setting mean to original value: " << mean << std::endl;
    client.set_value<double, xsd__double>("mod.1", "mean", mean);

    // Test runs
    int nevts = 100000;
    odaq << DAQpp::loglevel(DAQpp::Log::message) << "Start a " << nevts << " events run" << std::endl;
    client.setMaxEvents("main", nevts);
    client.GetReady("main");

    DAQpp::GTimer timer;
    timer.start();
    client.StartRunManager("main");
    while (timer()<5.0)
    {
        client.wait(0.5);

        client.getStatus("main", status);
        std::cout << '\r' << status << std::flush;
        if ( !status.isRunning() )
            break;
    }
    timer.stop();
    std::cout << std::endl;
    client.StopRunManager("main");

    // Now we start a long run that we will stop from here
    odaq << DAQpp::loglevel(DAQpp::Log::message) << "Start a run lasting 5 seconds" << std::endl;
    client.setMaxEvents("main", 0);
    client.StartRunManager("main");
    timer.start();
    while (timer()<5.0)
    {
        client.wait(0.5);
        client.getStatus("main", status);
        std::cout << '\r' << status << std::flush;
    }

    std::cout << "\n" << std::endl;
    client.StopRunManager("main");

    client.getStatus("main", status);
    odaq << "Status after stop:\n" << status << std::endl;


    odaq << "Start with monitor data loop" << std::endl;
    client.getMonitorData("mod.1");
    if (client.nhst())
    {
        client.get_histogram(0)->Print(std::cout);
    }


    /*
     * try the Update requests
     */
    daqppClient::StateList::iterator is;
    daqppClient::ParameterList::iterator pp;
    daqppClient::empty_list<DAQpp::RunState>(stateList);
    daqppClient::empty_list<DAQpp::Parameter>(parList);

    client.sendUpdate(parList, stateList);
    std::cout << "+-- State Updates" << std::endl;
    for (is=stateList.begin(); is != stateList.end(); ++is)
    {
        DAQpp::RunState *st = *is;
        std::cout << "|->  DAQobject " << st->get_daqid() << " / " << st->get_name() << std::endl
                  << "     " << *st << std::endl;
    }

    std::cout << "+-- Parameter Updates" << std::endl;
    for (pp=parList.begin(); pp!=parList.end(); ++pp)
    {
        DAQpp::Parameter *P = *pp;
        std::cout << "|->  Param. " << P->get_name() << " from " << P->get_id() << " : " << P->get_as_string() << std::endl;
    }
    if (argv[1])
        client.stopServer();

    return 0;
}

