/*
 * server.cc
 *
 *  Created on: 20 Dec 2010
 *      Author: lacasta
 */
#include <iostream>
#include <iomanip>
#include <cmath>
#include <sstream>
#include <cstdio>
#include <getopt.h>
#include <DAQ++/soap/daqppService.h>
#include <DAQ++/soap/logging.h>
#include "Histogram.h"

#include <DAQ++/DAQmanager.h>
#include <DAQ++/RunManager.h>
#include <DAQ++/Module.h>
#include <DAQ++/Monitor.h>
#include <DAQ++/GTimer.h>
#include <DAQ++/logger.h>

//#define printf(...)




double randflat(double a, double b)
{
    double x = ((double) rand()) / ((double) RAND_MAX);
    return a + x * (b - a);
}
double randuniform()
{
    double x = ((double) rand()) / ((double) RAND_MAX);
    return x;
}
double randgauss(double mean, double stdev)
{
    static int haveCachedNormal = 0;
    static double cachedNormal = 0.;
    static double pMean = 0., pStdDev = 0.;

    if (mean != pMean || pStdDev != stdev)
    {
        haveCachedNormal = 0;
        pMean = mean;
        pStdDev = stdev;
    }
    if (haveCachedNormal == 1)
    {
        haveCachedNormal = 0;
        return (cachedNormal * pStdDev + pMean);
    }
    else
    {
        for (;;)
        {
            double u1 = randuniform();
            double u2 = randuniform();
            double v1 = 2 * u1 - 1;
            double v2 = 2 * u2 - 1;
            double w = (v1 * v1) + (v2 * v2);

            /*
             *  We actually generate two IID normal distribution variables.
             *  We cache the one & return the other.
             */
            if (w <= 1)
            {
                double y = sqrt((-2 * log(w)) / w);
                double x1 = v1 * y;
                double x2 = v2 * y;

                haveCachedNormal = 1;
                cachedNormal = x2;
                return (x1 * pStdDev + pMean);
            }
        }
    }

}





class MyMonitor : public DAQpp::Monitor
{
    private:
        unsigned long counter;
        DAQpp::RunManager *rm;
        DAQpp::GTimer timer;
    public:
        MyMonitor(DAQpp::RunManager *r, int size = 50)
            : DAQpp::Monitor(size), counter(0), rm(r)
        {}
        void on_start();
        void on_stop();
        bool user_run(DAQpp::Event *evt);
};

void MyMonitor::on_start()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info) << "*** Starting the Monitor";
    timer.reset();
    timer.start();
}

void MyMonitor::on_stop()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::none) << std::endl
       << DAQpp::loglevel(DAQpp::Log::info)
       << "*** Monitor stopped"
       << std::endl;
}

bool MyMonitor::user_run(DAQpp::Event *evt)
{
    DAQpp::ostream os;
    if ( timer() > 0.1 )
    {

        DAQpp::RunState status = rm->GetStatus();
        os << DAQpp::loglevel(DAQpp::Log::none) << '\r' << status << std::flush;
        timer.reset();
        if (!status.isRunning())
            return false;

        counter ++;
    }
    return true;
}



class SimpleRunManager: public DAQpp::RunManager
{
    public:
        SimpleRunManager(const DAQpp::DAQid &id, DAQpp::RunCommand *rc = 0);
        virtual ~SimpleRunManager();

        void GetReady();
        void PrepareForRun();
        void Stop();
        void Start();
        void EndOfRun();
};

class SimpleModule: public DAQpp::Module
{
    private:
        Histogram *hst;
    public:
        SimpleModule(const DAQpp::DAQid &id, double mean, double std);
        virtual ~SimpleModule();

        bool hasData();
        char *getData(int &size);

        int PrepareForRun();
        int Start(int rc = 0);

        int Stop(int rc = 0);

        int Pause();
        int Continue();
        int Reset();
        int GetReady();
        int Trigger();
        void flush();

        bool get_monitor_data(int &size, char **data);
        void reset_monitor_data();
};

SimpleRunManager::SimpleRunManager(const DAQpp::DAQid &id, DAQpp::RunCommand *rc) :
    DAQpp::RunManager(id, rc)
{
    set_monitor(new MyMonitor(this, 500));
    get_monitor()->set_user_timeout(100000);
}

SimpleRunManager::~SimpleRunManager()
{
    DAQpp::ostream os;
        os << DAQpp::loglevel(DAQpp::Log::info)
           << "Destroying " << get_id()
           << std::endl;
}

void SimpleRunManager::GetReady()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << get_id() << "-> GetReady" << std::endl;
    RunManager::GetReady();
}
void SimpleRunManager::PrepareForRun()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << get_id() << "-> Starting the run" << std::endl;
    RunManager::PrepareForRun();
}
void SimpleRunManager::EndOfRun()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << get_id() << "-> End of run" << std::endl;
    RunManager::EndOfRun();
}
void SimpleRunManager::Stop()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << get_id() << "-> Stop" << std::endl;
    RunManager::Stop();
}
void SimpleRunManager::Start()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << get_id() << "-> Start" << std::endl;
    RunManager::Start();
}


SimpleModule::SimpleModule(const DAQpp::DAQid &id, double mean, double std) :
    DAQpp::Module(id)
{
    add_par("random", new DAQpp::Par<bool>(true), &id);
    add_par("mean", new DAQpp::Par<double>(mean));
    add_par("std", new DAQpp::Par<double>(std));
    add_par("name", new DAQpp::Par<std::string>(id));
    add_par("pedestals", new DAQpp::Par< std::vector<double> >());
    std::fill_n( std::back_inserter( get_par_value< std::vector<double> >("pedestals") ), 32, 0.);

    int i;
    std::vector<double> &P = get_par_value< std::vector<double> >("pedestals");
    for (i=0;i<32; i++)
        P[i] = double(i);

    hst = new Histogram(get_id(), "gauss", 100, mean-5.0*std, mean+5.0*std);
}

SimpleModule::~SimpleModule()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "Destroying " << get_id()
       << std::endl;

    delete hst;
}

bool SimpleModule::hasData()
{
    return true;
}

char *SimpleModule::getData(int &size)
{
    char   *buff = allocate(sizeof(double));
    double *val  = reinterpret_cast<double *>(buff);
    double _mean = DAQpp::get_par_value<double>(get_par("mean"));
    double _std  = DAQpp::get_par_value<double>(get_par("std"));
    *val = randgauss ( _mean, _std );

    hst->Fill( *val, 1.);

    size = sizeof(double);
    return buff;
}


int SimpleModule::Trigger()
{
    return 0;
}
void SimpleModule::flush()
{
}

int SimpleModule::PrepareForRun ()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.PrepareForRun [" << get_id() << "]"
       << std::endl;
    return 0;
}

int SimpleModule::Start (int rc)
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.Start         [" << get_id() << "]"
       << std::endl;
    return 0;
}

int SimpleModule::Stop (int rc)
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.Stop          [" << get_id() << "]"
       << std::endl;
    return 0;
}

int SimpleModule::Pause ()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.Pause         [" << get_id() << "]"
       << std::endl;
    return 0;
}
int SimpleModule::Continue ()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.Continue      [" << get_id() << "]"
       << std::endl;
    return 0;
}
int SimpleModule::Reset ()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.Reset         [" << get_id() << "]"
       << std::endl;
    return 0;
}
int SimpleModule::GetReady ()
{
    DAQpp::ostream os;
    os << DAQpp::loglevel(DAQpp::Log::info)
       << "\t+-> Module.GetReady      [" << get_id() << "]"
       << std::endl;
    return 0;
}


bool SimpleModule::get_monitor_data(int &size, char **data)
{
    std::ostringstream os;
    size = hst->write(os);

    *data = new char[size];

    memcpy(*data, os.str().c_str(), size);
    return true;
}



void SimpleModule::reset_monitor_data()
{
    DAQpp::ostream odaq;
    hst->reset();
}


int main(int argc, char **argv)
{
    // Create the server
    daqppService server;
    int debug_in=0, debug_out=0;
    int option_index;
    static struct option long_options[] =
    {
     { "debug_in", 0, &debug_in, 1 },
     { "debug_out", 0, &debug_out, 1 },
     { 0, 0, 0, 0 }
    };
    while ( 1)
    {
        char c;
        option_index = 0;
        c = getopt_long(argc, argv, "", long_options, &option_index);
        if (c == -1)
            break;

        switch (c)
        {
            case 0:
                break;
            default:
                std::cout << "Error " << argv[optind-1] << " ";
                if (optind<argc)
                    std::cout << argv[optind];
                std::cout << " [Ughhhh !!!!]" << std::endl;
                return 1;

        }
    }

    soap_register_plugin(&server, logging);
    if ( debug_out)
        soap_set_logging_outbound(&server, stdout);

    if (debug_in)
        soap_set_logging_inbound(&server, stdout);


    // Create the run manager
    SimpleRunManager *run_manager = new SimpleRunManager("main");

    // Add the modules
    run_manager->addModule( new SimpleModule("mod.1", -2., 0.25) );
    run_manager->addModule( new SimpleModule("mod.2", -0., 1.5) );
    run_manager->addModule( new SimpleModule("mod.3", 2., 0.25) );

    DAQpp::DAQmanager *daq_manager= DAQpp::DAQmanager::theDAQmanager();
    daq_manager->StartRun(DAQpp::DAQmanager::Loop);
    server.run(50000);
    daq_manager->StopRun();

    return 0;
}

