#include <iostream>
#include <iomanip>
#include <csignal>
#include <cstdlib>
#include <DAQ++/DAQmanager.h>
#include <sys/resource.h>

#include "DataLogger.h"
#include "utils.h"
#include "MyModule.h"


bool DoRun = true;
void quit(int )
{
//   signal(SIGINT,SIG_DFL);
    std::cout << std::endl << "Run stop forced" << std::endl;
    DoRun = false;
}


/**
   This is an implementation of a run manager.
   It mantains a pointer to a DataLogger that gets activated everytime
   a new run starts.
*/
class Rmanager : public DAQpp::RunManager
{
    private:
        DataLogger *logger;
        double freq;
        unsigned long nloop;
    public:
        Rmanager(DAQpp::DAQid id, DAQpp::RunCommand *rc = 0) ;
        ~Rmanager();
        void Start();
        void Stop();
        void Trigger();
        void set_freq(double x = -1);

        //---------------------------------------------------------------------
        // IO stuff
        //---------------------------------------------------------------------
        int writeData(int size, const char *data);
        void log_data(const char *file);
        void stop_logging();

        DataLogger *get_logger()
        {
            return logger;
        }
};
Rmanager::Rmanager(DAQpp::DAQid id, DAQpp::RunCommand *rc)
        : DAQpp::RunManager(id, rc), logger(0), nloop(0)
{
}

Rmanager::~Rmanager()
{
    std::cout << "...destroying RunManager " << get_id() << std::endl;
}

void Rmanager::Start()
{
    std::cout << "Starting the run" << std::endl;

    if ( get_logger() )
        get_logger()->start_logging();

    DAQpp::RunManager::Start();
}

void Rmanager::Stop()
{
    DAQpp::RunManager::Stop();

    stop_logging();

    std::cout << "\nEnd of run" << std::endl;
}

void Rmanager::log_data(const char *file)
{
    if ( file )
    {
        logger = new DataLogger(this, 131072*sizeof(unsigned long),
                                true, 50000000);
        logger->open(file);
    }
}

void Rmanager::stop_logging()
{
    if ( logger )
    {
        logger->stop_logging();
        delete logger;
        logger = 0;
    }
}

int Rmanager::writeData(int size, const char *data)
{
    int rc = 0;
    if ( get_logger() )
    {
        int sz  = get_logger()->writeData(size, data);
        if (sz<=0)
           return 0;
        rc += sz;
    }
    else
        rc = size;

    return rc;
}

void Rmanager::Trigger()
{
    volatile unsigned long iloop;
    for (iloop = 0;iloop < nloop;iloop++);
}

void Rmanager::set_freq(double x)
{
    DAQpp::GTimer timer;
    if (x > 0.)
        freq = x;

    nloop = 100000000;
    timer.start();
    Trigger();
    timer.stop();
    nloop = (unsigned long)(((double)nloop) / timer() / freq);
    std::cout << "nloop " << nloop << std::endl;
}

void run(Rmanager *run_manager, int nevts, const char *ofile)
{
    DAQpp::DAQmanager *daq_manager= DAQpp::DAQmanager::theDAQmanager();

    run_manager->set_max_events(nevts);
    run_manager->Activate();
    if (ofile)
        run_manager->log_data(ofile);

    daq_manager->StartRun(DAQpp::DAQmanager::Threaded);
    while ( DoRun &&  daq_manager->isRunning())
    {
        DAQpp::wait(0.1);
        if ( !daq_manager->isRunning() )
            break;
        std:: cout << '\r' << run_manager->GetStatus();
        if ( run_manager->get_logger() )
            std::cout << " size "<< std::setw(6) << std::setprecision(1)<< 100.*run_manager->get_logger()->relative_size() << '%';

        std::cout << std::flush;
    }
    daq_manager->StopRun();

    std::cout << std::endl<< "--> Run stats. \n"<< "---------------\n"<< std::endl;
    std::cout << *run_manager << std::endl;
}


int main (int argc, char **argv)
{
    int nwords = 131;
    double freq = 50000.;

    if (argv[1])
        nwords = atoi(argv[1]);

    if (argc > 2)
    {
        freq = atof(argv[2]);
        std::cout << "Setting frequency to: " << freq << std::endl;
    }

    std::cout << "Each event will have " << nwords << " words" << std::endl;
    // catch control-c
    signal(SIGINT, quit);

    Rmanager *run_manager = new Rmanager("main");
    run_manager->addModule( new MyModule("100", 10) );

    run_manager->set_freq(freq);


//---------------------------------------------------------------------
// Start the DAQ
//---------------------------------------------------------------------

    // We first run logging data
    std::cout << "\n\n### Start a run logging data" << std::endl;
    run(run_manager, 100000, "/tmp/tstfifo.dat");

    // Now we do not log data
    std::cout << "\n\b### Start a run without logging data" << std::endl;
    run(run_manager, 100000, 0);


    return 0;
}
