#include <iostream>
#include <string>
#include <DAQ++/Parameter.h>
/**
 * This program tests the Parameter objects and tools
 *
 */

// This class tests the ParameterCollection object to bridge the
// Parameter notifications with methods of this class
class ParamBridge: public DAQpp::ParameterHolder,
					public DAQpp::ParameterCollection<ParamBridge>
{
public:
    ParamBridge();

    void change_state(const int &);
    void change_value(const double &);
    void change_name(const std::string &);
};
ParamBridge::ParamBridge()
        :  DAQpp::ParameterCollection<ParamBridge>(this)
{
    add_par( "state",
             new_par<int>(&ParamBridge::change_state,
                          new DAQpp::Par<int>(0)));

    add_par( "value",
             new_par<double>(&ParamBridge::change_value,
                             new DAQpp::Par<double>(0.)));

    add_par( "name",
             new_par<std::string>(&ParamBridge::change_name,
                                  new DAQpp::Par<std::string>("name")));
}
void ParamBridge::change_state(const int &value)
{
    std::cout << "...ParamBridge: parameter state changed to "
		      << value
		      << std::endl;
}
void ParamBridge::change_value(const double &value)
{
    std::cout << "...ParamBridge: parameter value changed to "
		      << value
		      << std::endl;
}
void ParamBridge::change_name(const std::string &value)
{
    std::cout << "...ParamBridge: parameter name changed to '"
              << value << "'"
              << std::endl;
}


/*
 * This is an observer that will register to a ParameterHolder.
 * It should receive notifications whenever a parameter changes
 */
class ParHolderObserver : public Observer
{
    public:
        ParHolderObserver() {}
        void update(Observable *o, Argument *arg)
        {
            DAQpp::Parameter *par = dynamic_cast<DAQpp::Parameter *>(o);
            if (par)
            {
                std::cout << "Holder Observer: parameter " << par->get_name() << "[" << par->get_id() << "] changed" << std::endl;
            }
            else
                std::cout << "Mmmmm, something went wrong" << std::endl;
        }
};

// The testing entry
int main(int argc, char **argv)
{
    DAQpp::ParameterHolder pH;

    double mean = 10.;

    // Define a few parameters the hard way
    pH.add_par("mean", new DAQpp::Par<double>(mean));
    pH.add_par("std", new DAQpp::Par<double>(0.));
    pH.add_par("thrs", new DAQpp::Par<int>(0));

    // Check the exceptions
    DAQpp::Parameter *par;

    std::cout << "Getting a non existing parameter" << std::endl;
    try
    {
        par = pH.get_par("kkdvk");
    }
    catch ( DAQpp::ParameterNonExistent ex)
    {
        std::cout << "*** " << ex.what() << std::endl;
    }

    std::cout << "Setting an incorrect value" << std::endl;
    try
    {
        pH.set_par<int>("mean", 2);
    }
    catch (DAQpp::ParameterTypeDoesNotMatch ex)
    {
        std::cout << "*** " << ex.what() << std::endl;
    }

    std::cout << "Setting a proper value" << std::endl;
    pH.set_par<double>("mean", 2.);
    std::cout << "... new value is "
              << pH.get_par_value<double>("mean")
              << std::endl;


    // interception a generig DAQ++ exception
    std::cout << "Try again a non existen parameter" << std::endl;
    try
    {
        par = pH.get_par("xxxx");
    }
    catch ( DAQpp::Exception ex )
    {
        std::cout << "*** " << ex << std::endl;
    }

    std::cout << "Getting the list of parameters" << std::endl;
    DAQpp::ParameterHolder::const_iterator ip;
    for (ip = pH.begin();ip != pH.end();ip++)
        std::cout << ip->first << std::endl;

    // Now we test the ParameterObserver stuff
    std::cout << "** Testing the ParameterObserver" << std::endl;
    ParamBridge Opar;

    int ival = 5;
    std::cout << "Changing parameter state to " << ival << std::endl;
    Opar.set_par("state", ival, true);

    double dval = 10.;
    std::cout << "Changing parameter value to " << dval << std::endl;
    Opar.set_par("value", dval, true);

    std::string sval("This is a string");
    std::cout << "Changing parameter name to '" << sval << "'" << std::endl;
    Opar.set_par("name", sval, true);


    // Now we test the ParameterHolder observable interface
    std::cout << "*** Now we test the ParameterHolder observable interface" << std::endl;
    ParHolderObserver pobs;
    Opar.addObserver(pobs);

    Opar.set_par("state",10, true);
    Opar.set_par("value", 22., true);


    DAQpp::DAQid par_id("XXX");
    Opar.add_par("newpar", new DAQpp::Par<double>(0.), &par_id );
    Opar.set_par("newpar",2., true);

    return 0;
}

